// -*- C++ -*-
// $Id:$
#ifndef EXFLOAT_H
#define EXFLOAT_H

#ifndef _EXFLIB_COMPILE
const char EXFLOAT_H_VERSION[] = "Time-stamp: \"2018-06-20 16:17:16 fujiwara\" ";
#endif

//----------------------------------------------------------------------
//  Class Specification of extend float (exfloat)
//
//  Copyright (C) 1999-2017, FUJIWARA Hiroshi
//        <fujiwara@acs.i.kyoto-u.ac.jp>
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// The folowing identifiers are reserved as keyword.
//
//   exfloat exflib PRECISION exflib_* EXFLIB_*
//----------------------------------------------------------------------

#ifdef _EXFLIB_COMPILE_CONST
#define _EXFLIB_COMPILE
#endif

#ifndef _EXFLIB_COMPILE

#ifndef PRECISION
#error Please Define PRECISION before including "exfloat.h".
#endif

#include <fstream>
#include <iomanip>

#endif  // _EXFLIB_COMPILE

#if defined(__BORLANDC__) || defined(_MSC_VER)
typedef const            __int64 exflib_cQword;
typedef const unsigned   __int64 exflib_cuQword;
typedef       unsigned   __int64 exflib_uQword;
typedef                  __int64 exflib_Qword;
typedef       unsigned      long exflib_ulong;
#else
typedef const          long long exflib_cQword;
typedef const unsigned long long exflib_cuQword;
typedef       unsigned long long exflib_uQword;
typedef                long long exflib_Qword;
typedef       unsigned      long exflib_ulong;
#endif

/******************************************************************/
/* Constatnt Identifier                                           */
/******************************************************************/

/* Constants have 19728.3 digit precision in decimail */

enum {
    EXFLIB_CONST_MAX_PR 	= 1024,
    EXFLIB_FUNCARG_MAX_PR64 	= 160,
    EXFLIB_FUNCARG_EXP_LEAST 	= -32,
    EXFLIB_FUNCLOG_INVTBL_MAXN 	= 500
};

#ifdef _EXFLIB_COMPILE_CONST
static long EXFLIB_TMP_PR	= (long)(EXFLIB_CONST_MAX_PR * 1.05);
#endif

#if defined(__BORLANDC__) || defined(_MSC_VER)
static exflib_cuQword EXFLIB_TEN_19	= 0x8ac7230489e80000;
static exflib_cuQword EXFLIB_TEN_18	= 0x0de0b6b3a7640000;

static exflib_cuQword EXFLIB_MASK_SGN	= 0x8000000000000000;
static exflib_cuQword EXFLIB_MASK_EXP   = 0x7fffffffffffffff;
static exflib_cuQword EXFLIB_DBL_EXP	= 0x7ff0000000000000;
static exflib_cuQword EXFLIB_DBL_FLT	= 0x000fffffffffffff;
static exflib_cQword  EXFLIB_BIAS	= 0x3fffffffffffffff;
static exflib_cuQword EXFLIB_FULL_ONE   = 0xffffffffffffffff;
#else
static exflib_cuQword EXFLIB_TEN_19	= 0x8ac7230489e80000ULL;
static exflib_cuQword EXFLIB_TEN_18	= 0x0de0b6b3a7640000ULL;

static exflib_cuQword EXFLIB_MASK_SGN	= 0x8000000000000000ULL;
static exflib_cuQword EXFLIB_MASK_EXP   = 0x7fffffffffffffffULL;
static exflib_cuQword EXFLIB_DBL_EXP	= 0x7ff0000000000000ULL;
static exflib_cuQword EXFLIB_DBL_FLT	= 0x000fffffffffffffULL;
static exflib_cQword  EXFLIB_BIAS	= 0x3fffffffffffffffLL;
static exflib_cuQword EXFLIB_FULL_ONE   = 0xffffffffffffffffULL;
#endif

/******************************************************************/
/* Prototypes                                                     */
/******************************************************************/

#ifdef __cplusplus
#include <cstddef>	// definition of size_t
extern "C" {
#else
#include <stddef.h>	/* size_t */
#endif

    /* fixed point arithmetic */
    exflib_ulong exflib_array_add(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_array_sub(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_array_mul(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_array_mul_h(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_array_mul_k(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_uQword exflib_array_imul(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b, size_t n);
    exflib_ulong exflib_array_idiv(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b, size_t n, long *top);

    /* floating point arithmetic */
    exflib_ulong exflib_float_add(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_float_sub(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_float_mul(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);
    exflib_ulong exflib_float_imul(exflib_uQword c[], exflib_cuQword a[], exflib_Qword b, size_t n);
    exflib_ulong exflib_float_idiv(exflib_uQword c[], exflib_cuQword a[], exflib_Qword b, size_t n);
    exflib_ulong exflib_float_div(exflib_uQword c[], exflib_cuQword a[], exflib_cuQword b[], size_t n);

    /* floating point rounding */
    void exflib_float_round_chop(exflib_uQword a[], exflib_ulong hint, size_t n);
    void exflib_float_round_near(exflib_uQword a[], exflib_ulong hint, size_t n);
    void exflib_float_round_plus(exflib_uQword a[], exflib_ulong hint, size_t n);
    void exflib_float_round_minus(exflib_uQword a[], exflib_ulong hint, size_t n);

    /* floating point operation */
    long exflib_float_lt(exflib_cuQword a[], exflib_cuQword b[], size_t n);
    void exflib_float_ltoef(exflib_uQword a[], long b, size_t n);

    /* output ------------------------------*/
    void exflib_convert_pow(exflib_cuQword u[], exflib_Qword *d_int, exflib_uQword target[], long pr64);
    void exflib_scientific_format(exflib_uQword flag, exflib_uQword u[],
				 exflib_Qword u_pow, char output_string[]);
    void exflib_fixed_format(exflib_uQword flag, exflib_uQword u[],
			    exflib_Qword u_pow, char output_string[]);
    /* assign ------------------------------*/
    void exflib_float_etoef(const char *s, exflib_uQword u[], long pr64);
    void exflib_float_ctoef(const char *s, exflib_uQword u[], long pr64);
    /* 
       following functions are defined, and used in etoef.

       void exflib_float_htoef(const char *s, exflib_uQword u[], long pr64);
       void exflib_float_atoef(const char *s, exflib_uQword u[], long pr64);
    */

    /* exfloat <-> double */
    void exflib_float_dtoef(const double d, exflib_uQword u[], long pr64);
    double exflib_float_eftod(exflib_cuQword u[]);

    /* built-in functions */
    exflib_ulong exflib_float_sqrt(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_sin(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_cos(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_exp(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_pow_long(exflib_uQword y[], exflib_cuQword x[], long i, long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_pow(exflib_uQword p[], exflib_cuQword x[], exflib_cuQword y[], long pr64, exflib_uQword w[], exflib_uQword *inv_table);
    exflib_ulong exflib_float_log(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[], exflib_uQword *inv_table);

    exflib_ulong exflib_float_asin(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    exflib_ulong exflib_float_atan(exflib_uQword y[], exflib_cuQword x[], long pr64, exflib_uQword w[], exflib_uQword *inv_table);

    void exflib_init_reciprocal_table(exflib_uQword tbl[], long size);

    void exflib_float_modf(exflib_cuQword x[], exflib_uQword ip[], exflib_uQword fp[], long pr64);
    void exflib_float_fmod(exflib_uQword r[], exflib_cuQword d[], exflib_cuQword m[], long pr64, exflib_uQword w[]);
    void exflib_float_ceil(exflib_uQword ip[], exflib_cuQword x[], long pr64, exflib_uQword w[]);
    void exflib_float_floor(exflib_uQword ip[], exflib_cuQword x[], long pr64, exflib_uQword w[]);

    /* for atan functions */
    long exflib_atan_arg_reduction(exflib_uQword x[], long pr64);
    void exflib_atan_recover(exflib_uQword x[], long k, long pr64);
    void exflib_atan_horner(exflib_uQword x[], exflib_uQword v[], exflib_uQword table[], long pr64);

    /* utility */
//    void exflib_version();
    const char *exflib_version();
//    void exflib_float_rand(exflib_uQword x[], long pr64);

#ifdef __cplusplus
}; // end of extern "C"
#endif

/******************************************************************/
/* decimal output format flag                                     */
/*                                                                */
/* A flag is type unsigned long. It is consists of 6 parts.       */
/*   (Each characters represents 4 bit.)                          */
/*                                                                */
/*    0x|   0|     1|      23|       4567|      89ab|  cdef|      */
/*      |sign|format|(unused)|field width|decimal pr|hex pr|      */
/*                                                                */
/* Field width, decimal precision & hex precision is 16 bit.      */
/*   (Max precision is 65536.)                                    */
/******************************************************************/

#if defined(__BORLANDC__) || defined(_MSC_VER)
static exflib_cuQword EXFLIB_SIGN     = 0x8000000000000000;
static exflib_cuQword EXFLIB_PLUS     = 0x4000000000000000;	// not supported
static exflib_cuQword EXFLIB_SPACE    = 0x2000000000000000;	// not supported
static exflib_cuQword EXFLIB_LEFT     = 0x1000000000000000;	// not supported

static exflib_cuQword EXFLIB_G_FORMAT = 0x0800000000000000;
static exflib_cuQword EXFLIB_F_FORMAT = 0x0400000000000000;
static exflib_cuQword EXFLIB_E_FORMAT = 0x0200000000000000;
static exflib_cuQword EXFLIB_CAPITAL  = 0x0100000000000000;	// not supported
#else
static exflib_cuQword EXFLIB_SIGN     = 0x8000000000000000ULL;
static exflib_cuQword EXFLIB_PLUS     = 0x4000000000000000ULL;	// not supported
static exflib_cuQword EXFLIB_SPACE    = 0x2000000000000000ULL;	// not supported
static exflib_cuQword EXFLIB_LEFT     = 0x1000000000000000ULL;	// not supported

static exflib_cuQword EXFLIB_G_FORMAT = 0x0800000000000000ULL;
static exflib_cuQword EXFLIB_F_FORMAT = 0x0400000000000000ULL;
static exflib_cuQword EXFLIB_E_FORMAT = 0x0200000000000000ULL;
static exflib_cuQword EXFLIB_CAPITAL  = 0x0100000000000000ULL;	// not supported
#endif

#ifndef _EXFLIB_COMPILE

struct exfloat {

    // requested precision in decimal
    static const long precision10 = (PRECISION >= 20)? PRECISION: 20;

    // precision in 2^64 radix
    // log(10)/log(2)/64 = 0.051905126
#if ((!defined(__SUNPRO_CC)) && (!defined(__FCC_VERSION))) && (!defined(__CRAYXE))
    static const long precision64
      = static_cast<long>(precision10 * 0.051905126) + 1;
#else
    // for SunStudio, FUJITSU CC
    static const long precision64
      = static_cast<long>(precision10 * 51905126 / 1000000000) + 1;
#endif    

    // size (sum of exponent part and fraction part)
    static const long size = 1 + precision64;
    static const long byte = 8*size;

    exflib_uQword num[size];

    // effectively processing precision in decimal
    // 64*log(2) = 19.26591972
    static const double precision;

    //----------------------------------------------------------------------
    // rounding operations
    //----------------------------------------------------------------------
    static void (*round)(exflib_uQword a[], exflib_ulong hint, size_t n);
    static void setround(long mode){
        if(mode == 0)
            round = exflib_float_round_near;
        else if(mode > 0)
            round = exflib_float_round_plus;
        else
            round = exflib_float_round_minus;
        return;
    }

    //----------------------------------------------------------------------
    // constructor
    //----------------------------------------------------------------------
    exfloat() { num[0] = 0; }
    exfloat(const char *op){
	exflib_float_etoef(op, num, precision64);
	/* rounding */
    }
    exfloat(long op){ exflib_float_ltoef(num, op, precision64); }
    exfloat(int op){ exflib_float_ltoef(num, static_cast<long>(op), precision64); }
    explicit exfloat(const double op){ exflib_float_dtoef(op, num, precision64); }
    // exfloat(const exfloat& op);

    //----------------------------------------------------------------------
    // assignment operator
    //----------------------------------------------------------------------
    inline exfloat &operator=(const exfloat& op2);
    // inline exfloat operator=(double op2);
    inline exfloat &operator=(long op2);
    inline exfloat &operator=(int op2);
    inline exfloat &operator=(const char *s);

    inline exfloat operator+=(const exfloat& op2);
    inline exfloat operator+=(long op2);
    inline exfloat operator+=(int op2);

    inline exfloat operator-=(const exfloat& op2);
    inline exfloat operator-=(long op2);
    inline exfloat operator-=(int op2);

    inline exfloat operator*=(const exfloat& op2);
    inline exfloat operator*=(long op2);
    inline exfloat operator*=(int op2);

    inline exfloat operator/=(const exfloat& op2);
    inline exfloat operator/=(long op2);
    inline exfloat operator/=(int op2);

    inline exfloat operator<<=(int op2);
    inline exfloat operator<<=(long op2);
    inline exfloat operator>>=(int op2);
    inline exfloat operator>>=(long op2);

    //----------------------------------------------------------------------
    // conversion functions
    //----------------------------------------------------------------------
    operator double() const { return exflib_float_eftod(num); }
    // operator char*() const;
    operator bool() const { return (num[0] & EXFLIB_MASK_EXP); }

    //----------------------------------------------------------------------
    // member functions
    //----------------------------------------------------------------------
    // return true if *this == +0 or *this == -0.
    bool iszero() const {
	return !(num[0] & EXFLIB_MASK_EXP);
    }
    bool isZero() const {
	return !(num[0] & EXFLIB_MASK_EXP);
    }
    // return true if *this == +\infty or *this == -\infty.
    bool isInfty() const {
	exflib_uQword infty = num[0] & EXFLIB_MASK_EXP;
	return (infty == EXFLIB_MASK_EXP);
    }
    bool isinfty() const {
	exflib_uQword infty = num[0] & EXFLIB_MASK_EXP;
	return (infty == EXFLIB_MASK_EXP);
    }

    //----------------------------------------------------------------------
    // machine epsilon
    //----------------------------------------------------------------------
    static exfloat epsilon(){
	exfloat eps;
	eps.num[0] = EXFLIB_BIAS - 64*(exfloat::size - 1);
	for(int i = 1; i < exfloat::size; i++)
	    eps.num[i] = 0;
	return eps;
    }

    //----------------------------------------------------------------------
    // arithmetic operator
    //----------------------------------------------------------------------
    // unary
    inline exfloat operator+() const;
    inline exfloat operator-() const;

    // binary
    inline exfloat operator+(const exfloat& op2) const;
    inline exfloat operator+(long op2) const;
    inline exfloat operator+(int op2) const;
    friend inline exfloat operator+(long op, const exfloat& op2);
    friend inline exfloat operator+(int op, const exfloat& op2);

    inline exfloat operator-(const exfloat& op2) const;
    inline exfloat operator-(long op2) const;
    inline exfloat operator-(int op2) const;
    friend inline exfloat operator-(long op, const exfloat& op2);
    friend inline exfloat operator-(int op, const exfloat& op2);

    inline exfloat operator*(const exfloat& op2) const;
    inline exfloat operator*(long op2) const;
    inline exfloat operator*(int op2) const;
    friend inline exfloat operator*(long op, const exfloat& op2);
    friend inline exfloat operator*(int op, const exfloat& op2);

    inline exfloat operator/(const exfloat& op2) const;
    inline exfloat operator/(long op2) const;
    inline exfloat operator/(int op2) const;
    friend inline exfloat operator/(long op, const exfloat& op2);
    friend inline exfloat operator/(int op, const exfloat& op2);

    inline exfloat operator<<(long op2) const;
    inline exfloat operator<<(int op2) const;
    inline exfloat operator>>(long op2) const;
    inline exfloat operator>>(int op2) const;

    //----------------------------------------------------------------------
    // relation operator
    //----------------------------------------------------------------------
    inline bool operator==(const exfloat& op2) const;
    inline bool operator==(double op2) const;
    inline bool operator==(long op2) const;
    inline bool operator==(int op2) const;
    friend inline bool operator==(double op, const exfloat& op2);
    friend inline bool operator==(long op, const exfloat& op2);
    friend inline bool operator==(int op, const exfloat& op2);

    inline bool operator!=(const exfloat& op2) const;
    inline bool operator!=(double op2) const;
    inline bool operator!=(long op2) const;
    inline bool operator!=(int op2) const;
    friend inline bool operator!=(double op, const exfloat& op2);
    friend inline bool operator!=(long op, const exfloat& op2);
    friend inline bool operator!=(int op, const exfloat& op2);

    inline bool operator<(const exfloat& op2) const;
    inline bool operator<(double op2) const;
    inline bool operator<(long op2) const;
    inline bool operator<(int op2) const;
    friend inline bool operator<(double op, const exfloat& op2);
    friend inline bool operator<(long op, const exfloat& op2);
    friend inline bool operator<(int op, const exfloat& op2);

    inline bool operator<=(const exfloat& op2) const;
    inline bool operator<=(double op2) const;
    inline bool operator<=(long op2) const;
    inline bool operator<=(int op2) const;
    friend inline bool operator<=(double op, const exfloat& op2);
    friend inline bool operator<=(long op, const exfloat& op2);
    friend inline bool operator<=(int op, const exfloat& op2);

    inline bool operator>(const exfloat& op2) const;
    inline bool operator>(double op2) const;
    inline bool operator>(long op2) const;
    inline bool operator>(int op2) const;
    friend inline bool operator>(double op, const exfloat& op2);
    friend inline bool operator>(long op, const exfloat& op2);
    friend inline bool operator>(int op, const exfloat& op2);

    inline bool operator>=(const exfloat& op2) const;
    inline bool operator>=(double op2) const;
    inline bool operator>=(long op2) const;
    inline bool operator>=(int op2) const;
    friend inline bool operator>=(double op, const exfloat& op2);
    friend inline bool operator>=(long op, const exfloat& op2);
    friend inline bool operator>=(int op, const exfloat& op2);

    //----------------------------------------------------------------------
    // output
    //----------------------------------------------------------------------
    friend std::ostream& operator<<(std::ostream& strm, const exfloat& op);

    //----------------------------------------------------------------------
    // mathematical functions
    //----------------------------------------------------------------------
    friend exfloat abs(const exfloat& x);
    friend exfloat fabs(const exfloat& x);
    friend exfloat sqrt(const exfloat& x);
    friend exfloat sin(const exfloat& x);
    friend exfloat cos(const exfloat& x);
    friend exfloat tan(const exfloat& x);
    friend exfloat exp(const exfloat& x);
    friend exfloat pow(const exfloat& x, long y);
    friend exfloat pow(const exfloat& x, const exfloat& y);
    friend exfloat pow(long x, const exfloat& y);
    friend exfloat log(const exfloat& x);
    friend exfloat log10(const exfloat& x);

    //----------------------------------------------------------------------
    // random number
    //----------------------------------------------------------------------
//    friend exfloat exflib_rand();

    // now implemented in C++
    friend exfloat asin(const exfloat& x);
    friend exfloat acos(const exfloat& x);
    friend exfloat atan(const exfloat& x);
    friend exfloat atan2(const exfloat &y, const exfloat &x);
    friend exfloat sinh(const exfloat& x);
    friend exfloat cosh(const exfloat& x);
    friend exfloat tanh(const exfloat& x);
    friend exfloat modf(const exfloat &x, exfloat *ip);
    friend exfloat fmod(const exfloat &x, const exfloat &y);
    friend exfloat ceil(const exfloat &x);
    friend exfloat floor(const exfloat &x);
}; // class exfloat

//----------------------------------------------------------------------
//  rounding operations, name binding
//----------------------------------------------------------------------
namespace exflib {
    void (*near)(exflib_uQword [], exflib_ulong, size_t)   = exflib_float_round_near;
    void (*chop)(exflib_uQword [], exflib_ulong, size_t)   = exflib_float_round_chop;
    void (*plus)(exflib_uQword [], exflib_ulong, size_t)   = exflib_float_round_plus;
    void (*minus)(exflib_uQword [], exflib_ulong, size_t)  = exflib_float_round_minus;
};

//----------------------------------------------------------------------
// initialization of static member
//----------------------------------------------------------------------

const double exfloat::precision = 19.26591972 * exfloat::precision64;
void (*exfloat::round)(exflib_uQword [], exflib_ulong, size_t) = exflib_float_round_near;

//----------------------------------------------------------------------
// working memory
//----------------------------------------------------------------------

namespace exflib {
    exflib_uQword *inv_table_log = 0;
    exflib_uQword *inv_table_atan = 0;
};

//----------------------------------------------------------------------
//  Arithmetic Operators for extend float (exfloat)
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// assignment operator
//----------------------------------------------------------------------

inline exfloat &exfloat::operator=(const exfloat& op2)
{
    for(int i = 0; i <= exfloat::precision64; i++)
	num[i] = op2.num[i];
    return *this;
}

inline exfloat &exfloat::operator=(long op2)
{
    exflib_float_ltoef(num, op2, exfloat::precision64);
    return *this;
}

inline exfloat &exfloat::operator=(int op2)
{
    exflib_float_ltoef(num, static_cast<long>(op2), exfloat::precision64);
    return *this;
}

inline exfloat &exfloat::operator=(const char *s)
{
    exflib_float_etoef(s, num, exfloat::precision64);
    /* rounding */
    return *this;
}

inline exfloat exfloat::operator+=(const exfloat& op2)
{
    exflib_ulong hint;
    hint = exflib_float_add(num, num, op2.num, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator+=(long op2)
{
    exflib_ulong hint;
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    hint = exflib_float_add(num, num, tmp, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator+=(int op2)
{
    exflib_ulong hint;
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    hint = exflib_float_add(num, num, tmp, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator-=(const exfloat& op2)
{
    exflib_ulong hint;
    hint = exflib_float_sub(num, num, op2.num, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator-=(long op2)
{
    exflib_ulong hint;
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    hint = exflib_float_sub(num, num, tmp, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator-=(int op2)
{
    exflib_ulong hint;
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    hint = exflib_float_sub(num, num, tmp, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator*=(const exfloat& op2)
{
    exflib_ulong hint;
    hint = exflib_float_mul(num, num, op2.num, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator*=(long op2)
{
    exflib_ulong hint;
    hint = exflib_float_imul(num, num, op2, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator*=(int op2)
{
    exflib_ulong hint;
    hint = exflib_float_imul(num, num, static_cast<long>(op2), exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator/=(const exfloat& op2)
{
    exflib_ulong hint;
    hint = exflib_float_div(num, num, op2.num, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator/=(long op2)
{
    exflib_ulong hint;
    hint = exflib_float_idiv(num, num, op2, exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator/=(int op2)
{
    exflib_ulong hint;
    hint = exflib_float_idiv(num, num, static_cast<long>(op2), exfloat::precision64);
    exfloat::round(num, hint, exfloat::precision64);
    return *this;
}

inline exfloat exfloat::operator<<=(int op2)
{
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	num[0] += op2;
    return *this;
}

inline exfloat exfloat::operator<<=(long op2)
{
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	num[0] += op2;
    return *this;
}

inline exfloat exfloat::operator>>=(int op2)
{
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	num[0] -= op2;
    return *this;
}

inline exfloat exfloat::operator>>=(long op2)
{
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	num[0] -= op2;
    return *this;
}

//----------------------------------------------------------------------
// arithmetic operator
//----------------------------------------------------------------------

inline exfloat exfloat::operator+() const
{
    return *this;
}

inline exfloat exfloat::operator-() const
{
    exfloat local_work;
    local_work.num[0] = num[0] + EXFLIB_MASK_SGN;
    for(int i = 1; i <= exfloat::precision64; i++)
	local_work.num[i] = num[i];
    return local_work;
}

inline exfloat exfloat::operator+(const exfloat& op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_add(local_work.num, num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator+(long op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op2;
    hint = exflib_float_add(local_work.num, num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator+(int op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op2;
    hint = exflib_float_add(local_work.num, num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator+(long op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_add(local_work.num, op2.num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator+(int op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_add(local_work.num, op2.num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator-(const exfloat& op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_sub(local_work.num, num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator-(long op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op2;
    hint = exflib_float_sub(local_work.num, num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator-(int op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op2;
    hint = exflib_float_sub(local_work.num, num, local_work.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator-(long op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_sub(local_work.num, local_work.num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator-(int op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_sub(local_work.num, local_work.num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator*(const exfloat& op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_mul(local_work.num, num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator*(long op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint =exflib_float_imul(local_work.num, num, op2, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator*(int op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_imul(local_work.num, num, static_cast<long>(op2), exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator*(long op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_imul(local_work.num, op2.num, op, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator*(int op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_imul(local_work.num, op2.num, static_cast<long>(op), exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator/(const exfloat& op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_div(local_work.num, num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator/(long op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_idiv(local_work.num, num, op2, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator/(int op2) const
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    hint = exflib_float_idiv(local_work.num, num, static_cast<long>(op2), exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator/(long op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_div(local_work.num, local_work.num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat operator/(int op, const exfloat& op2)
{
    exfloat local_work;
    exflib_ulong hint;
    exfloat &w = local_work;
    local_work = op;
    hint = exflib_float_div(local_work.num, local_work.num, op2.num, exfloat::precision64);
    exfloat::round(w.num, hint, exfloat::precision64);
    return local_work;
}

inline exfloat exfloat::operator<<(long op2) const
{
    exfloat local_work;
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	local_work.num[0] = num[0] + op2;
    for(int i = 1; i < exfloat::precision64; i++)
        local_work.num[i] = num[i];
    return local_work;
}

inline exfloat exfloat::operator<<(int op2) const
{
    exfloat local_work;
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	local_work.num[0] = num[0] + op2;
    for(int i = 1; i < exfloat::precision64; i++)
        local_work.num[i] = num[i];
    return local_work;
}

inline exfloat exfloat::operator>>(long op2) const
{
    exfloat local_work;
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	local_work.num[0] = num[0] - op2;
    for(int i = 1; i < exfloat::precision64; i++)
        local_work.num[i] = num[i];
    return local_work;
}

inline exfloat exfloat::operator>>(int op2) const
{
    exfloat local_work;
    if(((num[0] & EXFLIB_MASK_EXP) != 0)
       && ((num[0] & EXFLIB_MASK_EXP) != EXFLIB_MASK_EXP))
	local_work.num[0] = num[0] - op2;
    for(int i = 1; i < exfloat::precision64; i++)
        local_work.num[i] = num[i];
    return local_work;
}

//----------------------------------------------------------------------
// relation operator
//----------------------------------------------------------------------
inline bool exflib_float_eq(exflib_cuQword array1[], exflib_cuQword array2[])
{
    // detecting Zero
    if(((array1[0] << 1) == 0) && ((array2[0] << 1) == 0))
	return true;

    int i;
    for(i = 0; (i <= exfloat::precision64) && (array1[i] == array2[i]); i++)
	continue;
    return (i > exfloat::precision64)? true: false;
}    

inline bool exfloat::operator==(const exfloat& op2) const
{
    return exflib_float_eq(num, op2.num)? true: false;
}

inline bool exfloat::operator==(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return exflib_float_eq(num, tmp)? true: false;
}

inline bool exfloat::operator==(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return exflib_float_eq(num, tmp)? true: false;
}

inline bool exfloat::operator==(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return exflib_float_eq(num, tmp)? true: false;
}

inline bool operator==(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? true: false;
}

inline bool operator==(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? true: false;
}

inline bool operator==(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? true: false;
}

inline bool exfloat::operator!=(const exfloat& op2) const
{
    return exflib_float_eq(num, op2.num)? false: true;
}

inline bool exfloat::operator!=(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return exflib_float_eq(num, tmp)? false: true;
}

inline bool exfloat::operator!=(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return exflib_float_eq(num, tmp)? false: true;
}

inline bool exfloat::operator!=(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return exflib_float_eq(num, tmp)? false: true;
}

inline bool operator!=(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? false: true;
}

inline bool operator!=(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? false: true;
}

inline bool operator!=(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return exflib_float_eq(tmp, op2.num)? false: true;
}

inline bool exfloat::operator<(const exfloat& op2) const
{
    return exflib_float_lt(num, op2.num, exfloat::precision64);
}

inline bool exfloat::operator<(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool exfloat::operator<(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool exfloat::operator<(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool operator<(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

inline bool operator<(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

inline bool operator<(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

inline bool exfloat::operator<=(const exfloat& op2) const
{
    return !exflib_float_lt(op2.num, num, exfloat::precision64);
}
    
inline bool exfloat::operator<=(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return !exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool exfloat::operator<=(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return !exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool exfloat::operator<=(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return !exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool operator<=(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return !exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool operator<=(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return !exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool operator<=(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return !exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool exfloat::operator>(const exfloat& op2) const
{
    return exflib_float_lt(op2.num, num, exfloat::precision64);
}

inline bool exfloat::operator>(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool exfloat::operator>(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool exfloat::operator>(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return exflib_float_lt(tmp, num, exfloat::precision64);
}

inline bool operator>(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool operator>(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool operator>(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return exflib_float_lt(op2.num, tmp, exfloat::precision64);
}

inline bool exfloat::operator>=(const exfloat& op2) const
{
    return !exflib_float_lt(num, op2.num, exfloat::precision64);
}

inline bool exfloat::operator>=(double op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op2, tmp, exfloat::precision64);
    return !exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool exfloat::operator>=(long op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op2, exfloat::precision64);
    return !exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool exfloat::operator>=(int op2) const
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op2), exfloat::precision64);
    return !exflib_float_lt(num, tmp, exfloat::precision64);
}

inline bool operator>=(double op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_dtoef(op, tmp, exfloat::precision64);
    return !exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

inline bool operator>=(long op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, op, exfloat::precision64);
    return !exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

inline bool operator>=(int op, const exfloat& op2)
{
    exflib_uQword tmp[exfloat::size];
    exflib_float_ltoef(tmp, static_cast<long>(op), exfloat::precision64);
    return !exflib_float_lt(tmp, op2.num, exfloat::precision64);
}

//----------------------------------------------------------------------
// output
//----------------------------------------------------------------------

std::ostream &operator<<(std::ostream &strm, const exfloat& op)
{
    std::ios::fmtflags strm_flags = strm.flags();

    //-----------------------------------------------------------------
    // dump
    //-----------------------------------------------------------------
    if(strm_flags & std::ios::hex){

	for(int i = 0; i < exfloat::precision64; i++)
	    strm << std::hex << std::setw(16) << std::setfill('0') << op.num[i] << ':';
	strm << std::hex << std::setw(16) << std::setfill('0') << op.num[exfloat::precision64] << std::dec;

	return strm;
    }

    //-----------------------------------------------------------------
    // decimal output
    //-----------------------------------------------------------------
    if(op.num[0] == EXFLIB_MASK_EXP){
	strm << "+Infty";
	return strm;
    }
    if(op.num[0] == EXFLIB_FULL_ONE){
	strm << "-Infty";
	return strm;
    }

    if(op.iszero()){

	long pr = strm.precision();
	char zero[65536];
	int i = 0;

	if(op.num[0] & EXFLIB_MASK_SGN)
	    zero[i++] = '-';

	if((strm_flags & std::ios::scientific) && !(strm_flags & std::ios::fixed)){
	    // scientific format
	    zero[i++] = '0';	    zero[i++] = '.';
	    for(int k = 0; k < pr; k++)
		zero[i++] = '0';
	    zero[i++] = 'e'; zero[i++] = '+'; zero[i++] = '0';
	    zero[i] = '\0';

	} else if((strm_flags & std::ios::fixed) && !(strm_flags & std::ios::scientific)){
	    // fixed format
	    zero[i++] = '0';	    zero[i++] = '.';
	    for(int k = 0; k < pr; k++)
		zero[i++] = '0';
	    zero[i] = '\0';

	} else {
	    // general format

	    zero[i++] = '0'; zero[i] = '\0';
	}

	strm << zero;
	return strm;
    }

    long strm_precision = strm.precision();

#if 0
    //-----------------------------------------------------------------
    // converted double is enough for output
    //       1. precision is less than 16
    //  and  2. the number (magnitude) is expressible in double
    //           0 < exp(double biased) < 0x7ff
    //        -> -0x3ff < exp(unbias) < 0x400
    //        -> 0x3ffffffffffffc00 < exp(exf.biased) < 0x40000000000003ff
    //-----------------------------------------------------------------
    if((strm_precision < 16) &&
       (0x3ffffffffffffc00ULL < (op.num[0] & (~EXFLIB_SIGN)))
       && ((op.num[0] & (~EXFLIB_SIGN)) < 0x40000000000003ffULL))
    {
	strm << exflib_float_eftod(op.num);
	return strm;
    }
#endif

    exflib_uQword b64[exfloat::precision64 + 2];
    exflib_Qword deci_pow;
    
    //-----------------------------------------------------------------
    // 1.(num) * 2^e --> i.(b64) * 10^d ('i' is 1 digit in radix 10)
    //-----------------------------------------------------------------
    exflib_convert_pow(op.num, &deci_pow, b64, exfloat::precision64);

    //-----------------------------------------------------------------
    // format
    //-----------------------------------------------------------------
    char deci_f[65536];
    exflib_uQword flag;

    if((strm_flags & std::ios::scientific) && !(strm_flags & std::ios::fixed)){

	// scientific format

	// make a format flag

	flag = (op.num[0] & EXFLIB_SIGN) | EXFLIB_E_FORMAT
	    | (strm_precision << 16) | (exfloat::precision64 + 2);
	
	exflib_scientific_format(flag, b64, deci_pow, deci_f);

    }else if((strm_flags & std::ios::fixed) && !(strm_flags & std::ios::scientific)){

	// fixed format

	// make a format flag
	
	flag = (op.num[0] & EXFLIB_SIGN) | EXFLIB_F_FORMAT
	    | (strm_precision << 16) | (exfloat::precision64 + 2);

	exflib_fixed_format(flag, b64, deci_pow, deci_f);

    } else {

	// general format

	if((deci_pow < -4) || (strm_precision <= deci_pow)){

	    // scientific format

	    flag = (op.num[0] & EXFLIB_SIGN) | EXFLIB_G_FORMAT
		| (((strm_precision - 1 > 0)? (strm_precision -1): 0) << 16)
		    | (exfloat::precision64 + 2);

	    exflib_scientific_format(flag, b64, deci_pow, deci_f);

	} else {

	    // fixed format

	    flag = (op.num[0] & EXFLIB_SIGN) | EXFLIB_G_FORMAT
		| ((strm_precision - deci_pow - 1) << 16)
		    | (exfloat::precision64 + 2);

	    exflib_fixed_format(flag, b64, deci_pow, deci_f);
	}

    }
    strm << deci_f;

    return strm;
}

//----------------------------------------------------------------------
//  Mathematical Functions for exfloat
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//  absolute value
//----------------------------------------------------------------------

exfloat abs(const exfloat& x)
{
    exfloat local_work;
    local_work.num[0] = x.num[0] & EXFLIB_MASK_EXP;
    for(int i = 1; i <= exfloat::precision64; i++)
	local_work.num[i] = x.num[i];
    return local_work;
}

exfloat fabs(const exfloat& x)
{
    exfloat local_work;
    local_work.num[0] = x.num[0] & EXFLIB_MASK_EXP;
    for(int i = 1; i <= exfloat::precision64; i++)
	local_work.num[i] = x.num[i];
    return local_work;
}

//----------------------------------------------------------------------
//  square root
//----------------------------------------------------------------------

exfloat sqrt(const exfloat& x)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_sqrt(local_work.num, x.num, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

//----------------------------------------------------------------------
//  trigonometric functions
//----------------------------------------------------------------------

exfloat sin(const exfloat &x)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_sin(local_work.num, x.num, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat cos(const exfloat& x)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_cos(local_work.num, x.num, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat tan(const exfloat& x)
{
    return sin(x)/cos(x);
}
    
//----------------------------------------------------------------------
//  exponential functions
//----------------------------------------------------------------------

exfloat exp(const exfloat& x)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_exp(local_work.num, x.num, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat pow(const exfloat& x, long y)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_pow_long(local_work.num, x.num, y, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat pow(const exfloat& x, int y)
{
    return pow(x, static_cast<long>(y));
}

exfloat pow(const exfloat& x, const exfloat& y)
{
    exfloat local_work;

    // comment out on 2017.09.23 (detecting in func_pow.c)
    // // inserted on 2016.09.29
    // if( (x.num[0] == 0) && ((y.num[0] & EXFLIB_MASK_EXP) != 0) ){ // x == +0 && y != 0
    //	local_work.num[0] = 0;
    //	return local_work;
    // }

#ifdef _OPENMP
    #pragma omp critical ( EXFLIB_INV_TABLE_LOG )
    {
	if(exflib::inv_table_log == 0){

            long size = exfloat::precision64 + 2;

	    exflib::inv_table_log
		= new exflib_uQword [size*EXFLIB_FUNCLOG_INVTBL_MAXN];

	    exflib_init_reciprocal_table(exflib::inv_table_log, size);
	}
    }
#else
    if(exflib::inv_table_log == 0){
	exflib::inv_table_log
	    = new exflib_uQword [(exfloat::precision64 + 2)*EXFLIB_FUNCLOG_INVTBL_MAXN];
	exflib::inv_table_log[0] = 0;
    }
#endif

    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_pow(local_work.num, x.num, y.num, exfloat::precision64, workmem, exflib::inv_table_log);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}
    
exfloat pow(long x, const exfloat& y)
{
    exfloat local_work;
    local_work = x;
    return pow(local_work, y);
}
    
exfloat pow(int x, const exfloat &y)
{
    exfloat local_work;
    local_work = static_cast<long>(x);
    return pow(local_work, y);
}

//----------------------------------------------------------------------
//  logarithmic functions
//----------------------------------------------------------------------

exfloat log(const exfloat& x)
{
#ifdef _OPENMP
    #pragma omp critical ( EXFLIB_INV_TABLE_LOG )
    {
	if(exflib::inv_table_log == 0){

            long size = exfloat::precision64 + 2;

	    exflib::inv_table_log
		= new exflib_uQword [size*EXFLIB_FUNCLOG_INVTBL_MAXN];

	    exflib_init_reciprocal_table(exflib::inv_table_log, size);
	}
    }
#else
    if(exflib::inv_table_log == 0){
	exflib::inv_table_log
	    = new exflib_uQword [(exfloat::precision64 + 2)*EXFLIB_FUNCLOG_INVTBL_MAXN];
	exflib::inv_table_log[0] = 0;
    }
#endif

    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_log(local_work.num, x.num, exfloat::precision64, workmem, exflib::inv_table_log);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat log10(const exfloat& x)
{
    exfloat local_work, work2;
    exflib_float_ctoef("ln10", work2.num, exfloat::precision64);
    local_work = log(x);
    local_work /= work2;
    return local_work;
}

//----------------------------------------------------------------------
// random number
//----------------------------------------------------------------------
//exfloat exflib_rand(){
//    exfloat local_work;
//    exflib_float_rand(local_work.num, exfloat::precision64);
//    return local_work;
//}

//----------------------------------------------------------------------
//  inverse trigonometric functions
//----------------------------------------------------------------------

exfloat asin(const exfloat& x)
{
    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_asin(local_work.num, x.num, exfloat::precision64, workmem);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat acos(const exfloat& x)
{
    exfloat local_work;
    exflib_uQword ret[exfloat::size];
    exflib_float_ctoef("pi", ret, exfloat::precision64);
    ret[0] -= 1;
    local_work = asin(x);
    exflib_float_sub(local_work.num, ret, local_work.num, exfloat::precision64);
    return local_work;
}

exfloat atan(const exfloat& x)
{
#ifdef _OPENMP
    #pragma omp critical ( EXFLIB_INV_TABLE_ATAN )
    {
	if(exflib::inv_table_atan == 0){

            long size = exfloat::precision64 + 2;

	    exflib::inv_table_atan
		= new exflib_uQword [size*EXFLIB_FUNCLOG_INVTBL_MAXN];

	    exflib_init_reciprocal_table(exflib::inv_table_atan, size);
	}
    }
#else
    if(exflib::inv_table_atan == 0){
	exflib::inv_table_atan
	    = new exflib_uQword [(exfloat::precision64+2)*EXFLIB_FUNCLOG_INVTBL_MAXN];
	exflib::inv_table_atan[0] = 0;
    }
#endif

    exfloat local_work;
    exflib_ulong hint;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    hint = exflib_float_atan(local_work.num, x.num, exfloat::precision64, workmem, exflib::inv_table_atan);
    exfloat::round(local_work.num, hint, exfloat::precision64);
    return local_work;
}

exfloat atan2(const exfloat &y, const exfloat &x)
{
    exfloat PI;
    exflib_float_ctoef("pi", PI.num, exfloat::precision64);

    if(x > 0)
	return atan(y/x);

    if( (y >= 0) && (x < 0) )
	return PI + atan(y/x);

    if( (y < 0) && (x < 0) )
	return -PI + atan(y/x);

    if( (y > 0) && (x == 0) )
	return PI / 2;

    if( (y < 0) && (x == 0) )
	return -PI / 2;

    // y == 0 && x == 0
    return static_cast<exfloat>(0);
}

//----------------------------------------------------------------------
//  hyperbolic trigonometric functions
//----------------------------------------------------------------------

exfloat sinh(const exfloat& x)
{
    exfloat local_work;

    if(x.iszero()){
	local_work.num[0] = 0; // 2016.09.30 debug
	return local_work;
    }
    exfloat e1 = exp(x);
    exfloat e2 = 1/e1;
    exflib_float_sub(local_work.num, e1.num, e2.num, exfloat::precision64);

    // 2016/08/08, debug
    if( (local_work.num[0] & EXFLIB_MASK_EXP) == 0 ){
	return local_work;
    }

    local_work.num[0] -= 1;	// (e1 - e2)/2;
    return local_work;
}
    
exfloat cosh(const exfloat& x)
{
    exfloat local_work;

    exfloat e1 = exp(x);
    exfloat e2 = 1/e1;
    exflib_float_add(local_work.num, e1.num, e2.num, exfloat::precision64);

    // 2016/08/08, debug
    if( (local_work.num[0] & EXFLIB_MASK_EXP) == 0 ){
	return local_work;
    }

    local_work.num[0] -= 1;	// (e1 + e2)/2;
    return local_work;
}
    
exfloat tanh(const exfloat& x)
{
    exfloat e = exp(x);
    exfloat e2 = e*e;
    return (e2 - 1)/(e2 + 1);
}

//----------------------------------------------------------------------
// separate integer part and fraction part
//----------------------------------------------------------------------

exfloat modf(const exfloat &x, exfloat *ip)
{
    exfloat local_work;
    exflib_float_modf(x.num, ip->num, local_work.num, exfloat::precision64);
    return local_work;
}

exfloat fmod(const exfloat &x, const exfloat &y)
{
    exfloat local_work;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    exflib_float_fmod(local_work.num, x.num, y.num, exfloat::precision64, workmem);
    return local_work;
}

exfloat ceil(const exfloat &x)
{
    exfloat local_work;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    exflib_float_ceil(local_work.num, x.num, exfloat::precision64, workmem);
    return local_work;
}

exfloat floor(const exfloat& x)
{
    exfloat local_work;
    exflib_uQword workmem[(exfloat::size + 1)*5];
    exflib_float_floor(local_work.num, x.num, exfloat::precision64, workmem);
    return local_work;
}

#endif	// _EXFLIB_COMPILE
#endif	// EXFLOAT_H

// End of exfloat.h
