// -*- C++ -*-
// $Id: matrix.C,v 1.4 2002/04/16 01:28:42 fujiwara Exp $

//----------------------------------------------------------------------
//  Copyright (C) 2002, FUJIWARA Hiroshi
//        <fujiwara@acs.i.kyoto-u.ac.jp>
//----------------------------------------------------------------------

#include <iostream>
#include <iomanip>
using namespace std;

#define PRECISION 100
#include "exfloat.h"
#include "exfutil.h"

template <class Tp>
void mulVector(const Tp *const *const a, long n, const Tp *const v, Tp* c)
{
    for(int i = 0; i < n; i++){
	c[i] = 0;
	for(int k = 0; k < n; k++){
	    c[i] += a[i][k] * v[k];
	}
    }
    return;
}

template <class Tp>
void printMatrix(const Tp *const *const a, long row, long col)
{
    for(int i = 0; i < row; i++){
	for(int j = 0; j < col; j++)
	    cout<< a[i][j] << ' ';
	cout << '\n';
    }
    
    return;
}

template <class Tp>
void printVector(const Tp *const a, long n)
{
    for(int i = 0; i < n; i++)
	cout<< a[i] << ' ';
    cout << '\n';
    
    return;
}

int main()
{
    const long N = 4;
    exfloat **a;

    // memory allocation function
    exflib::newMatrix(a, N);

    exfloat aa[N][N] = {
	{4,3,2,1}, {3,3,2,1}, {2,2,2,1}, {1,1,1,1}
    };

    // set elements
    for(int i = 0; i < N; i++)
	for(int j = 0; j < N; j++)
	    a[i][j] = aa[i][j];

    // another memory allocation
    exfloat **b, **c;
    b = new exfloat * [N];
    for(int i = 0; i < N; i++)
	b[i] = new exfloat [N];

    // another memory allocation function
    exflib::newMatrix(c, N, N);

    //------------------------------
    // QR decomposition : a = bc
    //------------------------------
    cout << "QR decomposition------------------------------\n";

    qr(a, N, N, b, c);

    // output
    cout << setiosflags(ios::fixed);

    cout << "A = \n";    printMatrix(a, N, N);    cout << '\n';
    cout << "Q = \n";    printMatrix(b, N, N);    cout << '\n';
    cout << "R = \n";    printMatrix(c, N, N);	  cout << '\n';

    // check
    exflib::mulMatrix(b, c, N, a);
    cout << "Q*R = \n";    printMatrix(a, N, N);    cout << '\n';

    cout << '\n';

    //------------------------------
    // LU decomposition
    //------------------------------
    cout << "LU Decomposition (Solve Ax = b) ------------------------------\n";

    exfloat v[N] = { 22, 1, 5, -1};
    exfloat x[N];

    cout << "A = \n";   printMatrix(a, N, N);   cout << '\n';
    cout << "b = \n";	printVector(v, N);	cout << '\n';

    lu(a, N, v, x);

    cout << "x = \n";    printVector(x, N);	cout << '\n';

    // a is destroyed by LU decomposition lu()
    // reset elements
    for(int i = 0; i < N; i++)
	for(int j = 0; j < N; j++)
	    a[i][j] = aa[i][j];

    // check
    cout << "A*x = \n";
    mulVector(a, N, x, v);
    printVector(v, N);

    return 0;
}

//----------------------------------------------------------------------
// Local Variables:
// compile-command: "g++ -Wall -O2"
// End:
//----------------------------------------------------------------------
