!----------------------------------------------------------------------
! Gauss Elimination for Hilbert Matrix
!   Time-stamp: "2006-03-05 21:34:18 fujiwara"
!----------------------------------------------------------------------

  ! main routine ************************************************************

  PROGRAM gauss
    
    USE exflib

    IMPLICIT NONE

    !---------------------------
    ! dynamic memory allocation
    !---------------------------

!    INTEGER*8 :: N
!    TYPE(exfloat), DIMENSION(:,:), ALLOCATABLE :: a
!    INTEGER*8 :: i, j
!!    WRITE(*,*) "N = "
!!    READ(*,*) N
!    N = 100
!    ALLOCATE(a(N, N+1))      ! add DEALLOCATE(A) before CONTAINS

    ! Static allocation
    INTEGER*8, PARAMETER :: N = 100
    TYPE(exfloat) :: a(N, N+1)
    INTEGER*8 :: i, j

    ! print information
    WRITE(*,*) 'Required precision = ', exflib_get_req_precision()
    WRITE(*,*) 'Current precision = ', exflib_get_precision()
    WRITE(*,*) 'Current libexfloat.a is ', TRIM(exflib_get_version())
    WRITE(*,*) 'An exfloat is', exflib_get_exfloat_byte(), 'bytes.'

    ! set a coefficient matrix (Hilbert Matrix)
    CALL set_hilbert_matrix()

    ! print the coefficient matrix
    ! CALL print_matrix()
    
    CALL forward_elimination()

    CALL backward_substitution()

    ! print the solution
    DO i = 1, N
       WRITE(*,*) i, ' : ', TRIM(exflib_format('F', a(i,N+1)))
    ENDDO

    ! You need the next in dynamic allocation
!    DEALLOCATE(a)

    ! internal procedure **************************************************

    CONTAINS

      ! ************************************************************
      SUBROUTINE pivoting(k)
        INTEGER*8, INTENT(IN) :: k
        TYPE(exfloat) :: tmp
        INTEGER*8 :: r, row, j

        row = k
	WRITE(*,*) k

        DO r = k+1, N
           IF ( abs(a(row,k)) < abs(a(r,k))) THEN
              row = r
           ENDIF
        ENDDO

        IF ( a(row,k) == 0 ) THEN
           WRITE(*,*) "The coefficient matrix is singular:", k
           STOP
        ENDIF

        IF ( row /= k ) THEN
           DO j = 1, N+1
              tmp = a(k, j)
              a(k, j) = a(row, j)
              a(row, j) = tmp
           ENDDO
        ENDIF

      END SUBROUTINE pivoting
 
      ! ************************************************************
      SUBROUTINE forward_elimination()

        INTEGER*8 :: i, j, k
        TYPE(exfloat) :: p

        DO k = 1, N
           CALL pivoting(k)

           DO i = k+1, N

              p = a(i,k)/a(k,k)
              
              DO j = k+1, N+1
                 a(i,j) = a(i,j) - a(k,j)*p
              ENDDO

              a(i,k) = 0
           ENDDO
        ENDDO

      END SUBROUTINE forward_elimination

      ! ************************************************************
      SUBROUTINE backward_substitution()

        INTEGER*8 :: k, i

        DO k = N, 1, -1

           a(k, N+1) = a(k, N+1) / a(k, k)

           DO i = 1, k-1
              a(i, N+1) = a(i, N+1) - a(i, k) * a(k, N+1)
           ENDDO

        ENDDO

      END SUBROUTINE backward_substitution

      ! **************************************************
      SUBROUTINE set_hilbert_matrix()

        INTEGER*8 :: i, j

        DO i = 1, N
           DO j = 1, N
              a(i, j) = 1
              a(i, j) = a(i,j) / (i+j-1)
           END DO
           
           ! RHS
           a(i, N+1) = 1
           
        END DO

      END SUBROUTINE set_hilbert_matrix
      
      ! **************************************************
      SUBROUTINE print_matrix()

        INTEGER*8 :: i

        DO i = 1, N
           DO j = 1, N+1
              write(*,*) i,':',j,':', exflib_double(a(i,j))
           ENDDO
        ENDDO

      END SUBROUTINE print_matrix

    END PROGRAM gauss

!----------------------------------------------------------------------
! End of File
!----------------------------------------------------------------------
