! -*- f90 -*-
!----------------------------------------------------------------------
! xlamch.f90
! Time-stamp: "2009-02-04 11:56:53 fujiwara"
!----------------------------------------------------------------------

  ! main ************************************************************
  PROGRAM main

    USE exflib

    WRITE(*,*) 'current digits = ', exflib_exfloat_precision
    WRITE(*,*) 'The size of exfloat is = ', exflib_exfloat_byte, 'bytes.'

    WRITE(*,*)

    WRITE(*,*) 'eps =      ', TRIM(exflib_format('E', xlamch('e')))
    WRITE(*,*) 'sfmin =    ', TRIM(exflib_format('E', xlamch('s')))
    WRITE(*,*) 'base =     ', TRIM(exflib_format('E', xlamch('b')))
    WRITE(*,*) 'eps*base = ', TRIM(exflib_format('E', xlamch('p')))
    WRITE(*,*) 't =        ', TRIM(exflib_format('E', xlamch('n')))
    WRITE(*,*) 'rnd =      ', TRIM(exflib_format('E', xlamch('r')))
    WRITE(*,*) 'emin =     ', TRIM(exflib_format('E', xlamch('m')))
    WRITE(*,*) 'rmin =     ', TRIM(exflib_format('E', xlamch('u')))
    WRITE(*,*) 'emax =     ', TRIM(exflib_format('E', xlamch('l')))
    WRITE(*,*) 'rmax =     ', TRIM(exflib_format('E', xlamch('o')))

  CONTAINS

!**********************************************************************
! exflib version of 's(d)lamch'
!  Copyright(C) 2009, FUJIWARA Hiroshi
!**********************************************************************
    FUNCTION xlamch(CMACH) RESULT(RET_VAL)

      USE exflib
      CHARACTER, INTENT(IN) :: CMACH
      TYPE(exfloat) :: RET_VAL
      TYPE(exfloat) :: x

      IF ((CMACH.EQ.'E').OR.(CMACH.EQ.'e')) THEN

!        eps   = relative machine precision
!        DLAMCH('e') = 1.11022302462515654042E-16 = 2.0D0**(-53.0D0)

!        The above eps is 'the half of the gap between the nearest
!        floating-point numbers', in other words,
!        the largest number that satisfies round(1.0+eps) = 1.0D0.
         RET_VAL = 1
         RET_VAL%num(0) = RET_VAL%num(0) - (1 + 64*(exflib_exfloat_size-1))
         
      ELSE IF ((CMACH.EQ.'S').OR.(CMACH.EQ.'s')) THEN

!        sfmin = safe minimum, such that 1/sfmin does not overflow
!        DLAMCH('s') = 0.00000000000000000000E+00
         RET_VAL = 0
         RET_VAL%num(0) = 1  ! ???

      ELSE IF ((CMACH.EQ.'B').OR.(CMACH.EQ.'b')) THEN

!        base  = base of the machine
!        DLAMCH('b') = 2.00000000000000000000E+00
         RET_VAL = 2

      ELSE IF ((CMACH.EQ.'P').OR.(CMACH.EQ.'p')) THEN

!        prec  = eps*base
!        DLAMCH('p') = 2.22044604925031308085E-16
         RET_VAL = 1
         RET_VAL%num(0) = RET_VAL%num(0) - 64*(exflib_exfloat_size-1)

      ELSE IF ((CMACH.EQ.'N').OR.(CMACH.EQ.'n')) THEN

!        t     = number of (base) digits in the mantissa
!        DLAMCH('t') = 5.30000000000000000000E+01
         RET_VAL = 1 + 64*(exflib_exfloat_size-1)

      ELSE IF ((CMACH.EQ.'R').OR.(CMACH.EQ.'r')) THEN

!        rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
!        DLAMCH('r') = 1.00000000000000000000E+00
         RET_VAL = 1

      ELSE IF ((CMACH.EQ.'M').OR.(CMACH.EQ.'m')) THEN

!        emin  = minimum exponent before (gradual) underflow
!        DLAMCH('m') = -1.02100000000000000000E+03

!        The exponent of the IEEE754 normalized double is in
!        (1 <= e+BIAS <= 0x7fe = 2046), where BIAS = 0x3ff = 1023.
!        Thus the normalized e is in (-1022 <= e <= 1023).

!        For exflib exfloat number, the exponent is in
!        (1 <= e+BIAS <= 0x7ffffffffffffffe), where BIAS = 0x3fffffffffffffff.
!        Thus -0x3ffffffffffffffe <= e <= 0x3fffffffffffffff.
         RET_VAL = -1
         RET_VAL%num(0) = RET_VAL%num(0) + 63
         ! gfortran does not support unsigned integer in hexadecimal.
!         RET_VAL%num(1) = Z'ffffffffffffffe8'  ! may be Z'fffffffffffffff0'
         RET_VAL%num(1) = Z'7ffffffffffffff4'
         RET_VAL%num(1) = RET_VAL%num(1) * 2

      ELSE IF ((CMACH.EQ.'U').OR.(CMACH.EQ.'u')) THEN

!        rmin  = underflow threshold - base**(emin-1)
!        DLAMCH('u') = 0.00000000000000000000E+00
         RET_VAL = 1
         RET_VAL%num(0) = 1  ! may be 0

      ELSE IF ((CMACH.EQ.'L').OR.(CMACH.EQ.'l')) THEN

!        emax  = largest exponent before overflow
!        DLAMCH('l') = 1.02400000000000000000E+03
         RET_VAL = 1
         RET_VAL%num(0) = RET_VAL%num(0) + 63
         ! gfortran does not support unsigned integer in hexadecimal.
!         RET_VAL%num(1) = Z'fffffffffffffff0'  ! may be Z'fffffffffffffff8'
         RET_VAL%num(1) = Z'7ffffffffffffff8'
         RET_VAL%num(1) = RET_VAL%num(1) * 2

      ELSE IF ((CMACH.EQ.'O').OR.(CMACH.EQ.'o')) THEN

!        rmax  = overflow threshold  - (base**emax)*(1-eps)
!        DLAMCH('o') = Infinity

         ! base**emax
         RET_VAL = 1
         RET_VAL%num(0) = Z'7ffffffffffffffd'  ! may be Z'7ffffffffffffffe'

         ! eps
         x = 1
         x%num(0) = x%num(0) - (1 + 64*(exflib_exfloat_size-1))

         RET_VAL = RET_VAL * (1 - x)

      ELSE

         RET_VAL = 0
         STOP 'xlamch() : UNKNOWN ARGUMENTS'

      END IF

    END FUNCTION xlamch
!**********************************************************************

  END PROGRAM main

! End of file
