// -*- C++ -*-
// Time-stamp: "2009-09-24 09:40:14 fujiwara"
// $Id$

//----------------------------------------------------------------------
//  sample of MPI (one-to-one communication) program with exfloat
//
//  Copyright (C) 2006, FUJIWARA Hiroshi
//        <fujiwara@acs.i.kyoto-u.ac.jp>
//----------------------------------------------------------------------

#define PRECISION 100
#include "exfloat.h"

#include <iostream>
#include "mpi.h"
using namespace std;

//----------------------------------------------------------------------
// send exfloat array 'v' (# of elements is 'length'), to process 'dst'
//----------------------------------------------------------------------
void send(const exfloat *const v, const size_t length, const int dst,
	  const int tag = 1)
{
    // buffer
    const size_t bufsize = exfloat::size * length;
    exflib_uQword *buf;
    buf = new exflib_uQword [bufsize];

    // packing
    long size = exfloat::size * length;
    for(unsigned int j = 0; j < length; j++)
	for(int i = 0; i < exfloat::size; i++)
	    buf[exfloat::size * j + i] = (v[j]).num[i];
	
    // send
    MPI::COMM_WORLD.Send(buf, size, MPI::DOUBLE, dst, tag);

    delete [] buf;	buf = NULL;
    return;
}
    
//----------------------------------------------------------------------
// recv exfloat array 'v' (# of elements is 'length'), from process 'src'
//----------------------------------------------------------------------
void recv(exfloat *const v, const size_t length, const int src,
	  int tag = 1)
{
    // buffer
    const size_t bufsize = exfloat::size * length;
    exflib_uQword *buf;
    buf = new exflib_uQword [bufsize];

    long size = exfloat::size * length;

    // recieve
    MPI::COMM_WORLD.Recv(buf, size, MPI::DOUBLE, src, tag);

    // unpack
    for(unsigned int j = 0; j < length; j++)
	for(int i = 0; i < exfloat::size; i++)
	    (v[j]).num[i] = buf[exfloat::size * j + i];

    delete [] buf;	buf = NULL;
    return;
}

//----------------------------------------------------------------------
// main routine
//----------------------------------------------------------------------

int main(int argc, char *argv[])
{
    //------------------------------------------------------------
    // initialization
    //------------------------------------------------------------
    MPI::Init(argc, argv);
    const long mynode = MPI::COMM_WORLD.Get_rank();
    const long numprocs = MPI::COMM_WORLD.Get_size();

    // from process '1' to process '0'
    const long dst = 0;
    const long src = 1;

    //------------------------------------------------------------
    // exfloat 1 Ĥ
    //
    // (1) 饹 exfloat  num ѿĤ (x.num, y.num)
    // (2)  exfloat::size 
    // (3)  MPI::DOUBLE ꤹ
    //------------------------------------------------------------
    // ͤν
    exfloat x = 0;
    if(mynode == src)
	x = static_cast<exfloat>(1)/(mynode+10);

    // 
    if(mynode == src)
	MPI::COMM_WORLD.Send(x.num, exfloat::size, MPI::DOUBLE, dst, 1);

    // 
    if(mynode == dst)
	MPI::COMM_WORLD.Recv(x.num, exfloat::size, MPI::DOUBLE, src, 1);

    // ͤɽ
    if(mynode == dst)
	cout << x << endl;

    //------------------------------------------------------------
    // 1Ĥ exfloat եκǽؿȤ
    //------------------------------------------------------------
    if(mynode == dst) x = 0;

    if(mynode == src) send(&x, 1, dst);
    if(mynode == dst) recv(&x, 1, src);

    if(mynode == dst) cout << x << endl;

    //------------------------------------------------------------
    // exfloat եκǽؿȤ
    //------------------------------------------------------------
    // N Ǥ
    const long N = 10;
    exfloat a[N];

    // 
    if(mynode == dst)
	for(long i = 0; i < N; i++)
	    a[i] = 0;
    if(mynode == src)
	for(long i = 0; i < N; i++)
	    a[i] = i + mynode*100;

    // 
    if(mynode == src)
	send(a, N, dst);

    // 
    if(mynode == dst)
	recv(a, N, src);

    // ͤɽ
    if(mynode == dst)
	for(long i = 0; i < N; i++)
	    cout << i << ' ' << a[i] << endl;
    
    //------------------------------------------------------------
    // λ
    //------------------------------------------------------------
    MPI::Finalize();
    return 0;
}

// End of file
