// -*- C++ -*-
// Time-stamp: "2006-04-11 12:16:22 fujiwara"
// $Id$

//----------------------------------------------------------------------
//  sample of MPI (Gather) program with exfloat
//
//  Copyright (C) 2006, FUJIWARA Hiroshi
//        <fujiwara@acs.i.kyoto-u.ac.jp>
//----------------------------------------------------------------------
#define PRECISION 100
#include "exfloat.h"

#include <iostream>
#include "mpi.h"
using namespace std;

//----------------------------------------------------------------------
// exfloat μ
// src  length Ĥ򡢥Ρ root  dst ˽
// src  Ĺ length dst Ĺ length*(Ρɿ) 
//----------------------------------------------------------------------
void gather(const exfloat *const src, exfloat *const dst,
	    const long length, const long root)
{
    const long ranks = MPI::COMM_WORLD.Get_size();

    // ѤΥХåե
    exflib_uQword *srcbuf, *dstbuf;
    const long bufsize = length * exfloat::size; 
    srcbuf = new exflib_uQword [bufsize];
    dstbuf = new exflib_uQword [bufsize * ranks];
    
    // Packing
    for(long i = 0; i < length; i++)
	for(long s = 0; s < exfloat::size; s++)
	    srcbuf[i*exfloat::size + s] = (src[i]).num[s];

    // Communicating
    MPI::COMM_WORLD.Gather(srcbuf, bufsize, MPI::DOUBLE,
			   dstbuf, bufsize, MPI::DOUBLE, root);

    // Unpacking
    if(root == MPI::COMM_WORLD.Get_rank())
	for(long r = 0; r < ranks; r++)
	    for(long i = 0; i < length; i++)
		for(long s = 0; s < exfloat::size; s++)
		    (dst[r*length + i]).num[s]
			= dstbuf[(r*length + i)*exfloat::size + s];
    
    delete [] srcbuf;	srcbuf = NULL;
    delete [] dstbuf;	dstbuf = NULL;
    return;
}

//----------------------------------------------------------------------
// main routine
//----------------------------------------------------------------------

int main(int argc, char *argv[])
{
    MPI::Init(argc, argv);
    const long numprocs = MPI::COMM_WORLD.Get_size();
    const long myproc = MPI::COMM_WORLD.Get_rank();

    // ǤϡΡ 0 ˽
    const long root = 0;

    // ¿Ĺ
    const long N = 5;
    exfloat a[N];
    exfloat *buf = new exfloat [N*numprocs];

    for(long i = 0; i < N; i++)
	a[i] = myproc*100+i;

    // 
    gather(a, buf, N, root);

    // Ƥγǧ
    if(myproc == root)
	for(long i = 0; i < N*numprocs; i++)
	    cout << i << ' ' << buf[i] << endl;

    delete [] buf;	buf = NULL;
    MPI::Finalize();
    return 0;
}

// End of file
