// -*- C++ -*-
// Time-stamp: "2006-04-12 00:29:45 fujiwara"
// $Id$

//----------------------------------------------------------------------
//  sample of MPI (Reduce, finding MAXIMUM with index) program with exfloat
//
//  Copyright (C) 2006, FUJIWARA Hiroshi
//        <fujiwara@acs.i.kyoto-u.ac.jp>
//----------------------------------------------------------------------

#define PRECISION 100
#include "exfloat.h"

#include <iostream>
#include "mpi.h"
using namespace std;

//----------------------------------------------------------------------
// 黻ؿ : ͤθտ魯륤ǥå
//----------------------------------------------------------------------

void exflib_lt_loc(const void *inbuf, void *outbuf,
		   const int length, const MPI::Datatype& type)
{
    exflib_cuQword *in = reinterpret_cast<exflib_cuQword*>(inbuf);
    exflib_uQword *out = reinterpret_cast<exflib_uQword *>(outbuf);

    // out < in ʤ顢out  in ˽񤭴
#if 1

    // C 롼θƤӽФ
    // ݤפʤ顢ǽʬ®Ǥ⤢롣
    if(exflib_float_lt(&(out[1]), &(in[1]), exfloat::precision64))
	for(long i = 0; i <= exfloat::size; i++)
	    out[i] = in[i];

#else

    // ݤޤȤϡö¿Ĺˤɤۤ褤
    exfloat in_val, out_val;
    for(long i = 0; i < exfloat::size; i++){
	in_val.num[i] = in[i];
	out_val.num[i] = out[i];
    }

    // ᤵ줿¿ĹФƱ黻¹
    if(out_val < in_val)
	for(long i = 0; i <= exfloat::size; i++)
	    out[i] = in[i];
#endif    
    return;
}

//----------------------------------------------------------------------
//  Reduce ؿ
//  ¿Ĺ 1 Ĥȡտ魯륤ǥå黻оݤȤ
//  *src_elem, *src_index  &function ܤ̤root ˽󤹤
//  root  res, res_index ̤ˤʤ
//
//
//  src_elem  : 黻оݤ¿Ĺ
//  src_index : src_elem տ魯륤ǥå
//  res_elem  : 黻̤Ǽ¿Ĺ
//  res_index : 黻̤οտ魯륤ǥå
//  function  : 黻
//  root      : 黻 (res) ݻΡɤֹ
//----------------------------------------------------------------------

template <class _index_type>
void reduce(const exfloat *src_elem, const _index_type *src_index,
	    exfloat *res_elem, _index_type *res_index,
	    MPI::Op &function, const long root)
{
    exflib_uQword *sbuf, *rbuf;
    sbuf = new exflib_uQword [exfloat::size+1];
    rbuf = new exflib_uQword [exfloat::size+1];

    // packing
    // buffer  0 Ǥǥå1 ְʹߤ¿Ĺ
    sbuf[0] = *src_index;
    for(long i = 1; i <= exfloat::size; i++)
	sbuf[i] = (*src_elem).num[i-1];

    // communication & operation
    MPI::COMM_WORLD.Reduce(sbuf, rbuf, exfloat::size, MPI::DOUBLE,
			   function, root);

    // unpacking
    if(MPI::COMM_WORLD.Get_rank() == root){
	*res_index = static_cast<_index_type>(rbuf[0]);
	for(long i = 1; i <= exfloat::size; i++)
	    (*res_elem).num[i-1] = rbuf[i];
    }

    return;
}

//----------------------------------------------------------------------
// main routine
//----------------------------------------------------------------------
int main(int argc, char *argv[])
{
    MPI::Init(argc, argv);
    const long numprocs = MPI::COMM_WORLD.Get_size();
    const long myproc   = MPI::COMM_WORLD.Get_rank();

    // ΥץǤϡƥΡɤ¿Ĺ x κͤȡ
    // տ魯 index 
    // Ρ root (= 0)  maximum  index ˵
    exfloat x = static_cast<exfloat>(1)/(myproc+1);
    exfloat maximum;
    long index;

    const long root = 0;

    // 黻Ͽ
    MPI::Op exflib_maxloc;
    exflib_maxloc.Init(exflib_lt_loc, true);

    // 黻
    reduce(&x, &myproc,
	   &maximum, &index,
	   exflib_maxloc, root);

    // ̤ɽ
    if(myproc == root)
	cout << index << ' ' << maximum << endl;

    MPI::Finalize();
    return 0;
}

// End of file
