! -*- f90 -*-
!**********************************************************************
! Time-stamp: "2009-10-28 12:19:03 fujiwara"
!
!  EXFLIB array operations
!
!  Copyright (C) 2009-, FUJIWARA Hiroshi <fujiwara@acs.i.kyoto-u.ac.jp>
!
!----------------------------------------------------------------------
! TODO : three or higher dimensional array
!      : array size check
!      : MERGE, PACK, SUPREAD, UNPACK, RESHAPE, CSHIFT, EOSHIFT
!      : are not implemented.
!**********************************************************************

    MODULE exflib_array

      USE exflib
      IMPLICIT NONE

      INTEGER, PARAMETER, PRIVATE :: EXFLIB_ARRAY_BEGIN = 0

      !--------------------------------------------------
      ! type name abbreviation
      !--------------------------------------------------
      ! i : INTEGER*4, integer
      ! q : INTEGER*8, quadword
      ! s : REAL*4, float
      ! d : REAL*8, double
      ! x : TYPE(exfloat)
      ! c : COMPLEX*8
      ! w : COMPLEX*16
      ! z : TYPE(ezfloat)
      ! l : LOGICAL
      !--------------------------------------------------
      ! one dimensional array functions
      !--------------------------------------------------
      INTERFACE ASSIGNMENT ( = )
         MODULE PROCEDURE exflib_float_assign_x1i0
         MODULE PROCEDURE exflib_float_assign_x1q0
         MODULE PROCEDURE exflib_float_assign_x1x0

         MODULE PROCEDURE exflib_float_assign_x1i1
         MODULE PROCEDURE exflib_float_assign_x1q1

         MODULE PROCEDURE exflib_float_assign_z1i0
         MODULE PROCEDURE exflib_float_assign_z1q0
         MODULE PROCEDURE exflib_float_assign_z1x0
         MODULE PROCEDURE exflib_float_assign_z1z0

         MODULE PROCEDURE exflib_float_assign_z1i1
         MODULE PROCEDURE exflib_float_assign_z1q1
         MODULE PROCEDURE exflib_float_assign_z1x1
      END INTERFACE

      INTERFACE exflib_double
         MODULE PROCEDURE exflib_float_cast_x1d1
      END INTERFACE
      INTERFACE DBLE
         MODULE PROCEDURE exflib_float_cast_x1d1
      END INTERFACE
!      INTERFACE REAL
!      END INTERFACE
!      INTERFACE QREAL
!      END INTERFACE
      INTERFACE exflib_complex
         MODULE PROCEDURE exflib_float_cast_z1w1
      END INTERFACE
      INTERFACE CMPLX
         MODULE PROCEDURE exflib_float_cast_z1w1
      END INTERFACE
      INTERFACE DCMPLX
         MODULE PROCEDURE exflib_float_cast_z1w1
      END INTERFACE
!      INTERFACE QCMPLX
!      END INTERFACE

      INTERFACE OPERATOR ( == )
         ! exfloat
         MODULE PROCEDURE exflib_float_eq_i0x1
         MODULE PROCEDURE exflib_float_eq_q0x1
         MODULE PROCEDURE exflib_float_eq_s0x1
         MODULE PROCEDURE exflib_float_eq_d0x1
         MODULE PROCEDURE exflib_float_eq_x0x1
         MODULE PROCEDURE exflib_float_eq_c0x1
         MODULE PROCEDURE exflib_float_eq_w0x1
         MODULE PROCEDURE exflib_float_eq_z0x1

         MODULE PROCEDURE exflib_float_eq_i1x0
         MODULE PROCEDURE exflib_float_eq_q1x0
         MODULE PROCEDURE exflib_float_eq_s1x0
         MODULE PROCEDURE exflib_float_eq_d1x0
         MODULE PROCEDURE exflib_float_eq_x1x0
         MODULE PROCEDURE exflib_float_eq_c1x0
         MODULE PROCEDURE exflib_float_eq_w1x0
         MODULE PROCEDURE exflib_float_eq_z1x0

         MODULE PROCEDURE exflib_float_eq_i1x1
         MODULE PROCEDURE exflib_float_eq_q1x1
         MODULE PROCEDURE exflib_float_eq_s1x1
         MODULE PROCEDURE exflib_float_eq_d1x1
         MODULE PROCEDURE exflib_float_eq_x1x1
         MODULE PROCEDURE exflib_float_eq_c1x1
         MODULE PROCEDURE exflib_float_eq_w1x1
         MODULE PROCEDURE exflib_float_eq_z1x1

         MODULE PROCEDURE exflib_float_eq_x1i0
         MODULE PROCEDURE exflib_float_eq_x1q0
         MODULE PROCEDURE exflib_float_eq_x1s0
         MODULE PROCEDURE exflib_float_eq_x1d0
         !MODULE PROCEDURE exflib_float_eq_x1x0
         MODULE PROCEDURE exflib_float_eq_x1c0
         MODULE PROCEDURE exflib_float_eq_x1w0
         MODULE PROCEDURE exflib_float_eq_x1z0

         MODULE PROCEDURE exflib_float_eq_x0i1
         MODULE PROCEDURE exflib_float_eq_x0q1
         MODULE PROCEDURE exflib_float_eq_x0s1
         MODULE PROCEDURE exflib_float_eq_x0d1
         !MODULE PROCEDURE exflib_float_eq_x0x1
         MODULE PROCEDURE exflib_float_eq_x0c1
         MODULE PROCEDURE exflib_float_eq_x0w1
         MODULE PROCEDURE exflib_float_eq_x0z1

         MODULE PROCEDURE exflib_float_eq_x1i1
         MODULE PROCEDURE exflib_float_eq_x1q1
         MODULE PROCEDURE exflib_float_eq_x1s1
         MODULE PROCEDURE exflib_float_eq_x1d1
         !MODULE PROCEDURE exflib_float_eq_x1x1
         MODULE PROCEDURE exflib_float_eq_x1c1
         MODULE PROCEDURE exflib_float_eq_x1w1
         MODULE PROCEDURE exflib_float_eq_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_eq_i0z1
         MODULE PROCEDURE exflib_float_eq_q0z1
         MODULE PROCEDURE exflib_float_eq_s0z1
         MODULE PROCEDURE exflib_float_eq_d0z1
         !MODULE PROCEDURE exflib_float_eq_x0z1
         MODULE PROCEDURE exflib_float_eq_c0z1
         MODULE PROCEDURE exflib_float_eq_w0z1
         MODULE PROCEDURE exflib_float_eq_z0z1

         MODULE PROCEDURE exflib_float_eq_i1z0
         MODULE PROCEDURE exflib_float_eq_q1z0
         MODULE PROCEDURE exflib_float_eq_s1z0
         MODULE PROCEDURE exflib_float_eq_d1z0
         !MODULE PROCEDURE exflib_float_eq_x1z0
         MODULE PROCEDURE exflib_float_eq_c1z0
         MODULE PROCEDURE exflib_float_eq_w1z0
         MODULE PROCEDURE exflib_float_eq_z1z0

         MODULE PROCEDURE exflib_float_eq_i1z1
         MODULE PROCEDURE exflib_float_eq_q1z1
         MODULE PROCEDURE exflib_float_eq_s1z1
         MODULE PROCEDURE exflib_float_eq_d1z1
         !MODULE PROCEDURE exflib_float_eq_x1z1
         MODULE PROCEDURE exflib_float_eq_c1z1
         MODULE PROCEDURE exflib_float_eq_w1z1
         MODULE PROCEDURE exflib_float_eq_z1z1

         MODULE PROCEDURE exflib_float_eq_z1i0
         MODULE PROCEDURE exflib_float_eq_z1q0
         MODULE PROCEDURE exflib_float_eq_z1s0
         MODULE PROCEDURE exflib_float_eq_z1d0
         !MODULE PROCEDURE exflib_float_eq_z1x0
         MODULE PROCEDURE exflib_float_eq_z1c0
         MODULE PROCEDURE exflib_float_eq_z1w0
         !MODULE PROCEDURE exflib_float_eq_z1z0

         MODULE PROCEDURE exflib_float_eq_z0i1
         MODULE PROCEDURE exflib_float_eq_z0q1
         MODULE PROCEDURE exflib_float_eq_z0s1
         MODULE PROCEDURE exflib_float_eq_z0d1
         !MODULE PROCEDURE exflib_float_eq_z0x1
         MODULE PROCEDURE exflib_float_eq_z0c1
         MODULE PROCEDURE exflib_float_eq_z0w1
         !MODULE PROCEDURE exflib_float_eq_z0z1

         MODULE PROCEDURE exflib_float_eq_z1i1
         MODULE PROCEDURE exflib_float_eq_z1q1
         MODULE PROCEDURE exflib_float_eq_z1s1
         MODULE PROCEDURE exflib_float_eq_z1d1
         !MODULE PROCEDURE exflib_float_eq_z1x1
         MODULE PROCEDURE exflib_float_eq_z1c1
         MODULE PROCEDURE exflib_float_eq_z1w1
         !MODULE PROCEDURE exflib_float_eq_z1z1
      END INTERFACE

      INTERFACE OPERATOR ( /= )
         ! exfloat
         MODULE PROCEDURE exflib_float_ne_i0x1
         MODULE PROCEDURE exflib_float_ne_q0x1
         MODULE PROCEDURE exflib_float_ne_s0x1
         MODULE PROCEDURE exflib_float_ne_d0x1
         MODULE PROCEDURE exflib_float_ne_x0x1
         MODULE PROCEDURE exflib_float_ne_c0x1
         MODULE PROCEDURE exflib_float_ne_w0x1
         MODULE PROCEDURE exflib_float_ne_z0x1

         MODULE PROCEDURE exflib_float_ne_i1x0
         MODULE PROCEDURE exflib_float_ne_q1x0
         MODULE PROCEDURE exflib_float_ne_s1x0
         MODULE PROCEDURE exflib_float_ne_d1x0
         MODULE PROCEDURE exflib_float_ne_x1x0
         MODULE PROCEDURE exflib_float_ne_c1x0
         MODULE PROCEDURE exflib_float_ne_w1x0
         MODULE PROCEDURE exflib_float_ne_z1x0

         MODULE PROCEDURE exflib_float_ne_i1x1
         MODULE PROCEDURE exflib_float_ne_q1x1
         MODULE PROCEDURE exflib_float_ne_s1x1
         MODULE PROCEDURE exflib_float_ne_d1x1
         MODULE PROCEDURE exflib_float_ne_x1x1
         MODULE PROCEDURE exflib_float_ne_c1x1
         MODULE PROCEDURE exflib_float_ne_w1x1
         MODULE PROCEDURE exflib_float_ne_z1x1

         MODULE PROCEDURE exflib_float_ne_x1i0
         MODULE PROCEDURE exflib_float_ne_x1q0
         MODULE PROCEDURE exflib_float_ne_x1s0
         MODULE PROCEDURE exflib_float_ne_x1d0
         !MODULE PROCEDURE exflib_float_ne_x1x0
         MODULE PROCEDURE exflib_float_ne_x1c0
         MODULE PROCEDURE exflib_float_ne_x1w0
         MODULE PROCEDURE exflib_float_ne_x1z0

         MODULE PROCEDURE exflib_float_ne_x0i1
         MODULE PROCEDURE exflib_float_ne_x0q1
         MODULE PROCEDURE exflib_float_ne_x0s1
         MODULE PROCEDURE exflib_float_ne_x0d1
         !MODULE PROCEDURE exflib_float_ne_x0x1
         MODULE PROCEDURE exflib_float_ne_x0c1
         MODULE PROCEDURE exflib_float_ne_x0w1
         MODULE PROCEDURE exflib_float_ne_x0z1

         MODULE PROCEDURE exflib_float_ne_x1i1
         MODULE PROCEDURE exflib_float_ne_x1q1
         MODULE PROCEDURE exflib_float_ne_x1s1
         MODULE PROCEDURE exflib_float_ne_x1d1
         !MODULE PROCEDURE exflib_float_ne_x1x1
         MODULE PROCEDURE exflib_float_ne_x1c1
         MODULE PROCEDURE exflib_float_ne_x1w1
         MODULE PROCEDURE exflib_float_ne_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_ne_i0z1
         MODULE PROCEDURE exflib_float_ne_q0z1
         MODULE PROCEDURE exflib_float_ne_s0z1
         MODULE PROCEDURE exflib_float_ne_d0z1
         !MODULE PROCEDURE exflib_float_ne_x0z1
         MODULE PROCEDURE exflib_float_ne_c0z1
         MODULE PROCEDURE exflib_float_ne_w0z1
         MODULE PROCEDURE exflib_float_ne_z0z1

         MODULE PROCEDURE exflib_float_ne_i1z0
         MODULE PROCEDURE exflib_float_ne_q1z0
         MODULE PROCEDURE exflib_float_ne_s1z0
         MODULE PROCEDURE exflib_float_ne_d1z0
         !MODULE PROCEDURE exflib_float_ne_x1z0
         MODULE PROCEDURE exflib_float_ne_c1z0
         MODULE PROCEDURE exflib_float_ne_w1z0
         MODULE PROCEDURE exflib_float_ne_z1z0

         MODULE PROCEDURE exflib_float_ne_i1z1
         MODULE PROCEDURE exflib_float_ne_q1z1
         MODULE PROCEDURE exflib_float_ne_s1z1
         MODULE PROCEDURE exflib_float_ne_d1z1
         !MODULE PROCEDURE exflib_float_ne_x1z1
         MODULE PROCEDURE exflib_float_ne_c1z1
         MODULE PROCEDURE exflib_float_ne_w1z1
         MODULE PROCEDURE exflib_float_ne_z1z1

         MODULE PROCEDURE exflib_float_ne_z1i0
         MODULE PROCEDURE exflib_float_ne_z1q0
         MODULE PROCEDURE exflib_float_ne_z1s0
         MODULE PROCEDURE exflib_float_ne_z1d0
         !MODULE PROCEDURE exflib_float_ne_z1x0
         MODULE PROCEDURE exflib_float_ne_z1c0
         MODULE PROCEDURE exflib_float_ne_z1w0
         !MODULE PROCEDURE exflib_float_ne_z1z0

         MODULE PROCEDURE exflib_float_ne_z0i1
         MODULE PROCEDURE exflib_float_ne_z0q1
         MODULE PROCEDURE exflib_float_ne_z0s1
         MODULE PROCEDURE exflib_float_ne_z0d1
         !MODULE PROCEDURE exflib_float_ne_z0x1
         MODULE PROCEDURE exflib_float_ne_z0c1
         MODULE PROCEDURE exflib_float_ne_z0w1
         !MODULE PROCEDURE exflib_float_ne_z0z1

         MODULE PROCEDURE exflib_float_ne_z1i1
         MODULE PROCEDURE exflib_float_ne_z1q1
         MODULE PROCEDURE exflib_float_ne_z1s1
         MODULE PROCEDURE exflib_float_ne_z1d1
         !MODULE PROCEDURE exflib_float_ne_z1x1
         MODULE PROCEDURE exflib_float_ne_z1c1
         MODULE PROCEDURE exflib_float_ne_z1w1
         !MODULE PROCEDURE exflib_float_ne_z1z1
      END INTERFACE

      INTERFACE OPERATOR ( < )
         MODULE PROCEDURE exflib_float_lt_i0x1
         MODULE PROCEDURE exflib_float_lt_q0x1
         MODULE PROCEDURE exflib_float_lt_s0x1
         MODULE PROCEDURE exflib_float_lt_d0x1
         MODULE PROCEDURE exflib_float_lt_x0x1

         MODULE PROCEDURE exflib_float_lt_i1x0
         MODULE PROCEDURE exflib_float_lt_q1x0
         MODULE PROCEDURE exflib_float_lt_s1x0
         MODULE PROCEDURE exflib_float_lt_d1x0
         MODULE PROCEDURE exflib_float_lt_x1x0

         MODULE PROCEDURE exflib_float_lt_i1x1
         MODULE PROCEDURE exflib_float_lt_q1x1
         MODULE PROCEDURE exflib_float_lt_s1x1
         MODULE PROCEDURE exflib_float_lt_d1x1
         MODULE PROCEDURE exflib_float_lt_x1x1

         MODULE PROCEDURE exflib_float_lt_x1i0
         MODULE PROCEDURE exflib_float_lt_x1q0
         MODULE PROCEDURE exflib_float_lt_x1s0
         MODULE PROCEDURE exflib_float_lt_x1d0

         MODULE PROCEDURE exflib_float_lt_x0i1
         MODULE PROCEDURE exflib_float_lt_x0q1
         MODULE PROCEDURE exflib_float_lt_x0s1
         MODULE PROCEDURE exflib_float_lt_x0d1

         MODULE PROCEDURE exflib_float_lt_x1i1
         MODULE PROCEDURE exflib_float_lt_x1q1
         MODULE PROCEDURE exflib_float_lt_x1s1
         MODULE PROCEDURE exflib_float_lt_x1d1
      END INTERFACE

      INTERFACE OPERATOR ( <= )
         MODULE PROCEDURE exflib_float_le_i0x1
         MODULE PROCEDURE exflib_float_le_q0x1
         MODULE PROCEDURE exflib_float_le_s0x1
         MODULE PROCEDURE exflib_float_le_d0x1
         MODULE PROCEDURE exflib_float_le_x0x1

         MODULE PROCEDURE exflib_float_le_i1x0
         MODULE PROCEDURE exflib_float_le_q1x0
         MODULE PROCEDURE exflib_float_le_s1x0
         MODULE PROCEDURE exflib_float_le_d1x0
         MODULE PROCEDURE exflib_float_le_x1x0

         MODULE PROCEDURE exflib_float_le_i1x1
         MODULE PROCEDURE exflib_float_le_q1x1
         MODULE PROCEDURE exflib_float_le_s1x1
         MODULE PROCEDURE exflib_float_le_d1x1
         MODULE PROCEDURE exflib_float_le_x1x1

         MODULE PROCEDURE exflib_float_le_x1i0
         MODULE PROCEDURE exflib_float_le_x1q0
         MODULE PROCEDURE exflib_float_le_x1s0
         MODULE PROCEDURE exflib_float_le_x1d0

         MODULE PROCEDURE exflib_float_le_x0i1
         MODULE PROCEDURE exflib_float_le_x0q1
         MODULE PROCEDURE exflib_float_le_x0s1
         MODULE PROCEDURE exflib_float_le_x0d1

         MODULE PROCEDURE exflib_float_le_x1i1
         MODULE PROCEDURE exflib_float_le_x1q1
         MODULE PROCEDURE exflib_float_le_x1s1
         MODULE PROCEDURE exflib_float_le_x1d1
      END INTERFACE

      INTERFACE OPERATOR ( > )
         MODULE PROCEDURE exflib_float_gt_i0x1
         MODULE PROCEDURE exflib_float_gt_q0x1
         MODULE PROCEDURE exflib_float_gt_s0x1
         MODULE PROCEDURE exflib_float_gt_d0x1
         MODULE PROCEDURE exflib_float_gt_x0x1

         MODULE PROCEDURE exflib_float_gt_i1x0
         MODULE PROCEDURE exflib_float_gt_q1x0
         MODULE PROCEDURE exflib_float_gt_s1x0
         MODULE PROCEDURE exflib_float_gt_d1x0
         MODULE PROCEDURE exflib_float_gt_x1x0

         MODULE PROCEDURE exflib_float_gt_i1x1
         MODULE PROCEDURE exflib_float_gt_q1x1
         MODULE PROCEDURE exflib_float_gt_s1x1
         MODULE PROCEDURE exflib_float_gt_d1x1
         MODULE PROCEDURE exflib_float_gt_x1x1

         MODULE PROCEDURE exflib_float_gt_x1i0
         MODULE PROCEDURE exflib_float_gt_x1q0
         MODULE PROCEDURE exflib_float_gt_x1s0
         MODULE PROCEDURE exflib_float_gt_x1d0

         MODULE PROCEDURE exflib_float_gt_x0i1
         MODULE PROCEDURE exflib_float_gt_x0q1
         MODULE PROCEDURE exflib_float_gt_x0s1
         MODULE PROCEDURE exflib_float_gt_x0d1

         MODULE PROCEDURE exflib_float_gt_x1i1
         MODULE PROCEDURE exflib_float_gt_x1q1
         MODULE PROCEDURE exflib_float_gt_x1s1
         MODULE PROCEDURE exflib_float_gt_x1d1
      END INTERFACE

      INTERFACE OPERATOR ( >= )
         MODULE PROCEDURE exflib_float_ge_i0x1
         MODULE PROCEDURE exflib_float_ge_q0x1
         MODULE PROCEDURE exflib_float_ge_s0x1
         MODULE PROCEDURE exflib_float_ge_d0x1
         MODULE PROCEDURE exflib_float_ge_x0x1

         MODULE PROCEDURE exflib_float_ge_i1x0
         MODULE PROCEDURE exflib_float_ge_q1x0
         MODULE PROCEDURE exflib_float_ge_s1x0
         MODULE PROCEDURE exflib_float_ge_d1x0
         MODULE PROCEDURE exflib_float_ge_x1x0

         MODULE PROCEDURE exflib_float_ge_i1x1
         MODULE PROCEDURE exflib_float_ge_q1x1
         MODULE PROCEDURE exflib_float_ge_s1x1
         MODULE PROCEDURE exflib_float_ge_d1x1
         MODULE PROCEDURE exflib_float_ge_x1x1

         MODULE PROCEDURE exflib_float_ge_x1i0
         MODULE PROCEDURE exflib_float_ge_x1q0
         MODULE PROCEDURE exflib_float_ge_x1s0
         MODULE PROCEDURE exflib_float_ge_x1d0

         MODULE PROCEDURE exflib_float_ge_x0i1
         MODULE PROCEDURE exflib_float_ge_x0q1
         MODULE PROCEDURE exflib_float_ge_x0s1
         MODULE PROCEDURE exflib_float_ge_x0d1

         MODULE PROCEDURE exflib_float_ge_x1i1
         MODULE PROCEDURE exflib_float_ge_x1q1
         MODULE PROCEDURE exflib_float_ge_x1s1
         MODULE PROCEDURE exflib_float_ge_x1d1
      END INTERFACE

      INTERFACE SUM
         MODULE PROCEDURE exflib_float_sum_x1
         MODULE PROCEDURE exflib_float_sum_z1
      END INTERFACE

      INTERFACE PRODUCT
         MODULE PROCEDURE exflib_float_product_x1
         MODULE PROCEDURE exflib_float_product_z1
      END INTERFACE

      INTERFACE MAXVAL
         MODULE PROCEDURE exflib_float_maxval_x1
      END INTERFACE

      INTERFACE MINVAL
         MODULE PROCEDURE exflib_float_minval_x1
      END INTERFACE

      INTERFACE MAXLOC
         MODULE PROCEDURE exflib_float_maxloc_x1
      END INTERFACE

      INTERFACE MINLOC
         MODULE PROCEDURE exflib_float_minloc_x1
      END INTERFACE

      INTERFACE DOT_PRODUCT
         MODULE PROCEDURE exflib_float_dot_product_x1x1
         MODULE PROCEDURE exflib_float_dot_product_x1z1
         MODULE PROCEDURE exflib_float_dot_product_z1x1
         MODULE PROCEDURE exflib_float_dot_product_z1z1
      END INTERFACE

      INTERFACE OPERATOR ( + )
         ! exfloat
         MODULE PROCEDURE exflib_float_add_i0x1
         MODULE PROCEDURE exflib_float_add_q0x1
         MODULE PROCEDURE exflib_float_add_x0x1
         MODULE PROCEDURE exflib_float_add_z0x1

         MODULE PROCEDURE exflib_float_add_i1x0
         MODULE PROCEDURE exflib_float_add_q1x0
         MODULE PROCEDURE exflib_float_add_x1x0
         MODULE PROCEDURE exflib_float_add_z1x0

         MODULE PROCEDURE exflib_float_add_i1x1
         MODULE PROCEDURE exflib_float_add_q1x1
         MODULE PROCEDURE exflib_float_add_x1x1
         MODULE PROCEDURE exflib_float_add_z1x1

         MODULE PROCEDURE exflib_float_add_x0i1
         MODULE PROCEDURE exflib_float_add_x0q1
         MODULE PROCEDURE exflib_float_add_x0z1

         MODULE PROCEDURE exflib_float_add_x1i0
         MODULE PROCEDURE exflib_float_add_x1q0
         MODULE PROCEDURE exflib_float_add_x1z0

         MODULE PROCEDURE exflib_float_add_x1i1
         MODULE PROCEDURE exflib_float_add_x1q1
         MODULE PROCEDURE exflib_float_add_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_add_i0z1
         MODULE PROCEDURE exflib_float_add_q0z1
         MODULE PROCEDURE exflib_float_add_z0z1

         MODULE PROCEDURE exflib_float_add_i1z0
         MODULE PROCEDURE exflib_float_add_q1z0
         MODULE PROCEDURE exflib_float_add_z1z0

         MODULE PROCEDURE exflib_float_add_i1z1
         MODULE PROCEDURE exflib_float_add_q1z1
         MODULE PROCEDURE exflib_float_add_z1z1

         MODULE PROCEDURE exflib_float_add_z0i1
         MODULE PROCEDURE exflib_float_add_z0q1

         MODULE PROCEDURE exflib_float_add_z1i0
         MODULE PROCEDURE exflib_float_add_z1q0

         MODULE PROCEDURE exflib_float_add_z1i1
         MODULE PROCEDURE exflib_float_add_z1q1

         ! unarray operator
         MODULE PROCEDURE exflib_float_pls_x1
         MODULE PROCEDURE exflib_float_pls_z1
      END INTERFACE

      INTERFACE OPERATOR ( - )
         ! exfloat
         MODULE PROCEDURE exflib_float_sub_i0x1
         MODULE PROCEDURE exflib_float_sub_q0x1
         MODULE PROCEDURE exflib_float_sub_x0x1
         MODULE PROCEDURE exflib_float_sub_z0x1

         MODULE PROCEDURE exflib_float_sub_i1x0
         MODULE PROCEDURE exflib_float_sub_q1x0
         MODULE PROCEDURE exflib_float_sub_x1x0
         MODULE PROCEDURE exflib_float_sub_z1x0

         MODULE PROCEDURE exflib_float_sub_i1x1
         MODULE PROCEDURE exflib_float_sub_q1x1
         MODULE PROCEDURE exflib_float_sub_x1x1
         MODULE PROCEDURE exflib_float_sub_z1x1

         MODULE PROCEDURE exflib_float_sub_x0i1
         MODULE PROCEDURE exflib_float_sub_x0q1
         MODULE PROCEDURE exflib_float_sub_x0z1

         MODULE PROCEDURE exflib_float_sub_x1i0
         MODULE PROCEDURE exflib_float_sub_x1q0
         MODULE PROCEDURE exflib_float_sub_x1z0

         MODULE PROCEDURE exflib_float_sub_x1i1
         MODULE PROCEDURE exflib_float_sub_x1q1
         MODULE PROCEDURE exflib_float_sub_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_sub_i0z1
         MODULE PROCEDURE exflib_float_sub_q0z1
         MODULE PROCEDURE exflib_float_sub_z0z1

         MODULE PROCEDURE exflib_float_sub_i1z0
         MODULE PROCEDURE exflib_float_sub_q1z0
         MODULE PROCEDURE exflib_float_sub_z1z0

         MODULE PROCEDURE exflib_float_sub_i1z1
         MODULE PROCEDURE exflib_float_sub_q1z1
         MODULE PROCEDURE exflib_float_sub_z1z1

         MODULE PROCEDURE exflib_float_sub_z0i1
         MODULE PROCEDURE exflib_float_sub_z0q1

         MODULE PROCEDURE exflib_float_sub_z1i0
         MODULE PROCEDURE exflib_float_sub_z1q0

         MODULE PROCEDURE exflib_float_sub_z1i1
         MODULE PROCEDURE exflib_float_sub_z1q1

         ! unarray operator
         MODULE PROCEDURE exflib_float_neg_x1
         MODULE PROCEDURE exflib_float_neg_z1
      END INTERFACE

      INTERFACE OPERATOR ( * )
         ! exfloat
         MODULE PROCEDURE exflib_float_mul_i0x1
         MODULE PROCEDURE exflib_float_mul_q0x1
         MODULE PROCEDURE exflib_float_mul_x0x1
         MODULE PROCEDURE exflib_float_mul_z0x1

         MODULE PROCEDURE exflib_float_mul_i1x0
         MODULE PROCEDURE exflib_float_mul_q1x0
         MODULE PROCEDURE exflib_float_mul_x1x0
         MODULE PROCEDURE exflib_float_mul_z1x0

         MODULE PROCEDURE exflib_float_mul_i1x1
         MODULE PROCEDURE exflib_float_mul_q1x1
         MODULE PROCEDURE exflib_float_mul_x1x1
         MODULE PROCEDURE exflib_float_mul_z1x1

         MODULE PROCEDURE exflib_float_mul_x0i1
         MODULE PROCEDURE exflib_float_mul_x0q1
         MODULE PROCEDURE exflib_float_mul_x0z1

         MODULE PROCEDURE exflib_float_mul_x1i0
         MODULE PROCEDURE exflib_float_mul_x1q0
         MODULE PROCEDURE exflib_float_mul_x1z0

         MODULE PROCEDURE exflib_float_mul_x1i1
         MODULE PROCEDURE exflib_float_mul_x1q1
         MODULE PROCEDURE exflib_float_mul_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_mul_i0z1
         MODULE PROCEDURE exflib_float_mul_q0z1
         MODULE PROCEDURE exflib_float_mul_z0z1

         MODULE PROCEDURE exflib_float_mul_i1z0
         MODULE PROCEDURE exflib_float_mul_q1z0
         MODULE PROCEDURE exflib_float_mul_z1z0

         MODULE PROCEDURE exflib_float_mul_i1z1
         MODULE PROCEDURE exflib_float_mul_q1z1
         MODULE PROCEDURE exflib_float_mul_z1z1

         MODULE PROCEDURE exflib_float_mul_z0i1
         MODULE PROCEDURE exflib_float_mul_z0q1

         MODULE PROCEDURE exflib_float_mul_z1i0
         MODULE PROCEDURE exflib_float_mul_z1q0

         MODULE PROCEDURE exflib_float_mul_z1i1
         MODULE PROCEDURE exflib_float_mul_z1q1
      END INTERFACE

      INTERFACE OPERATOR ( / )
         ! exfloat
         MODULE PROCEDURE exflib_float_div_i0x1
         MODULE PROCEDURE exflib_float_div_q0x1
         MODULE PROCEDURE exflib_float_div_x0x1
         MODULE PROCEDURE exflib_float_div_z0x1

         MODULE PROCEDURE exflib_float_div_i1x0
         MODULE PROCEDURE exflib_float_div_q1x0
         MODULE PROCEDURE exflib_float_div_x1x0
         MODULE PROCEDURE exflib_float_div_z1x0

         MODULE PROCEDURE exflib_float_div_i1x1
         MODULE PROCEDURE exflib_float_div_q1x1
         MODULE PROCEDURE exflib_float_div_x1x1
         MODULE PROCEDURE exflib_float_div_z1x1

         MODULE PROCEDURE exflib_float_div_x0i1
         MODULE PROCEDURE exflib_float_div_x0q1
         MODULE PROCEDURE exflib_float_div_x0z1

         MODULE PROCEDURE exflib_float_div_x1i0
         MODULE PROCEDURE exflib_float_div_x1q0
         MODULE PROCEDURE exflib_float_div_x1z0

         MODULE PROCEDURE exflib_float_div_x1i1
         MODULE PROCEDURE exflib_float_div_x1q1
         MODULE PROCEDURE exflib_float_div_x1z1

         ! ezfloat
         MODULE PROCEDURE exflib_float_div_i0z1
         MODULE PROCEDURE exflib_float_div_q0z1
         MODULE PROCEDURE exflib_float_div_z0z1

         MODULE PROCEDURE exflib_float_div_i1z0
         MODULE PROCEDURE exflib_float_div_q1z0
         MODULE PROCEDURE exflib_float_div_z1z0

         MODULE PROCEDURE exflib_float_div_i1z1
         MODULE PROCEDURE exflib_float_div_q1z1
         MODULE PROCEDURE exflib_float_div_z1z1

         MODULE PROCEDURE exflib_float_div_z0i1
         MODULE PROCEDURE exflib_float_div_z0q1

         MODULE PROCEDURE exflib_float_div_z1i0
         MODULE PROCEDURE exflib_float_div_z1q0

         MODULE PROCEDURE exflib_float_div_z1i1
         MODULE PROCEDURE exflib_float_div_z1q1
      END INTERFACE

      INTERFACE POW
         MODULE PROCEDURE exflib_float_pow_x1i0
         MODULE PROCEDURE exflib_float_pow_x1q0
         MODULE PROCEDURE exflib_float_pow_z1i0
         MODULE PROCEDURE exflib_float_pow_z1q0
      END INTERFACE

      INTERFACE OPERATOR ( ** )
         MODULE PROCEDURE exflib_float_pow_x1i0
         MODULE PROCEDURE exflib_float_pow_x1q0
         MODULE PROCEDURE exflib_float_pow_z1i0
         MODULE PROCEDURE exflib_float_pow_z1q0
      END INTERFACE

      !--------------------------------------------------
      ! two dimensional array (matrix) functions
      !--------------------------------------------------
      INTERFACE ASSIGNMENT ( = )
         MODULE PROCEDURE exflib_float_assign_x2i0
         MODULE PROCEDURE exflib_float_assign_x2q0
         MODULE PROCEDURE exflib_float_assign_x2x0

         MODULE PROCEDURE exflib_float_assign_x2i2
         MODULE PROCEDURE exflib_float_assign_x2q2

         MODULE PROCEDURE exflib_float_assign_z2i0
         MODULE PROCEDURE exflib_float_assign_z2q0
         MODULE PROCEDURE exflib_float_assign_z2x0
         MODULE PROCEDURE exflib_float_assign_z2z0

         MODULE PROCEDURE exflib_float_assign_z2i2
         MODULE PROCEDURE exflib_float_assign_z2q2
         MODULE PROCEDURE exflib_float_assign_z2x2
      END INTERFACE

      INTERFACE exflib_double
         MODULE PROCEDURE exflib_float_cast_x2d2
      END INTERFACE
      INTERFACE DBLE
         MODULE PROCEDURE exflib_float_cast_x2d2
      END INTERFACE
!      INTERFACE REAL
!      END INTERFACE
!      INTERFACE QREAL
!      END INTERFACE
      INTERFACE exflib_complex
         MODULE PROCEDURE exflib_float_cast_z2w2
      END INTERFACE
      INTERFACE CMPLX
         MODULE PROCEDURE exflib_float_cast_z2w2
      END INTERFACE
      INTERFACE DCMPLX
         MODULE PROCEDURE exflib_float_cast_z2w2
      END INTERFACE
!      INTERFACE QCMPLX
!      END INTERFACE

      INTERFACE OPERATOR ( == )
         ! exfloat
         MODULE PROCEDURE exflib_float_eq_i0x2
         MODULE PROCEDURE exflib_float_eq_q0x2
         MODULE PROCEDURE exflib_float_eq_s0x2
         MODULE PROCEDURE exflib_float_eq_d0x2
         MODULE PROCEDURE exflib_float_eq_x0x2
         MODULE PROCEDURE exflib_float_eq_c0x2
         MODULE PROCEDURE exflib_float_eq_w0x2
         MODULE PROCEDURE exflib_float_eq_z0x2

         MODULE PROCEDURE exflib_float_eq_i2x0
         MODULE PROCEDURE exflib_float_eq_q2x0
         MODULE PROCEDURE exflib_float_eq_s2x0
         MODULE PROCEDURE exflib_float_eq_d2x0
         MODULE PROCEDURE exflib_float_eq_x2x0
         MODULE PROCEDURE exflib_float_eq_c2x0
         MODULE PROCEDURE exflib_float_eq_w2x0
         MODULE PROCEDURE exflib_float_eq_z2x0

         MODULE PROCEDURE exflib_float_eq_i2x2
         MODULE PROCEDURE exflib_float_eq_q2x2
         MODULE PROCEDURE exflib_float_eq_s2x2
         MODULE PROCEDURE exflib_float_eq_d2x2
         MODULE PROCEDURE exflib_float_eq_x2x2
         MODULE PROCEDURE exflib_float_eq_c2x2
         MODULE PROCEDURE exflib_float_eq_w2x2
         MODULE PROCEDURE exflib_float_eq_z2x2

         MODULE PROCEDURE exflib_float_eq_x2i0
         MODULE PROCEDURE exflib_float_eq_x2q0
         MODULE PROCEDURE exflib_float_eq_x2s0
         MODULE PROCEDURE exflib_float_eq_x2d0
         !MODULE PROCEDURE exflib_float_eq_x2x0
         MODULE PROCEDURE exflib_float_eq_x2c0
         MODULE PROCEDURE exflib_float_eq_x2w0
         MODULE PROCEDURE exflib_float_eq_x2z0

         MODULE PROCEDURE exflib_float_eq_x0i2
         MODULE PROCEDURE exflib_float_eq_x0q2
         MODULE PROCEDURE exflib_float_eq_x0s2
         MODULE PROCEDURE exflib_float_eq_x0d2
         !MODULE PROCEDURE exflib_float_eq_x0x2
         MODULE PROCEDURE exflib_float_eq_x0c2
         MODULE PROCEDURE exflib_float_eq_x0w2
         MODULE PROCEDURE exflib_float_eq_x0z2

         MODULE PROCEDURE exflib_float_eq_x2i2
         MODULE PROCEDURE exflib_float_eq_x2q2
         MODULE PROCEDURE exflib_float_eq_x2s2
         MODULE PROCEDURE exflib_float_eq_x2d2
         !MODULE PROCEDURE exflib_float_eq_x2x2
         MODULE PROCEDURE exflib_float_eq_x2c2
         MODULE PROCEDURE exflib_float_eq_x2w2
         MODULE PROCEDURE exflib_float_eq_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_eq_i0z2
         MODULE PROCEDURE exflib_float_eq_q0z2
         MODULE PROCEDURE exflib_float_eq_s0z2
         MODULE PROCEDURE exflib_float_eq_d0z2
         !MODULE PROCEDURE exflib_float_eq_x0z2
         MODULE PROCEDURE exflib_float_eq_c0z2
         MODULE PROCEDURE exflib_float_eq_w0z2
         MODULE PROCEDURE exflib_float_eq_z0z2

         MODULE PROCEDURE exflib_float_eq_i2z0
         MODULE PROCEDURE exflib_float_eq_q2z0
         MODULE PROCEDURE exflib_float_eq_s2z0
         MODULE PROCEDURE exflib_float_eq_d2z0
         !MODULE PROCEDURE exflib_float_eq_x2z0
         MODULE PROCEDURE exflib_float_eq_c2z0
         MODULE PROCEDURE exflib_float_eq_w2z0
         MODULE PROCEDURE exflib_float_eq_z2z0

         MODULE PROCEDURE exflib_float_eq_i2z2
         MODULE PROCEDURE exflib_float_eq_q2z2
         MODULE PROCEDURE exflib_float_eq_s2z2
         MODULE PROCEDURE exflib_float_eq_d2z2
         !MODULE PROCEDURE exflib_float_eq_x2z2
         MODULE PROCEDURE exflib_float_eq_c2z2
         MODULE PROCEDURE exflib_float_eq_w2z2
         MODULE PROCEDURE exflib_float_eq_z2z2

         MODULE PROCEDURE exflib_float_eq_z2i0
         MODULE PROCEDURE exflib_float_eq_z2q0
         MODULE PROCEDURE exflib_float_eq_z2s0
         MODULE PROCEDURE exflib_float_eq_z2d0
         !MODULE PROCEDURE exflib_float_eq_z2x0
         MODULE PROCEDURE exflib_float_eq_z2c0
         MODULE PROCEDURE exflib_float_eq_z2w0
         !MODULE PROCEDURE exflib_float_eq_z2z0

         MODULE PROCEDURE exflib_float_eq_z0i2
         MODULE PROCEDURE exflib_float_eq_z0q2
         MODULE PROCEDURE exflib_float_eq_z0s2
         MODULE PROCEDURE exflib_float_eq_z0d2
         !MODULE PROCEDURE exflib_float_eq_z0x2
         MODULE PROCEDURE exflib_float_eq_z0c2
         MODULE PROCEDURE exflib_float_eq_z0w2
         !MODULE PROCEDURE exflib_float_eq_z0z2

         MODULE PROCEDURE exflib_float_eq_z2i2
         MODULE PROCEDURE exflib_float_eq_z2q2
         MODULE PROCEDURE exflib_float_eq_z2s2
         MODULE PROCEDURE exflib_float_eq_z2d2
         !MODULE PROCEDURE exflib_float_eq_z2x2
         MODULE PROCEDURE exflib_float_eq_z2c2
         MODULE PROCEDURE exflib_float_eq_z2w2
         !MODULE PROCEDURE exflib_float_eq_z2z2
      END INTERFACE

      INTERFACE OPERATOR ( /= )
         ! exfloat
         MODULE PROCEDURE exflib_float_ne_i0x2
         MODULE PROCEDURE exflib_float_ne_q0x2
         MODULE PROCEDURE exflib_float_ne_s0x2
         MODULE PROCEDURE exflib_float_ne_d0x2
         MODULE PROCEDURE exflib_float_ne_x0x2
         MODULE PROCEDURE exflib_float_ne_c0x2
         MODULE PROCEDURE exflib_float_ne_w0x2
         MODULE PROCEDURE exflib_float_ne_z0x2

         MODULE PROCEDURE exflib_float_ne_i2x0
         MODULE PROCEDURE exflib_float_ne_q2x0
         MODULE PROCEDURE exflib_float_ne_s2x0
         MODULE PROCEDURE exflib_float_ne_d2x0
         MODULE PROCEDURE exflib_float_ne_x2x0
         MODULE PROCEDURE exflib_float_ne_c2x0
         MODULE PROCEDURE exflib_float_ne_w2x0
         MODULE PROCEDURE exflib_float_ne_z2x0

         MODULE PROCEDURE exflib_float_ne_i2x2
         MODULE PROCEDURE exflib_float_ne_q2x2
         MODULE PROCEDURE exflib_float_ne_s2x2
         MODULE PROCEDURE exflib_float_ne_d2x2
         MODULE PROCEDURE exflib_float_ne_x2x2
         MODULE PROCEDURE exflib_float_ne_c2x2
         MODULE PROCEDURE exflib_float_ne_w2x2
         MODULE PROCEDURE exflib_float_ne_z2x2

         MODULE PROCEDURE exflib_float_ne_x2i0
         MODULE PROCEDURE exflib_float_ne_x2q0
         MODULE PROCEDURE exflib_float_ne_x2s0
         MODULE PROCEDURE exflib_float_ne_x2d0
         !MODULE PROCEDURE exflib_float_ne_x2x0
         MODULE PROCEDURE exflib_float_ne_x2c0
         MODULE PROCEDURE exflib_float_ne_x2w0
         MODULE PROCEDURE exflib_float_ne_x2z0

         MODULE PROCEDURE exflib_float_ne_x0i2
         MODULE PROCEDURE exflib_float_ne_x0q2
         MODULE PROCEDURE exflib_float_ne_x0s2
         MODULE PROCEDURE exflib_float_ne_x0d2
         !MODULE PROCEDURE exflib_float_ne_x0x2
         MODULE PROCEDURE exflib_float_ne_x0c2
         MODULE PROCEDURE exflib_float_ne_x0w2
         MODULE PROCEDURE exflib_float_ne_x0z2

         MODULE PROCEDURE exflib_float_ne_x2i2
         MODULE PROCEDURE exflib_float_ne_x2q2
         MODULE PROCEDURE exflib_float_ne_x2s2
         MODULE PROCEDURE exflib_float_ne_x2d2
         !MODULE PROCEDURE exflib_float_ne_x2x2
         MODULE PROCEDURE exflib_float_ne_x2c2
         MODULE PROCEDURE exflib_float_ne_x2w2
         MODULE PROCEDURE exflib_float_ne_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_ne_i0z2
         MODULE PROCEDURE exflib_float_ne_q0z2
         MODULE PROCEDURE exflib_float_ne_s0z2
         MODULE PROCEDURE exflib_float_ne_d0z2
         !MODULE PROCEDURE exflib_float_ne_x0z2
         MODULE PROCEDURE exflib_float_ne_c0z2
         MODULE PROCEDURE exflib_float_ne_w0z2
         MODULE PROCEDURE exflib_float_ne_z0z2

         MODULE PROCEDURE exflib_float_ne_i2z0
         MODULE PROCEDURE exflib_float_ne_q2z0
         MODULE PROCEDURE exflib_float_ne_s2z0
         MODULE PROCEDURE exflib_float_ne_d2z0
         !MODULE PROCEDURE exflib_float_ne_x2z0
         MODULE PROCEDURE exflib_float_ne_c2z0
         MODULE PROCEDURE exflib_float_ne_w2z0
         MODULE PROCEDURE exflib_float_ne_z2z0

         MODULE PROCEDURE exflib_float_ne_i2z2
         MODULE PROCEDURE exflib_float_ne_q2z2
         MODULE PROCEDURE exflib_float_ne_s2z2
         MODULE PROCEDURE exflib_float_ne_d2z2
         !MODULE PROCEDURE exflib_float_ne_x2z2
         MODULE PROCEDURE exflib_float_ne_c2z2
         MODULE PROCEDURE exflib_float_ne_w2z2
         MODULE PROCEDURE exflib_float_ne_z2z2

         MODULE PROCEDURE exflib_float_ne_z2i0
         MODULE PROCEDURE exflib_float_ne_z2q0
         MODULE PROCEDURE exflib_float_ne_z2s0
         MODULE PROCEDURE exflib_float_ne_z2d0
         !MODULE PROCEDURE exflib_float_ne_z2x0
         MODULE PROCEDURE exflib_float_ne_z2c0
         MODULE PROCEDURE exflib_float_ne_z2w0
         !MODULE PROCEDURE exflib_float_ne_z2z0

         MODULE PROCEDURE exflib_float_ne_z0i2
         MODULE PROCEDURE exflib_float_ne_z0q2
         MODULE PROCEDURE exflib_float_ne_z0s2
         MODULE PROCEDURE exflib_float_ne_z0d2
         !MODULE PROCEDURE exflib_float_ne_z0x2
         MODULE PROCEDURE exflib_float_ne_z0c2
         MODULE PROCEDURE exflib_float_ne_z0w2
         !MODULE PROCEDURE exflib_float_ne_z0z2

         MODULE PROCEDURE exflib_float_ne_z2i2
         MODULE PROCEDURE exflib_float_ne_z2q2
         MODULE PROCEDURE exflib_float_ne_z2s2
         MODULE PROCEDURE exflib_float_ne_z2d2
         !MODULE PROCEDURE exflib_float_ne_z2x2
         MODULE PROCEDURE exflib_float_ne_z2c2
         MODULE PROCEDURE exflib_float_ne_z2w2
         !MODULE PROCEDURE exflib_float_ne_z2z2
      END INTERFACE

      INTERFACE OPERATOR ( < )
         MODULE PROCEDURE exflib_float_lt_i0x2
         MODULE PROCEDURE exflib_float_lt_q0x2
         MODULE PROCEDURE exflib_float_lt_s0x2
         MODULE PROCEDURE exflib_float_lt_d0x2
         MODULE PROCEDURE exflib_float_lt_x0x2

         MODULE PROCEDURE exflib_float_lt_i2x0
         MODULE PROCEDURE exflib_float_lt_q2x0
         MODULE PROCEDURE exflib_float_lt_s2x0
         MODULE PROCEDURE exflib_float_lt_d2x0
         MODULE PROCEDURE exflib_float_lt_x2x0

         MODULE PROCEDURE exflib_float_lt_i2x2
         MODULE PROCEDURE exflib_float_lt_q2x2
         MODULE PROCEDURE exflib_float_lt_s2x2
         MODULE PROCEDURE exflib_float_lt_d2x2
         MODULE PROCEDURE exflib_float_lt_x2x2

         MODULE PROCEDURE exflib_float_lt_x2i0
         MODULE PROCEDURE exflib_float_lt_x2q0
         MODULE PROCEDURE exflib_float_lt_x2s0
         MODULE PROCEDURE exflib_float_lt_x2d0

         MODULE PROCEDURE exflib_float_lt_x0i2
         MODULE PROCEDURE exflib_float_lt_x0q2
         MODULE PROCEDURE exflib_float_lt_x0s2
         MODULE PROCEDURE exflib_float_lt_x0d2

         MODULE PROCEDURE exflib_float_lt_x2i2
         MODULE PROCEDURE exflib_float_lt_x2q2
         MODULE PROCEDURE exflib_float_lt_x2s2
         MODULE PROCEDURE exflib_float_lt_x2d2
      END INTERFACE

      INTERFACE OPERATOR ( <= )
         MODULE PROCEDURE exflib_float_le_i0x2
         MODULE PROCEDURE exflib_float_le_q0x2
         MODULE PROCEDURE exflib_float_le_s0x2
         MODULE PROCEDURE exflib_float_le_d0x2
         MODULE PROCEDURE exflib_float_le_x0x2

         MODULE PROCEDURE exflib_float_le_i2x0
         MODULE PROCEDURE exflib_float_le_q2x0
         MODULE PROCEDURE exflib_float_le_s2x0
         MODULE PROCEDURE exflib_float_le_d2x0
         MODULE PROCEDURE exflib_float_le_x2x0

         MODULE PROCEDURE exflib_float_le_i2x2
         MODULE PROCEDURE exflib_float_le_q2x2
         MODULE PROCEDURE exflib_float_le_s2x2
         MODULE PROCEDURE exflib_float_le_d2x2
         MODULE PROCEDURE exflib_float_le_x2x2

         MODULE PROCEDURE exflib_float_le_x2i0
         MODULE PROCEDURE exflib_float_le_x2q0
         MODULE PROCEDURE exflib_float_le_x2s0
         MODULE PROCEDURE exflib_float_le_x2d0

         MODULE PROCEDURE exflib_float_le_x0i2
         MODULE PROCEDURE exflib_float_le_x0q2
         MODULE PROCEDURE exflib_float_le_x0s2
         MODULE PROCEDURE exflib_float_le_x0d2

         MODULE PROCEDURE exflib_float_le_x2i2
         MODULE PROCEDURE exflib_float_le_x2q2
         MODULE PROCEDURE exflib_float_le_x2s2
         MODULE PROCEDURE exflib_float_le_x2d2
      END INTERFACE

      INTERFACE OPERATOR ( > )
         MODULE PROCEDURE exflib_float_gt_i0x2
         MODULE PROCEDURE exflib_float_gt_q0x2
         MODULE PROCEDURE exflib_float_gt_s0x2
         MODULE PROCEDURE exflib_float_gt_d0x2
         MODULE PROCEDURE exflib_float_gt_x0x2

         MODULE PROCEDURE exflib_float_gt_i2x0
         MODULE PROCEDURE exflib_float_gt_q2x0
         MODULE PROCEDURE exflib_float_gt_s2x0
         MODULE PROCEDURE exflib_float_gt_d2x0
         MODULE PROCEDURE exflib_float_gt_x2x0

         MODULE PROCEDURE exflib_float_gt_i2x2
         MODULE PROCEDURE exflib_float_gt_q2x2
         MODULE PROCEDURE exflib_float_gt_s2x2
         MODULE PROCEDURE exflib_float_gt_d2x2
         MODULE PROCEDURE exflib_float_gt_x2x2

         MODULE PROCEDURE exflib_float_gt_x2i0
         MODULE PROCEDURE exflib_float_gt_x2q0
         MODULE PROCEDURE exflib_float_gt_x2s0
         MODULE PROCEDURE exflib_float_gt_x2d0

         MODULE PROCEDURE exflib_float_gt_x0i2
         MODULE PROCEDURE exflib_float_gt_x0q2
         MODULE PROCEDURE exflib_float_gt_x0s2
         MODULE PROCEDURE exflib_float_gt_x0d2

         MODULE PROCEDURE exflib_float_gt_x2i2
         MODULE PROCEDURE exflib_float_gt_x2q2
         MODULE PROCEDURE exflib_float_gt_x2s2
         MODULE PROCEDURE exflib_float_gt_x2d2
      END INTERFACE

      INTERFACE OPERATOR ( >= )
         MODULE PROCEDURE exflib_float_ge_i0x2
         MODULE PROCEDURE exflib_float_ge_q0x2
         MODULE PROCEDURE exflib_float_ge_s0x2
         MODULE PROCEDURE exflib_float_ge_d0x2
         MODULE PROCEDURE exflib_float_ge_x0x2

         MODULE PROCEDURE exflib_float_ge_i2x0
         MODULE PROCEDURE exflib_float_ge_q2x0
         MODULE PROCEDURE exflib_float_ge_s2x0
         MODULE PROCEDURE exflib_float_ge_d2x0
         MODULE PROCEDURE exflib_float_ge_x2x0

         MODULE PROCEDURE exflib_float_ge_i2x2
         MODULE PROCEDURE exflib_float_ge_q2x2
         MODULE PROCEDURE exflib_float_ge_s2x2
         MODULE PROCEDURE exflib_float_ge_d2x2
         MODULE PROCEDURE exflib_float_ge_x2x2

         MODULE PROCEDURE exflib_float_ge_x2i0
         MODULE PROCEDURE exflib_float_ge_x2q0
         MODULE PROCEDURE exflib_float_ge_x2s0
         MODULE PROCEDURE exflib_float_ge_x2d0

         MODULE PROCEDURE exflib_float_ge_x0i2
         MODULE PROCEDURE exflib_float_ge_x0q2
         MODULE PROCEDURE exflib_float_ge_x0s2
         MODULE PROCEDURE exflib_float_ge_x0d2

         MODULE PROCEDURE exflib_float_ge_x2i2
         MODULE PROCEDURE exflib_float_ge_x2q2
         MODULE PROCEDURE exflib_float_ge_x2s2
         MODULE PROCEDURE exflib_float_ge_x2d2
      END INTERFACE

      INTERFACE SUM
         MODULE PROCEDURE exflib_float_sum_x2
         MODULE PROCEDURE exflib_float_sum_dim_x2
         MODULE PROCEDURE exflib_float_sum_z2
         MODULE PROCEDURE exflib_float_sum_dim_z2
      END INTERFACE

      INTERFACE PRODUCT
         MODULE PROCEDURE exflib_float_product_x2
         MODULE PROCEDURE exflib_float_product_dim_x2
         MODULE PROCEDURE exflib_float_product_z2
         MODULE PROCEDURE exflib_float_product_dim_z2
      END INTERFACE

      INTERFACE MAXVAL
         MODULE PROCEDURE exflib_float_maxval_x2
         MODULE PROCEDURE exflib_float_maxval_dim_x2
      END INTERFACE

      INTERFACE MINVAL
         MODULE PROCEDURE exflib_float_minval_x2
         MODULE PROCEDURE exflib_float_minval_dim_x2
      END INTERFACE

      INTERFACE MAXLOC
         MODULE PROCEDURE exflib_float_maxloc_x2
         MODULE PROCEDURE exflib_float_maxloc_dim_x2
      END INTERFACE

      INTERFACE MINLOC
         MODULE PROCEDURE exflib_float_minloc_x2
         MODULE PROCEDURE exflib_float_minloc_dim_x2
      END INTERFACE

      INTERFACE MATMUL
         ! no optimized
         MODULE PROCEDURE exflib_float_matmul_x2x2
         MODULE PROCEDURE exflib_float_matmul_x2z2
         MODULE PROCEDURE exflib_float_matmul_z2x2
         MODULE PROCEDURE exflib_float_matmul_z2z2
      END INTERFACE

      INTERFACE TRANSPOSE
         MODULE PROCEDURE exflib_float_transpose_x2
         MODULE PROCEDURE exflib_float_transpose_z2
      END INTERFACE

      INTERFACE OPERATOR ( + )
         ! exfloat
         MODULE PROCEDURE exflib_float_add_i0x2
         MODULE PROCEDURE exflib_float_add_q0x2
         MODULE PROCEDURE exflib_float_add_x0x2
         MODULE PROCEDURE exflib_float_add_z0x2

         MODULE PROCEDURE exflib_float_add_i2x0
         MODULE PROCEDURE exflib_float_add_q2x0
         MODULE PROCEDURE exflib_float_add_x2x0
         MODULE PROCEDURE exflib_float_add_z2x0

         MODULE PROCEDURE exflib_float_add_i2x2
         MODULE PROCEDURE exflib_float_add_q2x2
         MODULE PROCEDURE exflib_float_add_x2x2
         MODULE PROCEDURE exflib_float_add_z2x2

         MODULE PROCEDURE exflib_float_add_x0i2
         MODULE PROCEDURE exflib_float_add_x0q2
         MODULE PROCEDURE exflib_float_add_x0z2

         MODULE PROCEDURE exflib_float_add_x2i0
         MODULE PROCEDURE exflib_float_add_x2q0
         MODULE PROCEDURE exflib_float_add_x2z0

         MODULE PROCEDURE exflib_float_add_x2i2
         MODULE PROCEDURE exflib_float_add_x2q2
         MODULE PROCEDURE exflib_float_add_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_add_i0z2
         MODULE PROCEDURE exflib_float_add_q0z2
         MODULE PROCEDURE exflib_float_add_z0z2

         MODULE PROCEDURE exflib_float_add_i2z0
         MODULE PROCEDURE exflib_float_add_q2z0
         MODULE PROCEDURE exflib_float_add_z2z0

         MODULE PROCEDURE exflib_float_add_i2z2
         MODULE PROCEDURE exflib_float_add_q2z2
         MODULE PROCEDURE exflib_float_add_z2z2

         MODULE PROCEDURE exflib_float_add_z0i2
         MODULE PROCEDURE exflib_float_add_z0q2

         MODULE PROCEDURE exflib_float_add_z2i0
         MODULE PROCEDURE exflib_float_add_z2q0

         MODULE PROCEDURE exflib_float_add_z2i2
         MODULE PROCEDURE exflib_float_add_z2q2

         ! unarray operator
         MODULE PROCEDURE exflib_float_pls_x2
         MODULE PROCEDURE exflib_float_pls_z2
      END INTERFACE

      INTERFACE OPERATOR ( - )
         ! exfloat
         MODULE PROCEDURE exflib_float_sub_i0x2
         MODULE PROCEDURE exflib_float_sub_q0x2
         MODULE PROCEDURE exflib_float_sub_x0x2
         MODULE PROCEDURE exflib_float_sub_z0x2

         MODULE PROCEDURE exflib_float_sub_i2x0
         MODULE PROCEDURE exflib_float_sub_q2x0
         MODULE PROCEDURE exflib_float_sub_x2x0
         MODULE PROCEDURE exflib_float_sub_z2x0

         MODULE PROCEDURE exflib_float_sub_i2x2
         MODULE PROCEDURE exflib_float_sub_q2x2
         MODULE PROCEDURE exflib_float_sub_x2x2
         MODULE PROCEDURE exflib_float_sub_z2x2

         MODULE PROCEDURE exflib_float_sub_x0i2
         MODULE PROCEDURE exflib_float_sub_x0q2
         MODULE PROCEDURE exflib_float_sub_x0z2

         MODULE PROCEDURE exflib_float_sub_x2i0
         MODULE PROCEDURE exflib_float_sub_x2q0
         MODULE PROCEDURE exflib_float_sub_x2z0

         MODULE PROCEDURE exflib_float_sub_x2i2
         MODULE PROCEDURE exflib_float_sub_x2q2
         MODULE PROCEDURE exflib_float_sub_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_sub_i0z2
         MODULE PROCEDURE exflib_float_sub_q0z2
         MODULE PROCEDURE exflib_float_sub_z0z2

         MODULE PROCEDURE exflib_float_sub_i2z0
         MODULE PROCEDURE exflib_float_sub_q2z0
         MODULE PROCEDURE exflib_float_sub_z2z0

         MODULE PROCEDURE exflib_float_sub_i2z2
         MODULE PROCEDURE exflib_float_sub_q2z2
         MODULE PROCEDURE exflib_float_sub_z2z2

         MODULE PROCEDURE exflib_float_sub_z0i2
         MODULE PROCEDURE exflib_float_sub_z0q2

         MODULE PROCEDURE exflib_float_sub_z2i0
         MODULE PROCEDURE exflib_float_sub_z2q0

         MODULE PROCEDURE exflib_float_sub_z2i2
         MODULE PROCEDURE exflib_float_sub_z2q2

         ! unarray operator
         MODULE PROCEDURE exflib_float_neg_x2
         MODULE PROCEDURE exflib_float_neg_z2
      END INTERFACE

      INTERFACE OPERATOR ( * )
         ! exfloat
         MODULE PROCEDURE exflib_float_mul_i0x2
         MODULE PROCEDURE exflib_float_mul_q0x2
         MODULE PROCEDURE exflib_float_mul_x0x2
         MODULE PROCEDURE exflib_float_mul_z0x2

         MODULE PROCEDURE exflib_float_mul_i2x0
         MODULE PROCEDURE exflib_float_mul_q2x0
         MODULE PROCEDURE exflib_float_mul_x2x0
         MODULE PROCEDURE exflib_float_mul_z2x0

         MODULE PROCEDURE exflib_float_mul_i2x2
         MODULE PROCEDURE exflib_float_mul_q2x2
         MODULE PROCEDURE exflib_float_mul_x2x2
         MODULE PROCEDURE exflib_float_mul_z2x2

         MODULE PROCEDURE exflib_float_mul_x0i2
         MODULE PROCEDURE exflib_float_mul_x0q2
         MODULE PROCEDURE exflib_float_mul_x0z2

         MODULE PROCEDURE exflib_float_mul_x2i0
         MODULE PROCEDURE exflib_float_mul_x2q0
         MODULE PROCEDURE exflib_float_mul_x2z0

         MODULE PROCEDURE exflib_float_mul_x2i2
         MODULE PROCEDURE exflib_float_mul_x2q2
         MODULE PROCEDURE exflib_float_mul_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_mul_i0z2
         MODULE PROCEDURE exflib_float_mul_q0z2
         MODULE PROCEDURE exflib_float_mul_z0z2

         MODULE PROCEDURE exflib_float_mul_i2z0
         MODULE PROCEDURE exflib_float_mul_q2z0
         MODULE PROCEDURE exflib_float_mul_z2z0

         MODULE PROCEDURE exflib_float_mul_i2z2
         MODULE PROCEDURE exflib_float_mul_q2z2
         MODULE PROCEDURE exflib_float_mul_z2z2

         MODULE PROCEDURE exflib_float_mul_z0i2
         MODULE PROCEDURE exflib_float_mul_z0q2

         MODULE PROCEDURE exflib_float_mul_z2i0
         MODULE PROCEDURE exflib_float_mul_z2q0

         MODULE PROCEDURE exflib_float_mul_z2i2
         MODULE PROCEDURE exflib_float_mul_z2q2
      END INTERFACE

      INTERFACE OPERATOR ( / )
         ! exfloat
         MODULE PROCEDURE exflib_float_div_i0x2
         MODULE PROCEDURE exflib_float_div_q0x2
         MODULE PROCEDURE exflib_float_div_x0x2
         MODULE PROCEDURE exflib_float_div_z0x2

         MODULE PROCEDURE exflib_float_div_i2x0
         MODULE PROCEDURE exflib_float_div_q2x0
         MODULE PROCEDURE exflib_float_div_x2x0
         MODULE PROCEDURE exflib_float_div_z2x0

         MODULE PROCEDURE exflib_float_div_i2x2
         MODULE PROCEDURE exflib_float_div_q2x2
         MODULE PROCEDURE exflib_float_div_x2x2
         MODULE PROCEDURE exflib_float_div_z2x2

         MODULE PROCEDURE exflib_float_div_x0i2
         MODULE PROCEDURE exflib_float_div_x0q2
         MODULE PROCEDURE exflib_float_div_x0z2

         MODULE PROCEDURE exflib_float_div_x2i0
         MODULE PROCEDURE exflib_float_div_x2q0
         MODULE PROCEDURE exflib_float_div_x2z0

         MODULE PROCEDURE exflib_float_div_x2i2
         MODULE PROCEDURE exflib_float_div_x2q2
         MODULE PROCEDURE exflib_float_div_x2z2

         ! ezfloat
         MODULE PROCEDURE exflib_float_div_i0z2
         MODULE PROCEDURE exflib_float_div_q0z2
         MODULE PROCEDURE exflib_float_div_z0z2

         MODULE PROCEDURE exflib_float_div_i2z0
         MODULE PROCEDURE exflib_float_div_q2z0
         MODULE PROCEDURE exflib_float_div_z2z0

         MODULE PROCEDURE exflib_float_div_i2z2
         MODULE PROCEDURE exflib_float_div_q2z2
         MODULE PROCEDURE exflib_float_div_z2z2

         MODULE PROCEDURE exflib_float_div_z0i2
         MODULE PROCEDURE exflib_float_div_z0q2

         MODULE PROCEDURE exflib_float_div_z2i0
         MODULE PROCEDURE exflib_float_div_z2q0

         MODULE PROCEDURE exflib_float_div_z2i2
         MODULE PROCEDURE exflib_float_div_z2q2
      END INTERFACE

      INTERFACE POW
         MODULE PROCEDURE exflib_float_pow_x2i0
         MODULE PROCEDURE exflib_float_pow_x2q0
         MODULE PROCEDURE exflib_float_pow_z2i0
         MODULE PROCEDURE exflib_float_pow_z2q0
      END INTERFACE

      INTERFACE OPERATOR ( ** )
         MODULE PROCEDURE exflib_float_pow_x2i0
         MODULE PROCEDURE exflib_float_pow_x2q0
         MODULE PROCEDURE exflib_float_pow_z2i0
         MODULE PROCEDURE exflib_float_pow_z2q0
      END INTERFACE

    CONTAINS

      !--------------------------------------------------
      ! assignemnt
      !--------------------------------------------------
      ! exfloat(:) = int
      SUBROUTINE exflib_float_assign_x1i0(a, b)
        TYPE(exfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*4, INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b
        END DO
      END SUBROUTINE exflib_float_assign_x1i0

      ! exfloat(:) = quadword
      SUBROUTINE exflib_float_assign_x1q0(a, b)
        TYPE(exfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*8, INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b
        END DO
      END SUBROUTINE exflib_float_assign_x1q0

      ! exfloat(:) = exfloat
      SUBROUTINE exflib_float_assign_x1x0(a, b)
        TYPE(exfloat), DIMENSION(:), INTENT(OUT) :: a
        TYPE(exfloat), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b
        END DO
      END SUBROUTINE exflib_float_assign_x1x0

      ! exfloat(:) = int(:)
      SUBROUTINE exflib_float_assign_x1i1(a, b)
        TYPE(exfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b(i)
        END DO
      END SUBROUTINE exflib_float_assign_x1i1

      ! exfloat(:) = quadword(:)
      SUBROUTINE exflib_float_assign_x1q1(a, b)
        TYPE(exfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b(i)
        END DO
      END SUBROUTINE exflib_float_assign_x1q1

      ! ezfloat(:) = int
      SUBROUTINE exflib_float_assign_z1i0(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*4, INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b )
        END DO
      END SUBROUTINE exflib_float_assign_z1i0

      ! ezfloat(:) = quadword
      SUBROUTINE exflib_float_assign_z1q0(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*8, INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b )
        END DO
      END SUBROUTINE exflib_float_assign_z1q0

      ! ezfloat(:) = exfloat
      SUBROUTINE exflib_float_assign_z1x0(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        TYPE(exfloat), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b )
        END DO
      END SUBROUTINE exflib_float_assign_z1x0

      ! ezfloat(:) = ezfloat
      SUBROUTINE exflib_float_assign_z1z0(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = b
        END DO
      END SUBROUTINE exflib_float_assign_z1z0

      ! ezfloat(:) = int(:)
      SUBROUTINE exflib_float_assign_z1i1(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b(i) )
        END DO
      END SUBROUTINE exflib_float_assign_z1i1

      ! ezfloat(:) = quadword(:)
      SUBROUTINE exflib_float_assign_z1q1(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b(i) )
        END DO
      END SUBROUTINE exflib_float_assign_z1q1

      ! ezfloat(:) = exfloat(:)
      SUBROUTINE exflib_float_assign_z1x1(a, b)
        TYPE(ezfloat), DIMENSION(:), INTENT(OUT) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           a(i) = CMPLX( b(i) )
        END DO
      END SUBROUTINE exflib_float_assign_z1x1

      !--------------------------------------------------
      ! type conversion
      !--------------------------------------------------
      ! exfloat(:) to double(:)
      FUNCTION exflib_float_cast_x1d1(a) RESULT(b)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(SIZE(a)) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           b(i) = DBLE( a(i) )
        END DO
      END FUNCTION exflib_float_cast_x1d1

      ! ezfloat(:) to dcomplex(:)
      FUNCTION exflib_float_cast_z1w1(a) RESULT(b)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(SIZE(a)) :: b
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           b(i) = CMPLX( a(i) )
        END DO
      END FUNCTION exflib_float_cast_z1w1

      !--------------------------------------------------
      ! relation, == (equal)
      !--------------------------------------------------
      ! int == exfloat(:)
      FUNCTION exflib_float_eq_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_i0x1

      ! quadword == exfloat(:)
      FUNCTION exflib_float_eq_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_q0x1

      ! single == exfloat(:)
      FUNCTION exflib_float_eq_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_s0x1

      ! double == exfloat(:)
      FUNCTION exflib_float_eq_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_d0x1

      ! exfloat == exfloat(:)
      FUNCTION exflib_float_eq_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0x1

      ! scomplex == exfloat(:)
      FUNCTION exflib_float_eq_c0x1(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
            mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_c0x1

      ! dcomplex == exfloat(:)
      FUNCTION exflib_float_eq_w0x1(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_w0x1

      ! ezfloat == exfloat(:)
      FUNCTION exflib_float_eq_z0x1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0x1

      ! int(:) == exfloat
      FUNCTION exflib_float_eq_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_i1x0

      ! quadword(:) == exfloat
      FUNCTION exflib_float_eq_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_q1x0

      ! single(:) == exfloat
      FUNCTION exflib_float_eq_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_s1x0

      ! double(:) == exfloat
      FUNCTION exflib_float_eq_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_d1x0

      ! exfloat(:) == exfloat
      FUNCTION exflib_float_eq_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1x0

      ! scomplex(:) == exfloat
      FUNCTION exflib_float_eq_c1x0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_c1x0

      ! dcomplex(:) == exfloat
      FUNCTION exflib_float_eq_w1x0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_w1x0

      ! ezfloat(:) == exfloat
      FUNCTION exflib_float_eq_z1x0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1x0

      ! int(:) == exfloat(:)
      FUNCTION exflib_float_eq_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_i1x1

      ! quadword(:) == exfloat(:)
      FUNCTION exflib_float_eq_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_q1x1

      ! single(:) == exfloat(:)
      FUNCTION exflib_float_eq_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_s1x1

      ! double(:) == exfloat(:)
      FUNCTION exflib_float_eq_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_d1x1

      ! exfloat(:) == exfloat(:)
      FUNCTION exflib_float_eq_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1x1

      ! scomplex(:) == exfloat(:)
      FUNCTION exflib_float_eq_c1x1(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_c1x1

      ! dcomplex(:) == exfloat(:)
      FUNCTION exflib_float_eq_w1x1(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_w1x1

      ! ezfloat(:) == exfloat(:)
      FUNCTION exflib_float_eq_z1x1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1x1

      ! exfloat(:) == int
      FUNCTION exflib_float_eq_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1i0

      ! exfloat(:) == quadword
      FUNCTION exflib_float_eq_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1q0

      ! exfloat(:) == single
      FUNCTION exflib_float_eq_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1s0

      ! exfloat(:) == double
      FUNCTION exflib_float_eq_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1d0

      ! exfloat(:) == scomplex
      FUNCTION exflib_float_eq_x1c0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1c0

      ! exfloat(:) == dcomplex
      FUNCTION exflib_float_eq_x1w0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1w0

      ! exfloat(:) == ezfloat
      FUNCTION exflib_float_eq_x1z0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_x1z0

      ! exfloat == int(:)
      FUNCTION exflib_float_eq_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0i1

      ! exfloat == quadword(:)
      FUNCTION exflib_float_eq_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0q1

      ! exfloat == single(:)
      FUNCTION exflib_float_eq_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0s1

      ! exfloat == double(:)
      FUNCTION exflib_float_eq_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0d1

      ! exfloat == scomplex(:)
      FUNCTION exflib_float_eq_x0c1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0c1

      ! exfloat == dcomplex(:)
      FUNCTION exflib_float_eq_x0w1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0w1

      ! exfloat == ezfloat(:)
      FUNCTION exflib_float_eq_x0z1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x0z1

      ! exfloat(:) == int(:)
      FUNCTION exflib_float_eq_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1i1

      ! exfloat(:) == quadword(:)
      FUNCTION exflib_float_eq_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1q1

      ! exfloat(:) == single(:)
      FUNCTION exflib_float_eq_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1s1

      ! exfloat(:) == double(:)
      FUNCTION exflib_float_eq_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1d1

      ! exfloat(:) == scomplex(:)
      FUNCTION exflib_float_eq_x1c1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1c1

      ! exfloat(:) == dcomplex(:)
      FUNCTION exflib_float_eq_x1w1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1w1

      ! exfloat(:) == ezfloat(:)
      FUNCTION exflib_float_eq_x1z1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_x1z1

      ! int == ezfloat(:)
      FUNCTION exflib_float_eq_i0z1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_i0z1

      ! quadword == ezfloat(:)
      FUNCTION exflib_float_eq_q0z1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_q0z1

      ! single == ezfloat(:)
      FUNCTION exflib_float_eq_s0z1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_s0z1

      ! double == ezfloat(:)
      FUNCTION exflib_float_eq_d0z1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_d0z1

      ! scomplex == ezfloat(:)
      FUNCTION exflib_float_eq_c0z1(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_c0z1

      ! dcomplex == ezfloat(:)
      FUNCTION exflib_float_eq_w0z1(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_w0z1

      ! ezfloat == ezfloat(:)
      FUNCTION exflib_float_eq_z0z1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0z1

      ! int(:) == ezfloat
      FUNCTION exflib_float_eq_i1z0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_i1z0

      ! quadword(:) == ezfloat
      FUNCTION exflib_float_eq_q1z0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_q1z0

      ! single(:) == ezfloat
      FUNCTION exflib_float_eq_s1z0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_s1z0

      ! double(:) == ezfloat
      FUNCTION exflib_float_eq_d1z0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_d1z0

      ! scomplex(:) == ezfloat
      FUNCTION exflib_float_eq_c1z0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_c1z0

      ! dcomplex(:) == ezfloat
      FUNCTION exflib_float_eq_w1z0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_w1z0

      ! ezfloat(:) == ezfloat
      FUNCTION exflib_float_eq_z1z0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1z0

      ! int(:) == ezfloat(:)
      FUNCTION exflib_float_eq_i1z1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_i1z1

      ! quadword(:) == ezfloat(:)
      FUNCTION exflib_float_eq_q1z1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_q1z1

      ! single(:) == ezfloat(:)
      FUNCTION exflib_float_eq_s1z1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_s1z1

      ! double(:) == ezfloat(:)
      FUNCTION exflib_float_eq_d1z1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_d1z1

      ! scomplex(:) == ezfloat(:)
      FUNCTION exflib_float_eq_c1z1(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_c1z1

      ! dcomplex(:) == ezfloat(:)
      FUNCTION exflib_float_eq_w1z1(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_w1z1

      ! ezfloat(:) == ezfloat(:)
      FUNCTION exflib_float_eq_z1z1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1z1

      ! ezfloat(:) == int
      FUNCTION exflib_float_eq_z1i0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1i0

      ! ezfloat(:) == quadword
      FUNCTION exflib_float_eq_z1q0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1q0

      ! ezfloat(:) == single
      FUNCTION exflib_float_eq_z1s0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1s0

      ! ezfloat(:) == double
      FUNCTION exflib_float_eq_z1d0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1d0

      ! ezfloat(:) == scomplex
      FUNCTION exflib_float_eq_z1c0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1c0

      ! ezfloat(:) == dcomplex
      FUNCTION exflib_float_eq_z1w0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) == b )
        END DO
      END FUNCTION exflib_float_eq_z1w0

      ! ezfloat == int(:)
      FUNCTION exflib_float_eq_z0i1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0i1

      ! ezfloat == quadword(:)
      FUNCTION exflib_float_eq_z0q1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0q1

      ! ezfloat == single(:)
      FUNCTION exflib_float_eq_z0s1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0s1

      ! ezfloat == double(:)
      FUNCTION exflib_float_eq_z0d1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0d1

      ! ezfloat == scomplex(:)
      FUNCTION exflib_float_eq_z0c1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0c1

      ! ezfloat == dcomplex(:)
      FUNCTION exflib_float_eq_z0w1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z0w1

      ! ezfloat(:) == int(:)
      FUNCTION exflib_float_eq_z1i1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1i1

      ! ezfloat(:) == quadword(:)
      FUNCTION exflib_float_eq_z1q1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1q1

      ! ezfloat(:) == single(:)
      FUNCTION exflib_float_eq_z1s1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1s1

      ! ezfloat(:) == double(:)
      FUNCTION exflib_float_eq_z1d1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1d1

      ! ezfloat(:) == scomplex(:)
      FUNCTION exflib_float_eq_z1c1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1c1

      ! ezfloat(:) == dcomplex(:)
      FUNCTION exflib_float_eq_z1w1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) == b(i) )
        END DO
      END FUNCTION exflib_float_eq_z1w1

      !----------------------------------------
      ! relation, /= (not equal)
      !----------------------------------------
      ! int /= exfloat(:)
      FUNCTION exflib_float_ne_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_i0x1

      ! quadword /= exfloat(:)
      FUNCTION exflib_float_ne_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_q0x1

      ! single /= exfloat(:)
      FUNCTION exflib_float_ne_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_s0x1

      ! double /= exfloat(:)
      FUNCTION exflib_float_ne_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_d0x1

      ! exfloat /= exfloat(:)
      FUNCTION exflib_float_ne_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0x1

      ! scomplex /= exfloat(:)
      FUNCTION exflib_float_ne_c0x1(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_c0x1

      ! dcomplex /= exfloat(:)
      FUNCTION exflib_float_ne_w0x1(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_w0x1

      ! ezfloat /= exfloat(:)
      FUNCTION exflib_float_ne_z0x1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0x1

      ! int(:) /= exfloat
      FUNCTION exflib_float_ne_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_i1x0

      ! quadword(:) /= exfloat
      FUNCTION exflib_float_ne_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_q1x0

      ! single(:) /= exfloat
      FUNCTION exflib_float_ne_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_s1x0

      ! double(:) /= exfloat
      FUNCTION exflib_float_ne_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_d1x0

      ! exfloat(:) /= exfloat
      FUNCTION exflib_float_ne_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1x0

      ! scomplex(:) /= exfloat
      FUNCTION exflib_float_ne_c1x0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_c1x0

      ! dcomplex(:) /= exfloat
      FUNCTION exflib_float_ne_w1x0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_w1x0

      ! ezfloat(:) /= exfloat
      FUNCTION exflib_float_ne_z1x0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1x0

      ! int(:) /= exfloat(:)
      FUNCTION exflib_float_ne_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_i1x1

      ! quadword(:) /= exfloat(:)
      FUNCTION exflib_float_ne_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_q1x1

      ! single(:) /= exfloat(:)
      FUNCTION exflib_float_ne_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_s1x1

      ! double(:) /= exfloat(:)
      FUNCTION exflib_float_ne_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_d1x1

      ! exfloat(:) /= exfloat(:)
      FUNCTION exflib_float_ne_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1x1

      ! scomplex(:) /= exfloat(:)
      FUNCTION exflib_float_ne_c1x1(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_c1x1

      ! dcomplex(:) /= exfloat(:)
      FUNCTION exflib_float_ne_w1x1(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_w1x1

      ! ezfloat(:) /= exfloat(:)
      FUNCTION exflib_float_ne_z1x1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1x1

      ! exfloat(:) /= int
      FUNCTION exflib_float_ne_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1i0

      ! exfloat(:) /= quadword
      FUNCTION exflib_float_ne_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1q0

      ! exfloat(:) /= single
      FUNCTION exflib_float_ne_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1s0

      ! exfloat(:) /= double
      FUNCTION exflib_float_ne_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1d0

      ! exfloat(:) /= scomplex
      FUNCTION exflib_float_ne_x1c0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1c0

      ! exfloat(:) /= dcomplex
      FUNCTION exflib_float_ne_x1w0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1w0

      ! exfloat(:) /= ezfloat
      FUNCTION exflib_float_ne_x1z0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_x1z0

      ! exfloat /= int(:)
      FUNCTION exflib_float_ne_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0i1

      ! exfloat /= quadword(:)
      FUNCTION exflib_float_ne_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0q1

      ! exfloat /= single(:)
      FUNCTION exflib_float_ne_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0s1

      ! exfloat /= double(:)
      FUNCTION exflib_float_ne_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0d1

      ! exfloat /= scomplex(:)
      FUNCTION exflib_float_ne_x0c1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0c1

      ! exfloat /= dcomplex(:)
      FUNCTION exflib_float_ne_x0w1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0w1

      ! exfloat /= ezfloat(:)
      FUNCTION exflib_float_ne_x0z1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x0z1

      ! exfloat(:) /= int(:)
      FUNCTION exflib_float_ne_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1i1

      ! exfloat(:) /= quadword(:)
      FUNCTION exflib_float_ne_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1q1

      ! exfloat(:) /= single(:)
      FUNCTION exflib_float_ne_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1s1

      ! exfloat(:) /= double(:)
      FUNCTION exflib_float_ne_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1d1

      ! exfloat(:) /= scomplex(:)
      FUNCTION exflib_float_ne_x1c1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1c1

      ! exfloat(:) /= dcomplex(:)
      FUNCTION exflib_float_ne_x1w1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1w1

      ! exfloat(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_x1z1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_x1z1

      ! int /= ezfloat(:)
      FUNCTION exflib_float_ne_i0z1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_i0z1

      ! quadword /= ezfloat(:)
      FUNCTION exflib_float_ne_q0z1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_q0z1

      ! single /= ezfloat(:)
      FUNCTION exflib_float_ne_s0z1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_s0z1

      ! double /= ezfloat(:)
      FUNCTION exflib_float_ne_d0z1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_d0z1

      ! scomplex /= ezfloat(:)
      FUNCTION exflib_float_ne_c0z1(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_c0z1

      ! dcomplex /= ezfloat(:)
      FUNCTION exflib_float_ne_w0z1(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_w0z1

      ! ezfloat /= ezfloat(:)
      FUNCTION exflib_float_ne_z0z1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0z1

      ! int(:) /= ezfloat
      FUNCTION exflib_float_ne_i1z0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_i1z0

      ! quadword(:) /= ezfloat
      FUNCTION exflib_float_ne_q1z0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_q1z0

      ! single(:) /= ezfloat
      FUNCTION exflib_float_ne_s1z0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_s1z0

      ! double(:) /= ezfloat
      FUNCTION exflib_float_ne_d1z0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_d1z0

      ! scomplex(:) /= ezfloat
      FUNCTION exflib_float_ne_c1z0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_c1z0

      ! dcomplex(:) /= ezfloat
      FUNCTION exflib_float_ne_w1z0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_w1z0

      ! ezfloat(:) /= ezfloat
      FUNCTION exflib_float_ne_z1z0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1z0

      ! int(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_i1z1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_i1z1

      ! quadword(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_q1z1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_q1z1

      ! single(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_s1z1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_s1z1

      ! double(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_d1z1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_d1z1

      ! scomplex(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_c1z1(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_c1z1

      ! dcomplex(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_w1z1(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_w1z1

      ! ezfloat(:) /= ezfloat(:)
      FUNCTION exflib_float_ne_z1z1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1z1

      ! ezfloat(:) /= int
      FUNCTION exflib_float_ne_z1i0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1i0

      ! ezfloat(:) /= quadword
      FUNCTION exflib_float_ne_z1q0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1q0

      ! ezfloat(:) /= single
      FUNCTION exflib_float_ne_z1s0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1s0

      ! ezfloat(:) /= double
      FUNCTION exflib_float_ne_z1d0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1d0

      ! ezfloat(:) /= scomplex
      FUNCTION exflib_float_ne_z1c0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1c0

      ! ezfloat(:) /= dcomplex
      FUNCTION exflib_float_ne_z1w0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) /= b )
        END DO
      END FUNCTION exflib_float_ne_z1w0

      ! ezfloat /= int(:)
      FUNCTION exflib_float_ne_z0i1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0i1

      ! ezfloat /= quadword(:)
      FUNCTION exflib_float_ne_z0q1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0q1

      ! ezfloat /= single(:)
      FUNCTION exflib_float_ne_z0s1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0s1

      ! ezfloat /= double(:)
      FUNCTION exflib_float_ne_z0d1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0d1

      ! ezfloat /= scomplex(:)
      FUNCTION exflib_float_ne_z0c1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0c1

      ! ezfloat /= dcomplex(:)
      FUNCTION exflib_float_ne_z0w1(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z0w1

      ! ezfloat(:) /= int(:)
      FUNCTION exflib_float_ne_z1i1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1i1

      ! ezfloat(:) /= quadword(:)
      FUNCTION exflib_float_ne_z1q1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1q1

      ! ezfloat(:) /= single(:)
      FUNCTION exflib_float_ne_z1s1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1s1

      ! ezfloat(:) /= double(:)
      FUNCTION exflib_float_ne_z1d1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1d1

      ! ezfloat(:) /= scomplex(:)
      FUNCTION exflib_float_ne_z1c1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1c1

      ! ezfloat(:) /= dcomplex(:)
      FUNCTION exflib_float_ne_z1w1(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a(i) /= b(i) )
        END DO
      END FUNCTION exflib_float_ne_z1w1

      !----------------------------------------
      ! relation, < (less than)
      !----------------------------------------
      ! int < exfloat(:)
      FUNCTION exflib_float_lt_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_i0x1

      ! quadword < exfloat(:)
      FUNCTION exflib_float_lt_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_q0x1

      ! single < exfloat(:)
      FUNCTION exflib_float_lt_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_s0x1

      ! double < exfloat(:)
      FUNCTION exflib_float_lt_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_d0x1

      ! exfloat < exfloat(:)
      FUNCTION exflib_float_lt_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x0x1

      ! int(:) < exfloat
      FUNCTION exflib_float_lt_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_i1x0

      ! quadword(:) < exfloat
      FUNCTION exflib_float_lt_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_q1x0

      ! single(:) < exfloat
      FUNCTION exflib_float_lt_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_s1x0

      ! double(:) < exfloat
      FUNCTION exflib_float_lt_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_d1x0

      ! exfloat(:) < exfloat
      FUNCTION exflib_float_lt_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_x1x0

      ! int(:) < exfloat(:)
      FUNCTION exflib_float_lt_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_i1x1

      ! quadword(:) < exfloat(:)
      FUNCTION exflib_float_lt_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_q1x1

      ! single(:) < exfloat(:)
      FUNCTION exflib_float_lt_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_s1x1

      ! double(:) < exfloat(:)
      FUNCTION exflib_float_lt_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_d1x1

      ! exfloat(:) < exfloat(:)
      FUNCTION exflib_float_lt_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x1x1

      ! exfloat(:) < int
      FUNCTION exflib_float_lt_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_x1i0

      ! exfloat(:) < quadword
      FUNCTION exflib_float_lt_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_x1q0

      ! exfloat(:) < single
      FUNCTION exflib_float_lt_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_x1s0

      ! exfloat(:) < double
      FUNCTION exflib_float_lt_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b )
        END DO
      END FUNCTION exflib_float_lt_x1d0

      ! exfloat < int(:)
      FUNCTION exflib_float_lt_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x0i1

      ! exfloat < quadword(:)
      FUNCTION exflib_float_lt_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x0q1

      ! exfloat < single(:)
      FUNCTION exflib_float_lt_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x0s1

      ! exfloat < double(:)
      FUNCTION exflib_float_lt_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x0d1

      ! exfloat(:) < int(:)
      FUNCTION exflib_float_lt_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x1i1

      ! exfloat(:) < quadword(:)
      FUNCTION exflib_float_lt_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x1q1

      ! exfloat(:) < single(:)
      FUNCTION exflib_float_lt_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x1s1

      ! exfloat(:) < double(:)
      FUNCTION exflib_float_lt_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) < b(i) )
        END DO
      END FUNCTION exflib_float_lt_x1d1

      !----------------------------------------
      ! relation, <= (less than or equal)
      !----------------------------------------
      ! int <= exfloat(:)
      FUNCTION exflib_float_le_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_i0x1

      ! quadword <= exfloat(:)
      FUNCTION exflib_float_le_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_q0x1

      ! single <= exfloat(:)
      FUNCTION exflib_float_le_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_s0x1

      ! double <= exfloat(:)
      FUNCTION exflib_float_le_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_d0x1

      ! exfloat <= exfloat(:)
      FUNCTION exflib_float_le_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x0x1

      ! int(:) <= exfloat
      FUNCTION exflib_float_le_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_i1x0

      ! quadword(:) <= exfloat
      FUNCTION exflib_float_le_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_q1x0

      ! single(:) <= exfloat
      FUNCTION exflib_float_le_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_s1x0

      ! double(:) <= exfloat
      FUNCTION exflib_float_le_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_d1x0

      ! exfloat(:) <= exfloat
      FUNCTION exflib_float_le_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_x1x0

      ! int(:) <= exfloat(:)
      FUNCTION exflib_float_le_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_i1x1

      ! quadword(:) <= exfloat(:)
      FUNCTION exflib_float_le_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_q1x1

      ! single(:) <= exfloat(:)
      FUNCTION exflib_float_le_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_s1x1

      ! double(:) <= exfloat(:)
      FUNCTION exflib_float_le_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_d1x1

      ! exfloat(:) <= exfloat(:)
      FUNCTION exflib_float_le_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x1x1

      ! exfloat(:) <= int
      FUNCTION exflib_float_le_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_x1i0

      ! exfloat(:) <= quadword
      FUNCTION exflib_float_le_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_x1q0

      ! exfloat(:) <= single
      FUNCTION exflib_float_le_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_x1s0

      ! exfloat(:) <= double
      FUNCTION exflib_float_le_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b )
        END DO
      END FUNCTION exflib_float_le_x1d0

      ! exfloat <= int(:)
      FUNCTION exflib_float_le_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x0i1

      ! exfloat <= quadword(:)
      FUNCTION exflib_float_le_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x0q1

      ! exfloat <= single(:)
      FUNCTION exflib_float_le_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x0s1

      ! exfloat <= double(:)
      FUNCTION exflib_float_le_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x0d1

      ! exfloat(:) <= int(:)
      FUNCTION exflib_float_le_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x1i1

      ! exfloat(:) <= quadword(:)
      FUNCTION exflib_float_le_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x1q1

      ! exfloat(:) <= single(:)
      FUNCTION exflib_float_le_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x1s1

      ! exfloat(:) <= double(:)
      FUNCTION exflib_float_le_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) <= b(i) )
        END DO
      END FUNCTION exflib_float_le_x1d1

      !----------------------------------------
      ! relation, > (greater than)
      !----------------------------------------
      ! int > exfloat(:)
      FUNCTION exflib_float_gt_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_i0x1

      ! quadword > exfloat(:)
      FUNCTION exflib_float_gt_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_q0x1

      ! single > exfloat(:)
      FUNCTION exflib_float_gt_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_s0x1

      ! double > exfloat(:)
      FUNCTION exflib_float_gt_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_d0x1

      ! exfloat > exfloat(:)
      FUNCTION exflib_float_gt_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x0x1

      ! int(:) > exfloat
      FUNCTION exflib_float_gt_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_i1x0

      ! quadword(:) > exfloat
      FUNCTION exflib_float_gt_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_q1x0

      ! single(:) > exfloat
      FUNCTION exflib_float_gt_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_s1x0

      ! double(:) > exfloat
      FUNCTION exflib_float_gt_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_d1x0

      ! exfloat(:) > exfloat
      FUNCTION exflib_float_gt_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_x1x0

      ! int(:) > exfloat(:)
      FUNCTION exflib_float_gt_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_i1x1

      ! quadword(:) > exfloat(:)
      FUNCTION exflib_float_gt_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_q1x1

      ! single(:) > exfloat(:)
      FUNCTION exflib_float_gt_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_s1x1

      ! double(:) > exfloat(:)
      FUNCTION exflib_float_gt_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_d1x1

      ! exfloat(:) > exfloat(:)
      FUNCTION exflib_float_gt_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x1x1

      ! exfloat(:) > int
      FUNCTION exflib_float_gt_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_x1i0

      ! exfloat(:) > quadword
      FUNCTION exflib_float_gt_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_x1q0

      ! exfloat(:) > single
      FUNCTION exflib_float_gt_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_x1s0

      ! exfloat(:) > double
      FUNCTION exflib_float_gt_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b )
        END DO
      END FUNCTION exflib_float_gt_x1d0

      ! exfloat > int(:)
      FUNCTION exflib_float_gt_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x0i1

      ! exfloat > quadword(:)
      FUNCTION exflib_float_gt_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x0q1

      ! exfloat > single(:)
      FUNCTION exflib_float_gt_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x0s1

      ! exfloat > double(:)
      FUNCTION exflib_float_gt_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x0d1

      ! exfloat(:) > int(:)
      FUNCTION exflib_float_gt_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x1i1

      ! exfloat(:) > quadword(:)
      FUNCTION exflib_float_gt_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x1q1

      ! exfloat(:) > single(:)
      FUNCTION exflib_float_gt_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x1s1

      ! exfloat(:) > double(:)
      FUNCTION exflib_float_gt_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) > b(i) )
        END DO
      END FUNCTION exflib_float_gt_x1d1

      !----------------------------------------
      ! relation, >= (greater than or equal)
      !----------------------------------------
      ! int >= exfloat(:)
      FUNCTION exflib_float_ge_i0x1(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_i0x1

      ! quadword >= exfloat(:)
      FUNCTION exflib_float_ge_q0x1(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_q0x1

      ! single >= exfloat(:)
      FUNCTION exflib_float_ge_s0x1(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_s0x1

      ! double >= exfloat(:)
      FUNCTION exflib_float_ge_d0x1(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_d0x1

      ! exfloat >= exfloat(:)
      FUNCTION exflib_float_ge_x0x1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x0x1

      ! int(:) >= exfloat
      FUNCTION exflib_float_ge_i1x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_i1x0

      ! quadword(:) >= exfloat
      FUNCTION exflib_float_ge_q1x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_q1x0

      ! single(:) >= exfloat
      FUNCTION exflib_float_ge_s1x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_s1x0

      ! double(:) >= exfloat
      FUNCTION exflib_float_ge_d1x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_d1x0

      ! exfloat(:) >= exfloat
      FUNCTION exflib_float_ge_x1x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_x1x0

      ! int(:) >= exfloat(:)
      FUNCTION exflib_float_ge_i1x1(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_i1x1

      ! quadword(:) >= exfloat(:)
      FUNCTION exflib_float_ge_q1x1(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_q1x1

      ! single(:) >= exfloat(:)
      FUNCTION exflib_float_ge_s1x1(a, b) RESULT(mask)
        REAL*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_s1x1

      ! double(:) >= exfloat(:)
      FUNCTION exflib_float_ge_d1x1(a, b) RESULT(mask)
        REAL*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_d1x1

      ! exfloat(:) >= exfloat(:)
      FUNCTION exflib_float_ge_x1x1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x1x1

      ! exfloat(:) >= int
      FUNCTION exflib_float_ge_x1i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_x1i0

      ! exfloat(:) >= quadword
      FUNCTION exflib_float_ge_x1q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_x1q0

      ! exfloat(:) >= single
      FUNCTION exflib_float_ge_x1s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_x1s0

      ! exfloat(:) >= double
      FUNCTION exflib_float_ge_x1d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b )
        END DO
      END FUNCTION exflib_float_ge_x1d0

      ! exfloat >= int(:)
      FUNCTION exflib_float_ge_x0i1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x0i1

      ! exfloat >= quadword(:)
      FUNCTION exflib_float_ge_x0q1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x0q1

      ! exfloat >= single(:)
      FUNCTION exflib_float_ge_x0s1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x0s1

      ! exfloat >= double(:)
      FUNCTION exflib_float_ge_x0d1(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           mask(i) = ( a >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x0d1

      ! exfloat(:) >= int(:)
      FUNCTION exflib_float_ge_x1i1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x1i1

      ! exfloat(:) >= quadword(:)
      FUNCTION exflib_float_ge_x1q1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x1q1

      ! exfloat(:) >= single(:)
      FUNCTION exflib_float_ge_x1s1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*4, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x1s1

      ! exfloat(:) >= double(:)
      FUNCTION exflib_float_ge_x1d1(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        REAL*8, DIMENSION(:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a)) :: mask
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           mask(i) = ( a(i) >= b(i) )
        END DO
      END FUNCTION exflib_float_ge_x1d1

      !--------------------------------------------------
      ! one dimensional array functions
      !--------------------------------------------------
      ! sum of exfloat(:)
      FUNCTION exflib_float_sum_x1(a, mask) RESULT(s)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: s
        INTEGER*8 :: i

        s = 0
        IF( .NOT.PRESENT(mask) ) THEN

           DO i = 1, SIZE(a)
              s = s + a(i)
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( mask(i) ) THEN
                 s = s + a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_sum_x1

      ! sum of ezfloat(:)
      FUNCTION exflib_float_sum_z1(a, mask) RESULT(s)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat) :: s
        INTEGER*8 :: i

        s = 0
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              s = s + a(i)
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF( mask(i) ) THEN
                 s = s + a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_sum_z1

      ! product of exfloat(:)
      FUNCTION exflib_float_product_x1(a, mask) RESULT(p)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: p
        INTEGER*8 :: i

        p = 1
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              p = p * a(i)
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( mask(i) ) THEN
                 p = p * a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_product_x1

      ! product of ezfloat(:)
      FUNCTION exflib_float_product_z1(a, mask) RESULT(p)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat) :: p
        INTEGER*8 :: i

        p = 1
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              p = p * a(i)
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( mask(i) ) THEN
                 p = p * a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_product_z1

      !----------------------------------------
      ! one dimensional max/min functions
      !----------------------------------------
      ! find maximum in exfloat(:)
      FUNCTION exflib_float_maxval_x1(a, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: m
        INTEGER*8 :: i

        m = -exflib_float_huge_m(a(1))

        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              IF ( a(i) > m ) THEN
                 m = a(i)
              END IF
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( (mask(i)).AND.(a(i) > m) ) THEN
                 m = a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_maxval_x1

      ! find minimum in exfloat(:)
      FUNCTION exflib_float_minval_x1(a, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: m
        INTEGER*8 :: i

        m = exflib_float_huge_m(a(1))

        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              IF ( a(i) < m ) THEN
                 m = a(i)
              END IF
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( (mask(i)).AND.(a(i) < m) ) THEN
                 m = a(i)
              END IF
           END DO

        END IF

      END FUNCTION exflib_float_minval_x1

      ! find location of maximum in exfloat(:)
      FUNCTION exflib_float_maxloc_x1(a, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(1) :: idx
        TYPE(exfloat) :: m
        INTEGER*8 :: i

        idx(1) = EXFLIB_ARRAY_BEGIN
        m = -exflib_float_huge_m(a(1))

        IF( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              IF ( a(i) > m ) THEN
                 m = a(i)
                 idx(1) = i
              END IF
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( (mask(i)).AND.(a(i) > m) ) THEN
                 m = a(i)
                 idx(1) = i
              END IF
           END DO
        END IF

      END FUNCTION exflib_float_maxloc_x1

      ! find location of minimum in exfloat(:)
      FUNCTION exflib_float_minloc_x1(a, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(1) :: idx
        TYPE(exfloat) :: m
        INTEGER*8 :: i

        idx(1) = EXFLIB_ARRAY_BEGIN
        m = exflib_float_huge_m(a(1))

        IF ( .NOT.PRESENT( mask ) ) THEN

           DO i = 1, SIZE(a)
              IF ( a(i) < m ) THEN
                 m = a(i)
                 idx(1) = i
              END IF
           END DO

        ELSE

           DO i = 1, SIZE(a)
              IF ( (mask(i)).AND.(a(i) < m) ) THEN
                 m = a(i)
                 idx(1) = i
              END IF
           END DO
        END IF

      END FUNCTION exflib_float_minloc_x1

      !--------------------------------------------------
      ! one dimensional other operations
      !--------------------------------------------------
      ! inner-product of exfloat(:) and exfloat(:)
      FUNCTION exflib_float_dot_product_x1x1(a, b) RESULT(s)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a, b
        TYPE(exfloat) :: s
        INTEGER*8 :: i
        s = 0
        DO i = 1, SIZE(a)
           s = s + a(i)*b(i)
        END DO
      END FUNCTION exflib_float_dot_product_x1x1

      ! inner-product of exfloat(:) and ezfloat(:)
      FUNCTION exflib_float_dot_product_x1z1(a, b) RESULT(s)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat) :: s
        INTEGER*8 :: i
        s = 0
        DO i = 1, SIZE(a)
           s = s + a(i)*b(i)
        END DO
      END FUNCTION exflib_float_dot_product_x1z1

      ! inner-product of ezfloat(:) and exfloat(:)
      FUNCTION exflib_float_dot_product_z1x1(a, b) RESULT(s)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat) :: s
        INTEGER*8 :: i
        s = 0
        DO i = 1, SIZE(a)
           s = s + a(i)*b(i)
        END DO
      END FUNCTION exflib_float_dot_product_z1x1

      ! inner-product of ezfloat(:) and ezfloat(:)
      FUNCTION exflib_float_dot_product_z1z1(a, b) RESULT(s)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a, b
        TYPE(ezfloat) :: s
        INTEGER*8 :: i
        s = 0
        DO i = 1, SIZE(a)
           s = s + a(i)*b(i)
        END DO
      END FUNCTION exflib_float_dot_product_z1z1

      !----------------------------------------
      ! array addition
      !----------------------------------------
      ! int + exfloat(:)
      FUNCTION exflib_float_add_i0x1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_i0x1

      ! quadword + exfloat(:)
      FUNCTION exflib_float_add_q0x1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_q0x1

      ! exfloat + exfloat(:)
      FUNCTION exflib_float_add_x0x1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_x0x1

      ! ezfloat + exfloat(:)
      FUNCTION exflib_float_add_z0x1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_z0x1

      ! int(:) + exfloat
      FUNCTION exflib_float_add_i1x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_i1x0

      ! quadword(:) + exfloat
      FUNCTION exflib_float_add_q1x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_q1x0

      ! exfloat(:) + exfloat
      FUNCTION exflib_float_add_x1x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_x1x0

      ! ezfloat(:) + exfloat
      FUNCTION exflib_float_add_z1x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_z1x0

      ! int(:) + exfloat(:)
      FUNCTION exflib_float_add_i1x1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_i1x1

      ! quadword(:) + exfloat(:)
      FUNCTION exflib_float_add_q1x1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_q1x1

      ! exfloat(:) + exfloat(:)
      FUNCTION exflib_float_add_x1x1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_x1x1

      ! ezfloat(:) + exfloat(:)
      FUNCTION exflib_float_add_z1x1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_z1x1

      ! exfloat + int(:)
      FUNCTION exflib_float_add_x0i1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_x0i1

      ! exfloat + quadword(:)
      FUNCTION exflib_float_add_x0q1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_x0q1

      ! exfloat + ezfloat(:)
      FUNCTION exflib_float_add_x0z1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_x0z1

      ! exfloat(:) + int
      FUNCTION exflib_float_add_x1i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_x1i0

      ! exfloat(:) + quadword
      FUNCTION exflib_float_add_x1q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_x1q0

      ! exfloat(:) + ezfloat
      FUNCTION exflib_float_add_x1z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_x1z0

      ! exfloat(:) + int(:)
      FUNCTION exflib_float_add_x1i1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_x1i1

      ! exfloat(:) + quadword(:)
      FUNCTION exflib_float_add_x1q1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_x1q1

      ! exfloat(:) + ezfloat(:)
      FUNCTION exflib_float_add_x1z1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_x1z1

      ! int + ezfloat(:)
      FUNCTION exflib_float_add_i0z1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_i0z1

      ! quadword + ezfloat(:)
      FUNCTION exflib_float_add_q0z1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_q0z1

      ! ezfloat + ezfloat(:)
      FUNCTION exflib_float_add_z0z1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_z0z1

      ! int(:) + ezfloat
      FUNCTION exflib_float_add_i1z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_i1z0

      ! quadword(:) + ezfloat
      FUNCTION exflib_float_add_q1z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_q1z0

      ! ezfloat(:) + ezfloat
      FUNCTION exflib_float_add_z1z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_z1z0

      ! int(:) + ezfloat(:)
      FUNCTION exflib_float_add_i1z1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_i1z1

      ! quadword(:) + ezfloat(:)
      FUNCTION exflib_float_add_q1z1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_q1z1

      ! ezfloat(:) + ezfloat(:)
      FUNCTION exflib_float_add_z1z1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_z1z1

      ! ezfloat + int(:)
      FUNCTION exflib_float_add_z0i1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_z0i1

      ! ezfloat + quadword(:)
      FUNCTION exflib_float_add_z0q1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a + b(i)
        END DO
      END FUNCTION exflib_float_add_z0q1

      ! ezfloat(:) + int
      FUNCTION exflib_float_add_z1i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_z1i0

      ! ezfloat(:) + quadword
      FUNCTION exflib_float_add_z1q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b
        END DO
      END FUNCTION exflib_float_add_z1q0

      ! ezfloat(:) + int(:)
      FUNCTION exflib_float_add_z1i1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_z1i1

      ! ezfloat(:) + quadword(:)
      FUNCTION exflib_float_add_z1q1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) + b(i)
        END DO
      END FUNCTION exflib_float_add_z1q1

      ! unarray operator, +exfloat(:)
      FUNCTION exflib_float_pls_x1(a) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i)
        END DO
      END FUNCTION exflib_float_pls_x1

      ! unarray operator, +ezfloat(:)
      FUNCTION exflib_float_pls_z1(a) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i)
        END DO
      END FUNCTION exflib_float_pls_z1

      !----------------------------------------
      ! array subtraction
      !----------------------------------------
      ! int - exfloat(:)
      FUNCTION exflib_float_sub_i0x1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_i0x1

      ! quadword - exfloat(:)
      FUNCTION exflib_float_sub_q0x1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_q0x1

      ! exfloat - exfloat(:)
      FUNCTION exflib_float_sub_x0x1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_x0x1

      ! ezfloat - exfloat(:)
      FUNCTION exflib_float_sub_z0x1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_z0x1

      ! int(:) - exfloat
      FUNCTION exflib_float_sub_i1x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_i1x0

      ! quadword(:) - exfloat
      FUNCTION exflib_float_sub_q1x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_q1x0

      ! exfloat(:) - exfloat
      FUNCTION exflib_float_sub_x1x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_x1x0

      ! ezfloat(:) - exfloat
      FUNCTION exflib_float_sub_z1x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_z1x0

      ! int(:) - exfloat(:)
      FUNCTION exflib_float_sub_i1x1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_i1x1

      ! quadword(:) - exfloat(:)
      FUNCTION exflib_float_sub_q1x1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_q1x1

      ! exfloat(:) - exfloat(:)
      FUNCTION exflib_float_sub_x1x1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_x1x1

      ! ezfloat(:) - exfloat(:)
      FUNCTION exflib_float_sub_z1x1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_z1x1

      ! exfloat - int(:)
      FUNCTION exflib_float_sub_x0i1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_x0i1

      ! exfloat - quadword(:)
      FUNCTION exflib_float_sub_x0q1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_x0q1

      ! exfloat - ezfloat(:)
      FUNCTION exflib_float_sub_x0z1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_x0z1

      ! exfloat(:) - int
      FUNCTION exflib_float_sub_x1i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_x1i0

      ! exfloat(:) - quadword
      FUNCTION exflib_float_sub_x1q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_x1q0

      ! exfloat(:) - ezfloat
      FUNCTION exflib_float_sub_x1z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_x1z0

      ! exfloat(:) - int(:)
      FUNCTION exflib_float_sub_x1i1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_x1i1

      ! exfloat(:) - quadword(:)
      FUNCTION exflib_float_sub_x1q1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_x1q1

      ! exfloat(:) - ezfloat(:)
      FUNCTION exflib_float_sub_x1z1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_x1z1

      ! int - ezfloat(:)
      FUNCTION exflib_float_sub_i0z1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_i0z1

      ! quadword - ezfloat(:)
      FUNCTION exflib_float_sub_q0z1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_q0z1

      ! ezfloat - ezfloat(:)
      FUNCTION exflib_float_sub_z0z1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_z0z1

      ! int(:) - ezfloat
      FUNCTION exflib_float_sub_i1z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_i1z0

      ! quadword(:) - ezfloat
      FUNCTION exflib_float_sub_q1z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_q1z0

      ! ezfloat(:) - ezfloat
      FUNCTION exflib_float_sub_z1z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_z1z0

      ! int(:) - ezfloat(:)
      FUNCTION exflib_float_sub_i1z1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_i1z1

      ! quadword(:) - ezfloat(:)
      FUNCTION exflib_float_sub_q1z1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_q1z1

      ! ezfloat(:) - ezfloat(:)
      FUNCTION exflib_float_sub_z1z1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_z1z1

      ! ezfloat - int(:)
      FUNCTION exflib_float_sub_z0i1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_z0i1

      ! ezfloat - quadword(:)
      FUNCTION exflib_float_sub_z0q1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a - b(i)
        END DO
      END FUNCTION exflib_float_sub_z0q1

      ! ezfloat(:) - int
      FUNCTION exflib_float_sub_z1i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_z1i0

      ! ezfloat(:) - quadword
      FUNCTION exflib_float_sub_z1q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b
        END DO
      END FUNCTION exflib_float_sub_z1q0

      ! ezfloat(:) - int(:)
      FUNCTION exflib_float_sub_z1i1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_z1i1

      ! ezfloat(:) - quadword(:)
      FUNCTION exflib_float_sub_z1q1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) - b(i)
        END DO
      END FUNCTION exflib_float_sub_z1q1

      ! unarray operator, -exfloat(:)
      FUNCTION exflib_float_neg_x1(a) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = - a(i)
        END DO
      END FUNCTION exflib_float_neg_x1

      ! unarray operator, -ezfloat(:)
      FUNCTION exflib_float_neg_z1(a) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = - a(i)
        END DO
      END FUNCTION exflib_float_neg_z1

      !----------------------------------------
      ! array multiplication
      !----------------------------------------
      ! int * exfloat(:)
      FUNCTION exflib_float_mul_i0x1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_i0x1

      ! quadword * exfloat(:)
      FUNCTION exflib_float_mul_q0x1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_q0x1

      ! exfloat * exfloat(:)
      FUNCTION exflib_float_mul_x0x1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_x0x1

      ! ezfloat * exfloat(:)
      FUNCTION exflib_float_mul_z0x1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_z0x1

      ! int(:) * exfloat
      FUNCTION exflib_float_mul_i1x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_i1x0

      ! quadword(:) * exfloat
      FUNCTION exflib_float_mul_q1x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_q1x0

      ! exfloat(:) * exfloat
      FUNCTION exflib_float_mul_x1x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_x1x0

      ! ezfloat(:) * exfloat
      FUNCTION exflib_float_mul_z1x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_z1x0

      ! int(:) * exfloat(:)
      FUNCTION exflib_float_mul_i1x1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_i1x1

      ! quadword(:) * exfloat(:)
      FUNCTION exflib_float_mul_q1x1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_q1x1

      ! exfloat(:) * exfloat(:)
      FUNCTION exflib_float_mul_x1x1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_x1x1

      ! ezfloat(:) * exfloat(:)
      FUNCTION exflib_float_mul_z1x1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_z1x1

      ! exfloat * int(:)
      FUNCTION exflib_float_mul_x0i1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_x0i1

      ! exfloat * quadword(:)
      FUNCTION exflib_float_mul_x0q1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_x0q1

      ! exfloat * ezfloat(:)
      FUNCTION exflib_float_mul_x0z1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_x0z1

      ! exfloat(:) * int
      FUNCTION exflib_float_mul_x1i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_x1i0

      ! exfloat(:) * quadword
      FUNCTION exflib_float_mul_x1q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_x1q0

      ! exfloat(:) * ezfloat
      FUNCTION exflib_float_mul_x1z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_x1z0

      ! exfloat(:) * int(:)
      FUNCTION exflib_float_mul_x1i1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_x1i1

      ! exfloat(:) * quadword(:)
      FUNCTION exflib_float_mul_x1q1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_x1q1

      ! exfloat(:) * ezfloat(:)
      FUNCTION exflib_float_mul_x1z1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_x1z1

      ! int * ezfloat(:)
      FUNCTION exflib_float_mul_i0z1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_i0z1

      ! quadword * ezfloat(:)
      FUNCTION exflib_float_mul_q0z1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_q0z1

      ! ezfloat * ezfloat(:)
      FUNCTION exflib_float_mul_z0z1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_z0z1

      ! int(:) * ezfloat
      FUNCTION exflib_float_mul_i1z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_i1z0

      ! quadword(:) * ezfloat
      FUNCTION exflib_float_mul_q1z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_q1z0

      ! ezfloat(:) * ezfloat
      FUNCTION exflib_float_mul_z1z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_z1z0

      ! int(:) * ezfloat(:)
      FUNCTION exflib_float_mul_i1z1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_i1z1

      ! quadword(:) * ezfloat(:)
      FUNCTION exflib_float_mul_q1z1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_q1z1

      ! ezfloat(:) * ezfloat(:)
      FUNCTION exflib_float_mul_z1z1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_z1z1

      ! ezfloat * int(:)
      FUNCTION exflib_float_mul_z0i1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_z0i1

      ! ezfloat * quadword(:)
      FUNCTION exflib_float_mul_z0q1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a * b(i)
        END DO
      END FUNCTION exflib_float_mul_z0q1

      ! ezfloat(:) * int
      FUNCTION exflib_float_mul_z1i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_z1i0

      ! ezfloat(:) * quadword
      FUNCTION exflib_float_mul_z1q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b
        END DO
      END FUNCTION exflib_float_mul_z1q0

      ! ezfloat(:) * int(:)
      FUNCTION exflib_float_mul_z1i1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_z1i1

      ! ezfloat(:) * quadword(:)
      FUNCTION exflib_float_mul_z1q1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) * b(i)
        END DO
      END FUNCTION exflib_float_mul_z1q1

      !----------------------------------------
      ! array division
      !----------------------------------------
      ! int / exfloat(:)
      FUNCTION exflib_float_div_i0x1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_i0x1

      ! quadword / exfloat(:)
      FUNCTION exflib_float_div_q0x1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_q0x1

      ! exfloat / exfloat(:)
      FUNCTION exflib_float_div_x0x1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_x0x1

      ! ezfloat / exfloat(:)
      FUNCTION exflib_float_div_z0x1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_z0x1

      ! int(:) / exfloat
      FUNCTION exflib_float_div_i1x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_i1x0

      ! quadword(:) / exfloat
      FUNCTION exflib_float_div_q1x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_q1x0

      ! exfloat(:) / exfloat
      FUNCTION exflib_float_div_x1x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_x1x0

      ! ezfloat(:) / exfloat
      FUNCTION exflib_float_div_z1x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_z1x0

      ! int(:) / exfloat(:)
      FUNCTION exflib_float_div_i1x1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_i1x1

      ! quadword(:) / exfloat(:)
      FUNCTION exflib_float_div_q1x1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_q1x1

      ! exfloat(:) / exfloat(:)
      FUNCTION exflib_float_div_x1x1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_x1x1

      ! ezfloat(:) / exfloat(:)
      FUNCTION exflib_float_div_z1x1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_z1x1

      ! exfloat / int(:)
      FUNCTION exflib_float_div_x0i1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_x0i1

      ! exfloat / quadword(:)
      FUNCTION exflib_float_div_x0q1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_x0q1

      ! exfloat / ezfloat(:)
      FUNCTION exflib_float_div_x0z1(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_x0z1

      ! exfloat(:) / int
      FUNCTION exflib_float_div_x1i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_x1i0

      ! exfloat(:) / quadword
      FUNCTION exflib_float_div_x1q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_x1q0

      ! exfloat(:) / ezfloat
      FUNCTION exflib_float_div_x1z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_x1z0

      ! exfloat(:) / int(:)
      FUNCTION exflib_float_div_x1i1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_x1i1

      ! exfloat(:) / quadword(:)
      FUNCTION exflib_float_div_x1q1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_x1q1

      ! exfloat(:) / ezfloat(:)
      FUNCTION exflib_float_div_x1z1(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_x1z1

      ! int / ezfloat(:)
      FUNCTION exflib_float_div_i0z1(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_i0z1

      ! quadword / ezfloat(:)
      FUNCTION exflib_float_div_q0z1(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_q0z1

      ! ezfloat / ezfloat(:)
      FUNCTION exflib_float_div_z0z1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_z0z1

      ! int(:) / ezfloat
      FUNCTION exflib_float_div_i1z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_i1z0

      ! quadword(:) / ezfloat
      FUNCTION exflib_float_div_q1z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_q1z0

      ! ezfloat(:) / ezfloat
      FUNCTION exflib_float_div_z1z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_z1z0

      ! int(:) / ezfloat(:)
      FUNCTION exflib_float_div_i1z1(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_i1z1

      ! quadword(:) / ezfloat(:)
      FUNCTION exflib_float_div_q1z1(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_q1z1

      ! ezfloat(:) / ezfloat(:)
      FUNCTION exflib_float_div_z1z1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_z1z1

      ! ezfloat / int(:)
      FUNCTION exflib_float_div_z0i1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_z0i1

      ! ezfloat / quadword(:)
      FUNCTION exflib_float_div_z0q1(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(b)
           c(i) = a / b(i)
        END DO
      END FUNCTION exflib_float_div_z0q1

      ! ezfloat(:) / int
      FUNCTION exflib_float_div_z1i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_z1i0

      ! ezfloat(:) / quadword
      FUNCTION exflib_float_div_z1q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b
        END DO
      END FUNCTION exflib_float_div_z1q0

      ! ezfloat(:) / int(:)
      FUNCTION exflib_float_div_z1i1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_z1i1

      ! ezfloat(:) / quadword(:)
      FUNCTION exflib_float_div_z1q1(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: c
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           c(i) = a(i) / b(i)
        END DO
      END FUNCTION exflib_float_div_z1q1

      !----------------------------------------
      ! power
      !----------------------------------------
      ! pow( exfloat(:), int )
      FUNCTION exflib_float_pow_x1i0(a, n) RESULT(s)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: n
        TYPE(exfloat), DIMENSION(SIZE(a)) :: s
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           s(i) = a(i) ** n
        END DO
      END FUNCTION exflib_float_pow_x1i0

      ! pow( exfloat(:), quadword )
      FUNCTION exflib_float_pow_x1q0(a, n) RESULT(s)
        TYPE(exfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: n
        TYPE(exfloat), DIMENSION(SIZE(a)) :: s
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           s(i) = a(i) ** n
        END DO
      END FUNCTION exflib_float_pow_x1q0

      ! pow( ezfloat(:), int )
      FUNCTION exflib_float_pow_z1i0(a, n) RESULT(s)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: n
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: s
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           s(i) = a(i) ** n
        END DO
      END FUNCTION exflib_float_pow_z1i0

      ! pow( ezfloat(:), quadword )
      FUNCTION exflib_float_pow_z1q0(a, n) RESULT(s)
        TYPE(ezfloat), DIMENSION(:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: n
        TYPE(ezfloat), DIMENSION(SIZE(a)) :: s
        INTEGER*8 :: i
        DO i = 1, SIZE(a)
           s(i) = a(i) ** n
        END DO
      END FUNCTION exflib_float_pow_z1q0

      !--------------------------------------------------
      ! two dimensional array (matrix) functions
      !--------------------------------------------------
      ! assignemtn
      !--------------------------------------------------
      ! exfloat(:,:) = int
      SUBROUTINE exflib_float_assign_x2i0(a, b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*4, INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_x2i0

      ! exfloat(:,:) = quadword
      SUBROUTINE exflib_float_assign_x2q0(a, b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*8, INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_x2q0

      ! exfloat(:,:) = exfloat
      SUBROUTINE exflib_float_assign_x2x0(a, b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(OUT) :: a
        TYPE(exfloat), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_x2x0

      ! exfloat(:,:) = int(:,:)
      SUBROUTINE exflib_float_assign_x2i2(a, b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b(i,j)
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_x2i2

      ! exfloat(:,:) = quadword(:,:)
      SUBROUTINE exflib_float_assign_x2q2(a, b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b(i,j)
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_x2q2

      ! ezfloat(:,:) = int
      SUBROUTINE exflib_float_assign_z2i0(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*4, INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2i0

      ! ezfloat(:,:) = quadword
      SUBROUTINE exflib_float_assign_z2q0(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*8, INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2q0

      ! ezfloat(:,:) = exfloat
      SUBROUTINE exflib_float_assign_z2x0(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        TYPE(exfloat), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2x0

      ! ezfloat(:,:) = ezfloat
      SUBROUTINE exflib_float_assign_z2z0(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = b
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2z0

      ! ezfloat(:,:) = int(:,:)
      SUBROUTINE exflib_float_assign_z2i2(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b(i,j) )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2i2

      ! ezfloat(:,:) = quadword(:,:)
      SUBROUTINE exflib_float_assign_z2q2(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b(i,j) )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2q2

      ! ezfloat(:,:) = exfloat(:,:)
      SUBROUTINE exflib_float_assign_z2x2(a, b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(OUT) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              a(i,j) = CMPLX( b(i,j) )
           END DO
        END DO
      END SUBROUTINE exflib_float_assign_z2x2

      !--------------------------------------------------
      ! type conversion
      !--------------------------------------------------
      ! exfloat to double
      FUNCTION exflib_float_cast_x2d2(a) RESULT(b)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a, 1)
              b(i,j) = DBLE( a(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_cast_x2d2

      ! ezfloat(:,:) to dcomplex(:,:)
      FUNCTION exflib_float_cast_z2w2(a) RESULT(b)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(SIZE(a,1),SIZE(a,2)) :: b
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a, 1)
              b(i,j) = CMPLX( a(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_cast_z2w2

      !--------------------------------------------------
      ! relation, == (equal)
      !--------------------------------------------------
      ! int == exfloat(:,:)
      FUNCTION exflib_float_eq_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i0x2

      ! quadword == exfloat(:,:)
      FUNCTION exflib_float_eq_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_q0x2

      ! single == exfloat(:,:)
      FUNCTION exflib_float_eq_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s0x2

      ! double == exfloat(:,:)
      FUNCTION exflib_float_eq_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d0x2

      ! exfloat == exfloat(:,:)
      FUNCTION exflib_float_eq_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0x2

      ! scomplex == exfloat(:,:)
      FUNCTION exflib_float_eq_c0x2(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c0x2

      ! dcomplex == exfloat(:,:)
      FUNCTION exflib_float_eq_w0x2(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w0x2

      ! ezfloat == exfloat(:,:)
      FUNCTION exflib_float_eq_z0x2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0x2

      ! int(:,:) == exfloat
      FUNCTION exflib_float_eq_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i2x0

      ! quadword(:,:) == exfloat
      FUNCTION exflib_float_eq_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_q2x0

      ! single(:,:) == exfloat
      FUNCTION exflib_float_eq_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s2x0

      ! double(:,:) == exfloat
      FUNCTION exflib_float_eq_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d2x0

      ! exfloat(:,:) == exfloat
      FUNCTION exflib_float_eq_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2x0

      ! scomplex(:,:) == exfloat
      FUNCTION exflib_float_eq_c2x0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c2x0

      ! dcomplex(:,:) == exfloat
      FUNCTION exflib_float_eq_w2x0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w2x0

      ! ezfloat(:,:) == exfloat
      FUNCTION exflib_float_eq_z2x0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2x0

      ! int(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i2x2

      ! quadword(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
       END DO
      END FUNCTION exflib_float_eq_q2x2

      ! single(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s2x2

      ! double(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d2x2

      ! exfloat(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2x2

      ! scomplex(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_c2x2(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c2x2

      ! dcomplex(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_w2x2(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w2x2

      ! ezfloat(:,:) == exfloat(:,:)
      FUNCTION exflib_float_eq_z2x2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2x2

      ! exfloat(:,:) == int
      FUNCTION exflib_float_eq_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2i0

      ! exfloat(:,:) == quadword
      FUNCTION exflib_float_eq_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2q0

      ! exfloat(:,:) == single
      FUNCTION exflib_float_eq_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2s0

      ! exfloat(:,:) == double
      FUNCTION exflib_float_eq_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2d0

      ! exfloat(:,:) == scomplex
      FUNCTION exflib_float_eq_x2c0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2c0

      ! exfloat(:,:) == dcomplex
      FUNCTION exflib_float_eq_x2w0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2w0

      ! exfloat(:,:) == ezfloat
      FUNCTION exflib_float_eq_x2z0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2z0

      ! exfloat == int(:,:)
      FUNCTION exflib_float_eq_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0i2

      ! exfloat == quadword(:,:)
      FUNCTION exflib_float_eq_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0q2

      ! exfloat == single(:,:)
      FUNCTION exflib_float_eq_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0s2

      ! exfloat == double(:,:)
      FUNCTION exflib_float_eq_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0d2

      ! exfloat == scomplex(:,:)
      FUNCTION exflib_float_eq_x0c2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0c2

      ! exfloat == dcomplex(:,:)
      FUNCTION exflib_float_eq_x0w2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0w2

      ! exfloat == ezfloat(:,:)
      FUNCTION exflib_float_eq_x0z2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x0z2

      ! exfloat(:,:) == int(:,:)
      FUNCTION exflib_float_eq_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2i2

      ! exfloat(:,:) == quadword(:,:)
      FUNCTION exflib_float_eq_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2q2

      ! exfloat(:,:) == single(:,:)
      FUNCTION exflib_float_eq_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2s2

      ! exfloat(:,:) == double(:,:)
      FUNCTION exflib_float_eq_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2d2

      ! exfloat(:,:) == scomplex(:,:)
      FUNCTION exflib_float_eq_x2c2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2c2

      ! exfloat(:,:) == dcomplex(:,:)
      FUNCTION exflib_float_eq_x2w2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2w2

      ! exfloat(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_x2z2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_x2z2

      ! int == ezfloat(:,:)
      FUNCTION exflib_float_eq_i0z2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i0z2

      ! quadword == ezfloat(:,:)
      FUNCTION exflib_float_eq_q0z2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_q0z2

      ! single == ezfloat(:,:)
      FUNCTION exflib_float_eq_s0z2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s0z2

      ! double == ezfloat(:,:)
      FUNCTION exflib_float_eq_d0z2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d0z2

      ! scomplex == ezfloat(:,:)
      FUNCTION exflib_float_eq_c0z2(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c0z2

      ! dcomplex == ezfloat(:,:)
      FUNCTION exflib_float_eq_w0z2(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w0z2

      ! ezfloat == ezfloat(:,:)
      FUNCTION exflib_float_eq_z0z2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0z2

      ! int(:,:) == ezfloat
      FUNCTION exflib_float_eq_i2z0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i2z0

      ! quadword(:,:) == ezfloat
      FUNCTION exflib_float_eq_q2z0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_q2z0

      ! single(:,:) == ezfloat
      FUNCTION exflib_float_eq_s2z0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s2z0

      ! double(:,:) == ezfloat
      FUNCTION exflib_float_eq_d2z0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d2z0

      ! scomplex(:,:) == ezfloat
      FUNCTION exflib_float_eq_c2z0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c2z0

      ! dcomplex(:,:) == ezfloat
      FUNCTION exflib_float_eq_w2z0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w2z0

      ! ezfloat(:,:) == ezfloat
      FUNCTION exflib_float_eq_z2z0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2z0

      ! int(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_i2z2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_i2z2

      ! quadword(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_q2z2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_q2z2

      ! single(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_s2z2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_s2z2

      ! double(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_d2z2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_d2z2

      ! scomplex(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_c2z2(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_c2z2

      ! dcomplex(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_w2z2(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_w2z2

      ! ezfloat(:,:) == ezfloat(:,:)
      FUNCTION exflib_float_eq_z2z2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2z2

      ! ezfloat(:,:) == int
      FUNCTION exflib_float_eq_z2i0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2i0

      ! ezfloat(:,:) == quadword
      FUNCTION exflib_float_eq_z2q0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2q0

      ! ezfloat(:,:) == single
      FUNCTION exflib_float_eq_z2s0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2s0

      ! ezfloat(:,:) == double
      FUNCTION exflib_float_eq_z2d0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2d0

      ! ezfloat(:,:) == scomplex
      FUNCTION exflib_float_eq_z2c0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2c0

      ! ezfloat(:,:) == dcomplex
      FUNCTION exflib_float_eq_z2w0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) == b )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2w0

      ! ezfloat == int(:,:)
      FUNCTION exflib_float_eq_z0i2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0i2

      ! ezfloat == quadword(:,:)
      FUNCTION exflib_float_eq_z0q2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0q2

      ! ezfloat == single(:,:)
      FUNCTION exflib_float_eq_z0s2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0s2

      ! ezfloat == double(:,:)
      FUNCTION exflib_float_eq_z0d2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0d2

      ! ezfloat == scomplex(:,:)
      FUNCTION exflib_float_eq_z0c2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0c2

      ! ezfloat == dcomplex(:,:)
      FUNCTION exflib_float_eq_z0w2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z0w2

      ! ezfloat(:,:) == int(:,:)
      FUNCTION exflib_float_eq_z2i2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2i2

      ! ezfloat(:,:) == quadword(:,:)
      FUNCTION exflib_float_eq_z2q2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2q2

      ! ezfloat(:,:) == single(:,:)
      FUNCTION exflib_float_eq_z2s2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2s2

      ! ezfloat(:,:) == double(:,:)
      FUNCTION exflib_float_eq_z2d2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2d2

      ! ezfloat(:,:) == scomplex(:,:)
      FUNCTION exflib_float_eq_z2c2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2c2

      ! ezfloat(:,:) == dcomplex(:,:)
      FUNCTION exflib_float_eq_z2w2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) == b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_eq_z2w2

      !----------------------------------------
      ! relation, /= (not equal)
      !----------------------------------------
      ! int /= exfloat(:,:)
      FUNCTION exflib_float_ne_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i0x2

      ! quadword /= exfloat(:,:)
      FUNCTION exflib_float_ne_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q0x2

      ! single /= exfloat(:,:)
      FUNCTION exflib_float_ne_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s0x2

      ! double /= exfloat(:,:)
      FUNCTION exflib_float_ne_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d0x2

      ! exfloat /= exfloat(:,:)
      FUNCTION exflib_float_ne_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0x2

      ! scomplex /= exfloat(:,:)
      FUNCTION exflib_float_ne_c0x2(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c0x2

      ! dcomplex /= exfloat(:,:)
      FUNCTION exflib_float_ne_w0x2(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w0x2

      ! ezfloat /= exfloat(:,:)
      FUNCTION exflib_float_ne_z0x2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0x2

      ! int(:,:) /= exfloat
      FUNCTION exflib_float_ne_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i2x0

      ! quadword(:,:) /= exfloat
      FUNCTION exflib_float_ne_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q2x0

      ! single(:,:) /= exfloat
      FUNCTION exflib_float_ne_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s2x0

      ! double(:,:) /= exfloat
      FUNCTION exflib_float_ne_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d2x0

      ! exfloat(:,:) /= exfloat
      FUNCTION exflib_float_ne_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2x0

      ! scomplex(:,:) /= exfloat
      FUNCTION exflib_float_ne_c2x0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c2x0

      ! dcomplex(:,:) /= exfloat
      FUNCTION exflib_float_ne_w2x0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w2x0

      ! ezfloat(:,:) /= exfloat
      FUNCTION exflib_float_ne_z2x0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2x0

      ! int(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i2x2

      ! quadword(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q2x2

      ! single(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s2x2

      ! double(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d2x2

      ! exfloat(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2x2

      ! scomplex(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_c2x2(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c2x2

      ! dcomplex(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_w2x2(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w2x2

      ! ezfloat(:,:) /= exfloat(:,:)
      FUNCTION exflib_float_ne_z2x2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2x2

      ! exfloat(:,:) /= int
      FUNCTION exflib_float_ne_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2i0

      ! exfloat(:,:) /= quadword
      FUNCTION exflib_float_ne_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2q0

      ! exfloat(:,:) /= single
      FUNCTION exflib_float_ne_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2s0

      ! exfloat(:,:) /= double
      FUNCTION exflib_float_ne_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2d0

      ! exfloat(:,:) /= scomplex
      FUNCTION exflib_float_ne_x2c0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2c0

      ! exfloat(:,:) /= dcomplex
      FUNCTION exflib_float_ne_x2w0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2w0

      ! exfloat(:,:) /= ezfloat
      FUNCTION exflib_float_ne_x2z0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2z0

      ! exfloat /= int(:,:)
      FUNCTION exflib_float_ne_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0i2

      ! exfloat /= quadword(:,:)
      FUNCTION exflib_float_ne_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0q2

      ! exfloat /= single(:,:)
      FUNCTION exflib_float_ne_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0s2

      ! exfloat /= double(:,:)
      FUNCTION exflib_float_ne_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0d2

      ! exfloat /= scomplex(:,:)
      FUNCTION exflib_float_ne_x0c2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0c2

      ! exfloat /= dcomplex(:,:)
      FUNCTION exflib_float_ne_x0w2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0w2

      ! exfloat /= ezfloat(:,:)
      FUNCTION exflib_float_ne_x0z2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x0z2

      ! exfloat(:,:) /= int(:,:)
      FUNCTION exflib_float_ne_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2i2

      ! exfloat(:,:) /= quadword(:,:)
      FUNCTION exflib_float_ne_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2q2

      ! exfloat(:,:) /= single(:,:)
      FUNCTION exflib_float_ne_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2s2

      ! exfloat(:,:) /= double(:,:)
      FUNCTION exflib_float_ne_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2d2

      ! exfloat(:,:) /= scomplex(:,:)
      FUNCTION exflib_float_ne_x2c2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2c2

      ! exfloat(:,:) /= dcomplex(:,:)
      FUNCTION exflib_float_ne_x2w2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2w2

      ! exfloat(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_x2z2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_x2z2

      ! int /= ezfloat(:,:)
      FUNCTION exflib_float_ne_i0z2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i0z2

      ! quadword /= ezfloat(:,:)
      FUNCTION exflib_float_ne_q0z2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q0z2

      ! single /= ezfloat(:,:)
      FUNCTION exflib_float_ne_s0z2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s0z2

      ! double /= ezfloat(:,:)
      FUNCTION exflib_float_ne_d0z2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d0z2

      ! scomplex /= ezfloat(:,:)
      FUNCTION exflib_float_ne_c0z2(a, b) RESULT(mask)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c0z2

      ! dcomplex /= ezfloat(:,:)
      FUNCTION exflib_float_ne_w0z2(a, b) RESULT(mask)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w0z2

      ! ezfloat /= ezfloat(:,:)
      FUNCTION exflib_float_ne_z0z2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0z2

      ! int(:,:) /= ezfloat
      FUNCTION exflib_float_ne_i2z0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i2z0

      ! quadword(:,:) /= ezfloat
      FUNCTION exflib_float_ne_q2z0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q2z0

      ! single(:,:) /= ezfloat
      FUNCTION exflib_float_ne_s2z0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s2z0

      ! double(:,:) /= ezfloat
      FUNCTION exflib_float_ne_d2z0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d2z0

      ! scomplex(:,:) /= ezfloat
      FUNCTION exflib_float_ne_c2z0(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c2z0

      ! dcomplex(:,:) /= ezfloat
      FUNCTION exflib_float_ne_w2z0(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w2z0

      ! ezfloat(:,:) /= ezfloat
      FUNCTION exflib_float_ne_z2z0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2z0

      ! int(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_i2z2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_i2z2

      ! quadword(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_q2z2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_q2z2

      ! single(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_s2z2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_s2z2

      ! double(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_d2z2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_d2z2

      ! scomplex(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_c2z2(a, b) RESULT(mask)
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_c2z2

      ! dcomplex(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_w2z2(a, b) RESULT(mask)
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_w2z2

      ! ezfloat(:,:) /= ezfloat(:,:)
      FUNCTION exflib_float_ne_z2z2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2z2

      ! ezfloat(:,:) /= int
      FUNCTION exflib_float_ne_z2i0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2i0

      ! ezfloat(:,:) /= quadword
      FUNCTION exflib_float_ne_z2q0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2q0

      ! ezfloat(:,:) /= single
      FUNCTION exflib_float_ne_z2s0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2s0

      ! ezfloat(:,:) /= double
      FUNCTION exflib_float_ne_z2d0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2d0

      ! ezfloat(:,:) /= scomplex
      FUNCTION exflib_float_ne_z2c0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2c0

      ! ezfloat(:,:) /= dcomplex
      FUNCTION exflib_float_ne_z2w0(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) /= b )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2w0

      ! ezfloat /= int(:,:)
      FUNCTION exflib_float_ne_z0i2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0i2

      ! ezfloat /= quadword(:,:)
      FUNCTION exflib_float_ne_z0q2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0q2

      ! ezfloat /= single(:,:)
      FUNCTION exflib_float_ne_z0s2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0s2

      ! ezfloat /= double(:,:)
      FUNCTION exflib_float_ne_z0d2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0d2

      ! ezfloat /= scomplex(:,:)
      FUNCTION exflib_float_ne_z0c2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0c2

      ! ezfloat /= dcomplex(:,:)
      FUNCTION exflib_float_ne_z0w2(a, b) RESULT(mask)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z0w2

      ! ezfloat(:,:) /= int(:,:)
      FUNCTION exflib_float_ne_z2i2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2i2

      ! ezfloat(:,:) /= quadword(:,:)
      FUNCTION exflib_float_ne_z2q2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2q2

      ! ezfloat(:,:) /= single(:,:)
      FUNCTION exflib_float_ne_z2s2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2s2

      ! ezfloat(:,:) /= double(:,:)
      FUNCTION exflib_float_ne_z2d2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2d2

      ! ezfloat(:,:) /= scomplex(:,:)
      FUNCTION exflib_float_ne_z2c2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2c2

      ! ezfloat(:,:) /= dcomplex(:,:)
      FUNCTION exflib_float_ne_z2w2(a, b) RESULT(mask)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        COMPLEX*16, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a(i,j) /= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ne_z2w2

      !----------------------------------------
      ! relation, < (less than)
      !----------------------------------------
      ! int < exfloat(:,:)
      FUNCTION exflib_float_lt_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_i0x2

      ! quadword < exfloat(:,:)
      FUNCTION exflib_float_lt_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_q0x2

      ! single < exfloat(:,:)
      FUNCTION exflib_float_lt_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_s0x2

      ! double < exfloat(:,:)
      FUNCTION exflib_float_lt_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_d0x2

      ! exfloat < exfloat(:,:)
      FUNCTION exflib_float_lt_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x0x2

      ! int(:,:) < exfloat
      FUNCTION exflib_float_lt_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_i2x0

      ! quadword(:,:) < exfloat
      FUNCTION exflib_float_lt_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_q2x0

      ! single(:,:) < exfloat
      FUNCTION exflib_float_lt_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_s2x0

      ! double(:,:) < exfloat
      FUNCTION exflib_float_lt_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_d2x0

      ! exfloat(:,:) < exfloat
      FUNCTION exflib_float_lt_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2x0

      ! int(:,:) < exfloat(:,:)
      FUNCTION exflib_float_lt_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_i2x2

      ! quadword(:,:) < exfloat(:,:)
      FUNCTION exflib_float_lt_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_q2x2

      ! single(:,:) < exfloat(:,:)
      FUNCTION exflib_float_lt_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_s2x2

      ! double(:,:) < exfloat(:,:)
      FUNCTION exflib_float_lt_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_d2x2

      ! exfloat(:,:) < exfloat(:,:)
      FUNCTION exflib_float_lt_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2x2

      ! exfloat(:,:) < int
      FUNCTION exflib_float_lt_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2i0

      ! exfloat(:,:) < quadword
      FUNCTION exflib_float_lt_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2q0

      ! exfloat(:,:) < single
      FUNCTION exflib_float_lt_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2s0

      ! exfloat(:,:) < double
      FUNCTION exflib_float_lt_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2d0

      ! exfloat < int(:,:)
      FUNCTION exflib_float_lt_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x0i2

      ! exfloat < quadword(:,:)
      FUNCTION exflib_float_lt_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x0q2

      ! exfloat < single(:,:)
      FUNCTION exflib_float_lt_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x0s2

      ! exfloat < double(:,:)
      FUNCTION exflib_float_lt_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x0d2

      ! exfloat(:,:) < int(:,:)
      FUNCTION exflib_float_lt_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2i2

      ! exfloat(:,:) < quadword(:,:)
      FUNCTION exflib_float_lt_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2q2

      ! exfloat(:,:) < single(:,:)
      FUNCTION exflib_float_lt_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2s2

      ! exfloat(:,:) < double(:,:)
      FUNCTION exflib_float_lt_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) < b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_lt_x2d2

      !----------------------------------------
      ! relation, <= (less than or equal)
      !----------------------------------------
      ! int <= exfloat(:,:)
      FUNCTION exflib_float_le_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_i0x2

      ! quadword <= exfloat(:,:)
      FUNCTION exflib_float_le_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_q0x2

      ! single <= exfloat(:,:)
      FUNCTION exflib_float_le_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_s0x2

      ! double <= exfloat(:,:)
      FUNCTION exflib_float_le_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_d0x2

      ! exfloat <= exfloat(:,:)
      FUNCTION exflib_float_le_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x0x2

      ! int(:,:) <= exfloat
      FUNCTION exflib_float_le_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_i2x0

      ! quadword(:,:) <= exfloat
      FUNCTION exflib_float_le_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_q2x0

      ! single(:,:) <= exfloat
      FUNCTION exflib_float_le_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_s2x0

      ! double(:,:) <= exfloat
      FUNCTION exflib_float_le_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_d2x0

      ! exfloat(:,:) <= exfloat
      FUNCTION exflib_float_le_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2x0

      ! int(:,:) <= exfloat(:,:)
      FUNCTION exflib_float_le_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_i2x2

      ! quadword(:,:) <= exfloat(:,:)
      FUNCTION exflib_float_le_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_q2x2

      ! single(:,:) <= exfloat(:,:)
      FUNCTION exflib_float_le_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_s2x2

      ! double(:,:) <= exfloat(:,:)
      FUNCTION exflib_float_le_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_d2x2

      ! exfloat(:,:) <= exfloat(:,:)
      FUNCTION exflib_float_le_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2x2

      ! exfloat(:,:) <= int
      FUNCTION exflib_float_le_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2i0

      ! exfloat(:,:) <= quadword
      FUNCTION exflib_float_le_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2q0

      ! exfloat(:,:) <= single
      FUNCTION exflib_float_le_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2s0

      ! exfloat(:,:) <= double
      FUNCTION exflib_float_le_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2d0

      ! exfloat <= int(:,:)
      FUNCTION exflib_float_le_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x0i2

      ! exfloat <= quadword(:,:)
      FUNCTION exflib_float_le_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x0q2

      ! exfloat <= single(:,:)
      FUNCTION exflib_float_le_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x0s2

      ! exfloat <= double(:,:)
      FUNCTION exflib_float_le_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x0d2

      ! exfloat(:,:) <= int(:,:)
      FUNCTION exflib_float_le_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2i2

      ! exfloat(:,:) <= quadword(:,:)
      FUNCTION exflib_float_le_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2q2

      ! exfloat(:,:) <= single(:,:)
      FUNCTION exflib_float_le_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2s2

      ! exfloat(:,:) <= double(:,:)
      FUNCTION exflib_float_le_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) <= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_le_x2d2

      !----------------------------------------
      ! relation, > (greater than)
      !----------------------------------------
      ! int > exfloat(:,:)
      FUNCTION exflib_float_gt_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_i0x2

      ! quadword > exfloat(:,:)
      FUNCTION exflib_float_gt_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_q0x2

      ! single > exfloat(:,:)
      FUNCTION exflib_float_gt_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_s0x2

      ! double > exfloat(:,:)
      FUNCTION exflib_float_gt_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_d0x2

      ! exfloat > exfloat(:,:)
      FUNCTION exflib_float_gt_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x0x2

      ! int(:,:) > exfloat
      FUNCTION exflib_float_gt_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_i2x0

      ! quadword(:,:) > exfloat
      FUNCTION exflib_float_gt_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_q2x0

      ! single(:,:) > exfloat
      FUNCTION exflib_float_gt_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_s2x0

      ! double(:,:) > exfloat
      FUNCTION exflib_float_gt_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_d2x0

      ! exfloat(:,:) > exfloat
      FUNCTION exflib_float_gt_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2x0

      ! int(:,:) > exfloat(:,:)
      FUNCTION exflib_float_gt_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_i2x2

      ! quadword(:,:) > exfloat(:,:)
      FUNCTION exflib_float_gt_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_q2x2

      ! single(:,:) > exfloat(:,:)
      FUNCTION exflib_float_gt_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_s2x2

      ! double(:,:) > exfloat(:,:)
      FUNCTION exflib_float_gt_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_d2x2

      ! exfloat(:,:) > exfloat(:,:)
      FUNCTION exflib_float_gt_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2x2

      ! exfloat(:,:) > int
      FUNCTION exflib_float_gt_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2i0

      ! exfloat(:,:) > quadword
      FUNCTION exflib_float_gt_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2q0

      ! exfloat(:,:) > single
      FUNCTION exflib_float_gt_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2s0

      ! exfloat(:,:) > double
      FUNCTION exflib_float_gt_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2d0

      ! exfloat > int(:,:)
      FUNCTION exflib_float_gt_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x0i2

      ! exfloat > quadword(:,:)
      FUNCTION exflib_float_gt_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x0q2

      ! exfloat > single(:,:)
      FUNCTION exflib_float_gt_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x0s2

      ! exfloat > double(:,:)
      FUNCTION exflib_float_gt_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x0d2

      ! exfloat(:,:) > int(:,:)
      FUNCTION exflib_float_gt_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2i2

      ! exfloat(:,:) > quadword(:,:)
      FUNCTION exflib_float_gt_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2q2

      ! exfloat(:,:) > single(:,:)
      FUNCTION exflib_float_gt_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2s2

      ! exfloat(:,:) > double(:,:)
      FUNCTION exflib_float_gt_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) > b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_gt_x2d2

      !----------------------------------------
      ! relation, >= (greater than or equal)
      !----------------------------------------
      ! int >= exfloat(:,:)
      FUNCTION exflib_float_ge_i0x2(a, b) RESULT(mask)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_i0x2

      ! quadword >= exfloat(:,:)
      FUNCTION exflib_float_ge_q0x2(a, b) RESULT(mask)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_q0x2

      ! single >= exfloat(:,:)
      FUNCTION exflib_float_ge_s0x2(a, b) RESULT(mask)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_s0x2

      ! double >= exfloat(:,:)
      FUNCTION exflib_float_ge_d0x2(a, b) RESULT(mask)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_d0x2

      ! exfloat >= exfloat(:,:)
      FUNCTION exflib_float_ge_x0x2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x0x2

      ! int(:,:) >= exfloat
      FUNCTION exflib_float_ge_i2x0(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_i2x0

      ! quadword(:,:) >= exfloat
      FUNCTION exflib_float_ge_q2x0(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_q2x0

      ! single(:,:) >= exfloat
      FUNCTION exflib_float_ge_s2x0(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_s2x0

      ! double(:,:) >= exfloat
      FUNCTION exflib_float_ge_d2x0(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_d2x0

      ! exfloat(:,:) >= exfloat
      FUNCTION exflib_float_ge_x2x0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2x0

      ! int(:,:) >= exfloat(:,:)
      FUNCTION exflib_float_ge_i2x2(a, b) RESULT(mask)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_i2x2

      ! quadword(:,:) >= exfloat(:,:)
      FUNCTION exflib_float_ge_q2x2(a, b) RESULT(mask)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_q2x2

      ! single(:,:) >= exfloat(:,:)
      FUNCTION exflib_float_ge_s2x2(a, b) RESULT(mask)
        REAL*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_s2x2

      ! double(:,:) >= exfloat(:,:)
      FUNCTION exflib_float_ge_d2x2(a, b) RESULT(mask)
        REAL*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_d2x2

      ! exfloat(:,:) >= exfloat(:,:)
      FUNCTION exflib_float_ge_x2x2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2x2

      ! exfloat(:,:) >= int
      FUNCTION exflib_float_ge_x2i0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2i0

      ! exfloat(:,:) >= quadword
      FUNCTION exflib_float_ge_x2q0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2q0

      ! exfloat(:,:) >= single
      FUNCTION exflib_float_ge_x2s0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2s0

      ! exfloat(:,:) >= double
      FUNCTION exflib_float_ge_x2d0(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2d0

      ! exfloat >= int(:,:)
      FUNCTION exflib_float_ge_x0i2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x0i2

      ! exfloat >= quadword(:,:)
      FUNCTION exflib_float_ge_x0q2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x0q2

      ! exfloat >= single(:,:)
      FUNCTION exflib_float_ge_x0s2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x0s2

      ! exfloat >= double(:,:)
      FUNCTION exflib_float_ge_x0d2(a, b) RESULT(mask)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(b,1),SIZE(b,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              mask(i,j) = ( a >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x0d2

      ! exfloat(:,:) >= int(:,:)
      FUNCTION exflib_float_ge_x2i2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2i2

      ! exfloat(:,:) >= quadword(:,:)
      FUNCTION exflib_float_ge_x2q2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2q2

      ! exfloat(:,:) >= single(:,:)
      FUNCTION exflib_float_ge_x2s2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*4, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2s2

      ! exfloat(:,:) >= double(:,:)
      FUNCTION exflib_float_ge_x2d2(a, b) RESULT(mask)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        REAL*8, DIMENSION(:,:),INTENT(IN) :: b
        LOGICAL*8, DIMENSION(SIZE(a,1),SIZE(a,2)) :: mask
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              mask(i,j) = ( a(i,j) >= b(i,j) )
           END DO
        END DO
      END FUNCTION exflib_float_ge_x2d2

      !--------------------------------------------------
      ! two dimensional array functions
      !--------------------------------------------------
      ! sum of exfloat(:,:)
      FUNCTION exflib_float_sum_x2(a, mask) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: s
        INTEGER*8 :: i, j

        s = 0
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 s = s + a(i,j)
              END DO
           END DO

        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( mask(i,j) ) THEN
                    s = s + a(i,j)
                 END IF
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_sum_x2

      ! sum of exfloat(:,:) along dimension
      FUNCTION exflib_float_sum_dim_x2(a, dim, mask) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: s
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN
              DO j = 1, SIZE(a,2)
                 s(j) = 0
                 DO i = 1, SIZE(a,1)
                    s(j) = s(j) + a(i,j)
                 END DO
              END DO
           ElSE

              DO j = 1, SIZE(a,2)
                 s(j) = 0
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(j) = s(j) + a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              s(i) = 0
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    s(i) = s(i) + a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(i) = s(i) + a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        END IF

      END FUNCTION exflib_float_sum_dim_x2

      ! sum of ezfloat(:,:)
      FUNCTION exflib_float_sum_z2(a, mask) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat) :: s
        INTEGER*8 :: i, j

        s = 0
        IF ( .NOT.PRESENT( mask ) ) THEN
           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 s = s + a(i,j)
              END DO
           END DO
        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( mask(i,j) ) THEN
                    s = s + a(i,j)
                 END IF
              END DO
           END DO
        END IF

      END FUNCTION exflib_float_sum_z2

      ! sum of ezfloat(:,:) along dimension
      FUNCTION exflib_float_sum_dim_z2(a, dim, mask) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: s
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 s(j) = 0
                 DO i = 1, SIZE(a,1)
                    s(j) = s(j) + a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 s(j) = 0
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(j) = s(j) + a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              s(i) = 0
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    s(i) = s(i) + a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(i) = s(i) + a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        END IF

      END FUNCTION exflib_float_sum_dim_z2

      ! product of exfloat(:,:)
      FUNCTION exflib_float_product_x2(a, mask) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: s
        INTEGER*8 :: i, j

        s = 1
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 s = s * a(i,j)
              END DO
           END DO

        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( mask(i,j) ) THEN
                    s = s * a(i,j)
                 END IF
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_product_x2

      ! product of exfloat(:,:) along dimension
      FUNCTION exflib_float_product_dim_x2(a, dim, mask) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: s
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 s(j) = 1
                 DO i = 1, SIZE(a,1)
                    s(j) = s(j) * a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 s(j) = 1
                 DO i = 1, SIZE(a,1)
                    IF( mask(i,j) ) THEN
                       s(j) = s(j) * a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              s(i) = 1
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    s(i) = s(i) * a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(i) = s(i) * a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        END IF

      END FUNCTION exflib_float_product_dim_x2

      ! product of ezfloat(:,:)
      FUNCTION exflib_float_product_z2(a, mask) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat) :: s
        INTEGER*8 :: i, j

        s = 1
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 s = s * a(i,j)
              END DO
           END DO

        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 s = s * a(i,j)
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_product_z2

      ! product of ezfloat(:,:) along dimension
      FUNCTION exflib_float_product_dim_z2(a, dim, mask) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(ezfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: s
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 s(j) = 1
                 DO i = 1, SIZE(a,1)
                    s(j) = s(j) * a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 s(j) = 1
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i, j) ) THEN
                       s(j) = s(j) * a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              s(i) = 1
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    s(i) = s(i) * a(i,j)
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( mask(i,j) ) THEN
                       s(i) = s(i) * a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        END IF

      END FUNCTION exflib_float_product_dim_z2

      !----------------------------------------
      ! two dimensional max/min functions
      !----------------------------------------
      ! find maximum in exfloat(:,:)
      FUNCTION exflib_float_maxval_x2(a, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: m
        INTEGER*8 :: i, j

        m = -exflib_float_huge_m(a(1,1))

        ! column oriented search
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) > m ) THEN
                    m = a(i,j)
                 END IF
              END DO
           END DO

        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( (mask(i,j)).AND.(a(i,j) > m) ) THEN
                    m = a(i,j)
                 END IF
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_maxval_x2

      ! find maximum in exfloat(:,:) along dimension
      FUNCTION exflib_float_maxval_dim_x2(a, dim, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: m
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 m(j) = -exflib_float_huge_m(a(1,1))
                 DO i = 1, SIZE(a,1)
                    IF ( a(i,j) > m(j) ) THEN
                       m(j) = a(i,j)
                    END IF
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 m(j) = -exflib_float_huge_m(a(1,1))
                 DO i = 1, SIZE(a,1)
                    IF ( (mask(i,j)).AND.(a(i,j) > m(j)) ) THEN
                       m(j) = a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              m(i) = -exflib_float_huge_m(a(1,1))
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( a(i,j) > m(i) ) THEN
                       m(i) = a(i,j)
                    END IF
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( (mask(i,j)).AND.(a(i,j) > m(i)) ) THEN
                       m(i) = a(i,j)
                    END IF
                 END DO
              END DO

           END IF

        END IF

      END FUNCTION exflib_float_maxval_dim_x2

      ! find minimum in exfloat(:,:)
      FUNCTION exflib_float_minval_x2(a, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat) :: m
        INTEGER*8 :: i, j

        m = exflib_float_huge_m(a(1,1))

        ! column oriented search
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              IF ( a(i,j) < m ) THEN
                 m = a(i,j)
              END IF
           END DO
        END DO

      END FUNCTION exflib_float_minval_x2

      ! find minimum in exfloat(:,:) along dimension
      FUNCTION exflib_float_minval_dim_x2(a, dim, mask) RESULT(m)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        TYPE(exfloat), DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: m
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           DO j = 1, SIZE(a,2)
              m(j) = exflib_float_huge_m(a(1,1))
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) < m(j) ) THEN
                    m(j) = a(i,j)
                 END IF
              END DO
           END DO

        ELSE IF ( dim == 2 ) THEN

           DO i = 1, SIZE(a,1)
              m(i) = exflib_float_huge_m(a(1,1))
           END DO

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) < m(i) ) THEN
                    m(i) = a(i,j)
                 END IF
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_minval_dim_x2

      ! find location of maximum in exfloat(:,:)
      FUNCTION exflib_float_maxloc_x2(a, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(2) :: idx
        TYPE(exfloat) :: m
        INTEGER*8 :: i, j

        idx = (/EXFLIB_ARRAY_BEGIN, EXFLIB_ARRAY_BEGIN/)
        m = -exflib_float_huge_m(a(1,1))

        ! column oriented search
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              IF ( a(i,j) > m ) THEN
                 m = a(i,j)
                 idx = (/i,j/)
              END IF
           END DO
        END DO

      END FUNCTION exflib_float_maxloc_x2

      ! find location of maximum in exfloat(:,:) along dimension
      FUNCTION exflib_float_maxloc_dim_x2(a, dim, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: idx
        TYPE(exfloat) :: m
        TYPE(exfloat), DIMENSION(:), ALLOCATABLE :: m2
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           DO j = 1, SIZE(a,2)
              idx(j) = EXFLIB_ARRAY_BEGIN
              m = -exflib_float_huge_m(a(1,1))
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) > m ) THEN
                    m = a(i,j)
                    idx(j) = i
                 END IF
              END DO
           END DO

        ELSE IF ( dim == 2 ) THEN

           ALLOCATE( m2(SIZE(a,1)) )

           DO i = 1, SIZE(a,1)
              idx(i) = EXFLIB_ARRAY_BEGIN
              m2(i) = -exflib_float_huge_m(a(1,1))
           END DO

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) > m2(i) ) THEN
                    m2(i) = a(i,j)
                    idx(i) = j
                 END IF
              END DO
           END DO

           DEALLOCATE( m2 )

        END IF

      END FUNCTION exflib_float_maxloc_dim_x2

      ! find location of minimum in exfloat(:,:)
      FUNCTION exflib_float_minloc_x2(a, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(2) :: idx
        TYPE(exfloat) :: m
        INTEGER*8 :: i, j

        idx = (/EXFLIB_ARRAY_BEGIN, EXFLIB_ARRAY_BEGIN/)
        m = exflib_float_huge_m(a(1,1))

        ! column oriented search
        IF ( .NOT.PRESENT( mask ) ) THEN

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( a(i,j) < m ) THEN
                    m = a(i,j)
                    idx = (/i,j/)
                 END IF
              END DO
           END DO

        ELSE

           DO j = 1, SIZE(a,2)
              DO i = 1, SIZE(a,1)
                 IF ( (mask(i,j)).AND.(a(i,j) < m) ) THEN
                    m = a(i,j)
                    idx = (/i,j/)
                 END IF
              END DO
           END DO

        END IF

      END FUNCTION exflib_float_minloc_x2

      ! find location of minimum in exfloat(:,:) along dimension
      FUNCTION exflib_float_minloc_dim_x2(a, dim, mask) RESULT(idx)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER, INTENT(IN) :: dim
        LOGICAL*8, DIMENSION(:,:), INTENT(IN), OPTIONAL :: mask
        INTEGER, DIMENSION(MAX(SIZE(a,1),SIZE(a,2))) :: idx
        TYPE(exfloat) :: m
        TYPE(exfloat), DIMENSION(:), ALLOCATABLE :: m2
        INTEGER*8 :: i, j

        IF ( dim == 1 ) THEN

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 idx(j) = EXFLIB_ARRAY_BEGIN
                 m = exflib_float_huge_m(a(1,1))
                 DO i = 1, SIZE(a,1)
                    IF ( a(i,j) < m ) THEN
                       m = a(i,j)
                       idx(j) = i
                    END IF
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 idx(j) = EXFLIB_ARRAY_BEGIN
                 m = exflib_float_huge_m(a(1,1))
                 DO i = 1, SIZE(a,1)
                    IF ( (mask(i,j)).AND.(a(i,j) < m) ) THEN
                       m = a(i,j)
                       idx(j) = i
                    END IF
                 END DO
              END DO

           END IF

        ELSE IF ( dim == 2 ) THEN

           ALLOCATE( m2(SIZE(a,1)) )

           DO i = 1, SIZE(a,1)
              idx(i) = EXFLIB_ARRAY_BEGIN
              m2(i) = exflib_float_huge_m(a(1,1))
           END DO

           IF ( .NOT.PRESENT( mask ) ) THEN

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( a(i,j) < m2(i) ) THEN
                       m2(i) = a(i,j)
                       idx(i) = j
                    END IF
                 END DO
              END DO

           ELSE

              DO j = 1, SIZE(a,2)
                 DO i = 1, SIZE(a,1)
                    IF ( (mask(i,j)).AND.(a(i,j) < m2(i)) ) THEN
                       m2(i) = a(i,j)
                       idx(i) = j
                    END IF
                 END DO
              END DO

           END IF

           DEALLOCATE( m2 )

        END IF

      END FUNCTION exflib_float_minloc_dim_x2

      !--------------------------------------------------
      ! two dimensional (matrix) operations
      !--------------------------------------------------
      ! matrix(exfloat) * matrix(exfloat)
      FUNCTION exflib_float_matmul_x2x2(a, b) RESULT(p)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a, b
        TYPE(exfloat), DIMENSION(SIZE(a,1), SIZE(a,2)) :: p
        INTEGER*8 :: i, j, k

        !IF ( SIZE(a,2) /= SIZE(b,1) ) THEN
        !   STOP
        !END IF

        DO i = 1, SIZE(a,1)
           DO j = 1, SIZE(b,2)
              p(i,j) = 0
              DO k = 1, SIZE(a,2)
                 p(i,j) = p(i,j) + a(i,k)*b(k,j)
              END DO
           END DO
        END DO

      END FUNCTION exflib_float_matmul_x2x2

      ! matrix(exfloat) * matrix(ezfloat)
      FUNCTION exflib_float_matmul_x2z2(a, b) RESULT(p)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1), SIZE(a,2)) :: p
        INTEGER*8 :: i, j, k

        !IF ( SIZE(a,2) /= SIZE(b,1) ) THEN
        !   STOP
        !END IF

        DO i = 1, SIZE(a,1)
           DO j = 1, SIZE(b,2)
              p(i,j) = 0
              DO k = 1, SIZE(a,2)
                 p(i,j) = p(i,j) + a(i,k)*b(k,j)
              END DO
           END DO
        END DO

      END FUNCTION exflib_float_matmul_x2z2

      ! matrix(ezfloat) * matrix(exfloat)
      FUNCTION exflib_float_matmul_z2x2(a, b) RESULT(p)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1), SIZE(a,2)) :: p
        INTEGER*8 :: i, j, k

        !IF ( SIZE(a,2) /= SIZE(b,1) ) THEN
        !   STOP
        !END IF

        DO i = 1, SIZE(a,1)
           DO j = 1, SIZE(b,2)
              p(i,j) = 0
              DO k = 1, SIZE(a,2)
                 p(i,j) = p(i,j) + a(i,k)*b(k,j)
              END DO
           END DO
        END DO

      END FUNCTION exflib_float_matmul_z2x2

      ! matrix(ezfloat) * matrix(ezfloat)
      FUNCTION exflib_float_matmul_z2z2(a, b) RESULT(p)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a, b
        TYPE(ezfloat), DIMENSION(SIZE(a,1), SIZE(a,2)) :: p
        INTEGER*8 :: i, j, k

        !IF ( SIZE(a,2) /= SIZE(b,1) ) THEN
        !   STOP
        !END IF

        DO i = 1, SIZE(a,1)
           DO j = 1, SIZE(b,2)
              p(i,j) = 0
              DO k = 1, SIZE(a,2)
                 p(i,j) = p(i,j) + a(i,k)*b(k,j)
              END DO
           END DO
        END DO

      END FUNCTION exflib_float_matmul_z2z2

      ! transpose of matrix(exfloat)
      FUNCTION exflib_float_transpose_x2(a) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(SIZE(a,2),SIZE(a,1)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(j,i) = a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_transpose_x2

      ! transpose of matrix(ezfloat)
      FUNCTION exflib_float_transpose_z2(a) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(SIZE(a,2),SIZE(a,1)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(j,i) = a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_transpose_z2

      !----------------------------------------
      ! array addition
      !----------------------------------------
      ! int + exfloat(:,:)
      FUNCTION exflib_float_add_i0x2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_i0x2

      ! quadword + exfloat(:,:)
      FUNCTION exflib_float_add_q0x2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_q0x2

      ! exfloat + exfloat(:,:)
      FUNCTION exflib_float_add_x0x2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x0x2

      ! ezfloat + exfloat(:,:)
      FUNCTION exflib_float_add_z0x2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z0x2

      ! int(:,:) + exfloat
      FUNCTION exflib_float_add_i2x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_i2x0

      ! quadword(:,:) + exfloat
      FUNCTION exflib_float_add_q2x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_q2x0

      ! exfloat(:,:) + exfloat
      FUNCTION exflib_float_add_x2x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_x2x0

      ! ezfloat(:,:) + exfloat
      FUNCTION exflib_float_add_z2x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_z2x0

      ! int(:,:) + exfloat(:,:)
      FUNCTION exflib_float_add_i2x2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_i2x2

      ! quadword(:,:) + exfloat(:,:)
      FUNCTION exflib_float_add_q2x2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_q2x2

      ! exfloat(:,:) + exfloat(:,:)
      FUNCTION exflib_float_add_x2x2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x2x2

      ! ezfloat(:,:) + exfloat(:,:)
      FUNCTION exflib_float_add_z2x2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z2x2

      ! exfloat + int(:,:)
      FUNCTION exflib_float_add_x0i2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x0i2

      ! exfloat + quadword(:,:)
      FUNCTION exflib_float_add_x0q2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x0q2

      ! exfloat + ezfloat(:,:)
      FUNCTION exflib_float_add_x0z2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x0z2

      ! exfloat(:,:) + int
      FUNCTION exflib_float_add_x2i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_x2i0

      ! exfloat(:,:) + quadword
      FUNCTION exflib_float_add_x2q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_x2q0

      ! exfloat(:,:) + ezfloat
      FUNCTION exflib_float_add_x2z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_x2z0

      ! exfloat(:,:) + int(:,:)
      FUNCTION exflib_float_add_x2i2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x2i2

      ! exfloat(:,:) + quadword(:,:)
      FUNCTION exflib_float_add_x2q2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x2q2

      ! exfloat(:,:) + ezfloat(:,:)
      FUNCTION exflib_float_add_x2z2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_x2z2

      ! int + ezfloat(:,:)
      FUNCTION exflib_float_add_i0z2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_i0z2

      ! quadword + ezfloat(:,:)
      FUNCTION exflib_float_add_q0z2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_q0z2

      ! ezfloat + ezfloat(:,:)
      FUNCTION exflib_float_add_z0z2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z0z2

      ! int(:,:) + ezfloat
      FUNCTION exflib_float_add_i2z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_i2z0

      ! quadword(:,:) + ezfloat
      FUNCTION exflib_float_add_q2z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_q2z0

      ! ezfloat(:,:) + ezfloat
      FUNCTION exflib_float_add_z2z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_z2z0

      ! int(:,:) + ezfloat(:,:)
      FUNCTION exflib_float_add_i2z2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_i2z2

      ! quadword(:,:) + ezfloat(:,:)
      FUNCTION exflib_float_add_q2z2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_q2z2

      ! ezfloat(:,:) + ezfloat(:,:)
      FUNCTION exflib_float_add_z2z2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z2z2

      ! ezfloat + int(:,:)
      FUNCTION exflib_float_add_z0i2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z0i2

      ! ezfloat + quadword(:,:)
      FUNCTION exflib_float_add_z0q2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z0q2

      ! ezfloat(:,:) + int
      FUNCTION exflib_float_add_z2i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_z2i0

      ! ezfloat(:,:) + quadword
      FUNCTION exflib_float_add_z2q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b
           END DO
        END DO
      END FUNCTION exflib_float_add_z2q0

      ! ezfloat(:,:) + int(:,:)
      FUNCTION exflib_float_add_z2i2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z2i2

      ! ezfloat(:,:) + quadword(:,:)
      FUNCTION exflib_float_add_z2q2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) + b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_add_z2q2

      ! unarray operator, +exfloat(:,:)
      FUNCTION exflib_float_pls_x2(a) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_pls_x2

      ! unarray operator, +ezfloat(:,:)
      FUNCTION exflib_float_pls_z2(a) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_pls_z2

      !----------------------------------------
      ! array subtraction
      !----------------------------------------
      ! int - exfloat(:,:)
      FUNCTION exflib_float_sub_i0x2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_i0x2

      ! quadword - exfloat(:,:)
      FUNCTION exflib_float_sub_q0x2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_q0x2

      ! exfloat - exfloat(:,:)
      FUNCTION exflib_float_sub_x0x2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x0x2

      ! ezfloat - exfloat(:,:)
      FUNCTION exflib_float_sub_z0x2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z0x2

      ! int(:,:) - exfloat
      FUNCTION exflib_float_sub_i2x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_i2x0

      ! quadword(:,:) - exfloat
      FUNCTION exflib_float_sub_q2x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_q2x0

      ! exfloat(:,:) - exfloat
      FUNCTION exflib_float_sub_x2x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2x0

      ! ezfloat(:,:) - exfloat
      FUNCTION exflib_float_sub_z2x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2x0

      ! int(:,:) - exfloat(:,:)
      FUNCTION exflib_float_sub_i2x2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_i2x2

      ! quadword(:,:) - exfloat(:,:)
      FUNCTION exflib_float_sub_q2x2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_q2x2

      ! exfloat(:,:) - exfloat(:,:)
      FUNCTION exflib_float_sub_x2x2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2x2

      ! ezfloat(:,:) - exfloat(:,:)
      FUNCTION exflib_float_sub_z2x2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2x2

      ! exfloat - int(:,:)
      FUNCTION exflib_float_sub_x0i2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x0i2

      ! exfloat - quadword(:,:)
      FUNCTION exflib_float_sub_x0q2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x0q2

      ! exfloat - ezfloat(:,:)
      FUNCTION exflib_float_sub_x0z2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x0z2

      ! exfloat(:,:) - int
      FUNCTION exflib_float_sub_x2i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2i0

      ! exfloat(:,:) - quadword
      FUNCTION exflib_float_sub_x2q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2q0

      ! exfloat(:,:) - ezfloat
      FUNCTION exflib_float_sub_x2z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2z0

      ! exfloat(:,:) - int(:,:)
      FUNCTION exflib_float_sub_x2i2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2i2

      ! exfloat(:,:) - quadword(:,:)
      FUNCTION exflib_float_sub_x2q2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2q2

      ! exfloat(:,:) - ezfloat(:,:)
      FUNCTION exflib_float_sub_x2z2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_x2z2

      ! int - ezfloat(:,:)
      FUNCTION exflib_float_sub_i0z2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_i0z2

      ! quadword - ezfloat(:,:)
      FUNCTION exflib_float_sub_q0z2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_q0z2

      ! ezfloat - ezfloat(:,:)
      FUNCTION exflib_float_sub_z0z2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z0z2

      ! int(:,:) - ezfloat
      FUNCTION exflib_float_sub_i2z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_i2z0

      ! quadword(:,:) - ezfloat
      FUNCTION exflib_float_sub_q2z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_q2z0

      ! ezfloat(:,:) - ezfloat
      FUNCTION exflib_float_sub_z2z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2z0

      ! int(:,:) - ezfloat(:,:)
      FUNCTION exflib_float_sub_i2z2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_i2z2

      ! quadword(:,:) - ezfloat(:,:)
      FUNCTION exflib_float_sub_q2z2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_q2z2

      ! ezfloat(:,:) - ezfloat(:,:)
      FUNCTION exflib_float_sub_z2z2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2z2

      ! ezfloat - int(:,:)
      FUNCTION exflib_float_sub_z0i2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z0i2

      ! ezfloat - quadword(:,:)
      FUNCTION exflib_float_sub_z0q2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z0q2

      ! ezfloat(:,:) - int
      FUNCTION exflib_float_sub_z2i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2i0

      ! ezfloat(:,:) - quadword
      FUNCTION exflib_float_sub_z2q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2q0

      ! ezfloat(:,:) - int(:,:)
      FUNCTION exflib_float_sub_z2i2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2i2

      ! ezfloat(:,:) - quadword(:,:)
      FUNCTION exflib_float_sub_z2q2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) - b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_sub_z2q2

      ! unarray operator, -exfloat(:,:)
      FUNCTION exflib_float_neg_x2(a) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = - a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_neg_x2

      ! unarray operator, -ezfloat(:,:)
      FUNCTION exflib_float_neg_z2(a) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = - a(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_neg_z2

      !----------------------------------------
      ! array multiplication
      !----------------------------------------
      ! int * exfloat(:,:)
      FUNCTION exflib_float_mul_i0x2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_i0x2

      ! quadword * exfloat(:,:)
      FUNCTION exflib_float_mul_q0x2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_q0x2

      ! exfloat * exfloat(:,:)
      FUNCTION exflib_float_mul_x0x2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x0x2

      ! ezfloat * exfloat(:,:)
      FUNCTION exflib_float_mul_z0x2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z0x2

      ! int(:,:) * exfloat
      FUNCTION exflib_float_mul_i2x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_i2x0

      ! quadword(:,:) * exfloat
      FUNCTION exflib_float_mul_q2x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_q2x0

      ! exfloat(:,:) * exfloat
      FUNCTION exflib_float_mul_x2x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2x0

      ! ezfloat(:,:) * exfloat
      FUNCTION exflib_float_mul_z2x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2x0

      ! int(:,:) * exfloat(:,:)
      FUNCTION exflib_float_mul_i2x2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_i2x2

      ! quadword(:,:) * exfloat(:,:)
      FUNCTION exflib_float_mul_q2x2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_q2x2

      ! exfloat(:,:) * exfloat(:,:)
      FUNCTION exflib_float_mul_x2x2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2x2

      ! ezfloat(:,:) * exfloat(:,:)
      FUNCTION exflib_float_mul_z2x2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2x2

      ! exfloat * int(:,:)
      FUNCTION exflib_float_mul_x0i2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x0i2

      ! exfloat * quadword(:,:)
      FUNCTION exflib_float_mul_x0q2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x0q2

      ! exfloat * ezfloat(:,:)
      FUNCTION exflib_float_mul_x0z2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x0z2

      ! exfloat(:,:) * int
      FUNCTION exflib_float_mul_x2i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2i0

      ! exfloat(:,:) * quadword
      FUNCTION exflib_float_mul_x2q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2q0

      ! exfloat(:,:) * ezfloat
      FUNCTION exflib_float_mul_x2z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2z0

      ! exfloat(:,:) * int(:,:)
      FUNCTION exflib_float_mul_x2i2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2i2

      ! exfloat(:,:) * quadword(:,:)
      FUNCTION exflib_float_mul_x2q2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2q2

      ! exfloat(:,:) * ezfloat(:,:)
      FUNCTION exflib_float_mul_x2z2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_x2z2

      ! int * ezfloat(:,:)
      FUNCTION exflib_float_mul_i0z2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_i0z2

      ! quadword * ezfloat(:,:)
      FUNCTION exflib_float_mul_q0z2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_q0z2

      ! ezfloat * ezfloat(:,:)
      FUNCTION exflib_float_mul_z0z2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z0z2

      ! int(:,:) * ezfloat
      FUNCTION exflib_float_mul_i2z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_i2z0

      ! quadword(:,:) * ezfloat
      FUNCTION exflib_float_mul_q2z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_q2z0

      ! ezfloat(:,:) * ezfloat
      FUNCTION exflib_float_mul_z2z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2z0

      ! int(:,:) * ezfloat(:,:)
      FUNCTION exflib_float_mul_i2z2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_i2z2

      ! quadword(:,:) * ezfloat(:,:)
      FUNCTION exflib_float_mul_q2z2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_q2z2

      ! ezfloat(:,:) * ezfloat(:,:)
      FUNCTION exflib_float_mul_z2z2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2z2

      ! ezfloat * int(:,:)
      FUNCTION exflib_float_mul_z0i2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z0i2

      ! ezfloat * quadword(:,:)
      FUNCTION exflib_float_mul_z0q2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z0q2

      ! ezfloat(:,:) * int
      FUNCTION exflib_float_mul_z2i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2i0

      ! ezfloat(:,:) * quadword
      FUNCTION exflib_float_mul_z2q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2q0

      ! ezfloat(:,:) * int(:,:)
      FUNCTION exflib_float_mul_z2i2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2i2

      ! ezfloat(:,:) * quadword(:,:)
      FUNCTION exflib_float_mul_z2q2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) * b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_mul_z2q2

      !----------------------------------------
      ! array division
      !----------------------------------------
      ! int / exfloat(:,:)
      FUNCTION exflib_float_div_i0x2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_i0x2

      ! quadword / exfloat(:,:)
      FUNCTION exflib_float_div_q0x2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_q0x2

      ! exfloat / exfloat(:,:)
      FUNCTION exflib_float_div_x0x2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x0x2

      ! ezfloat / exfloat(:,:)
      FUNCTION exflib_float_div_z0x2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z0x2

      ! int(:,:) / exfloat
      FUNCTION exflib_float_div_i2x0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_i2x0

      ! quadword(:,:) / exfloat
      FUNCTION exflib_float_div_q2x0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_q2x0

      ! exfloat(:,:) / exfloat
      FUNCTION exflib_float_div_x2x0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_x2x0

      ! ezfloat(:,:) / exfloat
      FUNCTION exflib_float_div_z2x0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_z2x0

      ! int(:,:) / exfloat(:,:)
      FUNCTION exflib_float_div_i2x2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_i2x2

      ! quadword(:,:) / exfloat(:,:)
      FUNCTION exflib_float_div_q2x2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_q2x2

      ! exfloat(:,:) / exfloat(:,:)
      FUNCTION exflib_float_div_x2x2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x2x2

      ! ezfloat(:,:) / exfloat(:,:)
      FUNCTION exflib_float_div_z2x2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z2x2

      ! exfloat / int(:,:)
      FUNCTION exflib_float_div_x0i2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x0i2

      ! exfloat / quadword(:,:)
      FUNCTION exflib_float_div_x0q2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x0q2

      ! exfloat / ezfloat(:,:)
      FUNCTION exflib_float_div_x0z2(a, b) RESULT(c)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x0z2

      ! exfloat(:,:) / int
      FUNCTION exflib_float_div_x2i0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_x2i0

      ! exfloat(:,:) / quadword
      FUNCTION exflib_float_div_x2q0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_x2q0

      ! exfloat(:,:) / ezfloat
      FUNCTION exflib_float_div_x2z0(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_x2z0

      ! exfloat(:,:) / int(:,:)
      FUNCTION exflib_float_div_x2i2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x2i2

      ! exfloat(:,:) / quadword(:,:)
      FUNCTION exflib_float_div_x2q2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x2q2

      ! exfloat(:,:) / ezfloat(:,:)
      FUNCTION exflib_float_div_x2z2(a, b) RESULT(c)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_x2z2

      ! int / ezfloat(:,:)
      FUNCTION exflib_float_div_i0z2(a, b) RESULT(c)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_i0z2

      ! quadword / ezfloat(:,:)
      FUNCTION exflib_float_div_q0z2(a, b) RESULT(c)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_q0z2

      ! ezfloat / ezfloat(:,:)
      FUNCTION exflib_float_div_z0z2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z0z2

      ! int(:,:) / ezfloat
      FUNCTION exflib_float_div_i2z0(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_i2z0

      ! quadword(:,:) / ezfloat
      FUNCTION exflib_float_div_q2z0(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_q2z0

      ! ezfloat(:,:) / ezfloat
      FUNCTION exflib_float_div_z2z0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_z2z0

      ! int(:,:) / ezfloat(:,:)
      FUNCTION exflib_float_div_i2z2(a, b) RESULT(c)
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_i2z2

      ! quadword(:,:) / ezfloat(:,:)
      FUNCTION exflib_float_div_q2z2(a, b) RESULT(c)
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_q2z2

      ! ezfloat(:,:) / ezfloat(:,:)
      FUNCTION exflib_float_div_z2z2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z2z2

      ! ezfloat / int(:,:)
      FUNCTION exflib_float_div_z0i2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z0i2

      ! ezfloat / quadword(:,:)
      FUNCTION exflib_float_div_z0q2(a, b) RESULT(c)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(b,1),SIZE(b,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(b,2)
           DO i = 1, SIZE(b,1)
              c(i,j) = a / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z0q2

      ! ezfloat(:,:) / int
      FUNCTION exflib_float_div_z2i0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_z2i0

      ! ezfloat(:,:) / quadword
      FUNCTION exflib_float_div_z2q0(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b
           END DO
        END DO
      END FUNCTION exflib_float_div_z2q0

      ! ezfloat(:,:) / int(:,:)
      FUNCTION exflib_float_div_z2i2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z2i2

      ! ezfloat(:,:) / quadword(:,:)
      FUNCTION exflib_float_div_z2q2(a, b) RESULT(c)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, DIMENSION(:,:), INTENT(IN) :: b
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: c
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              c(i,j) = a(i,j) / b(i,j)
           END DO
        END DO
      END FUNCTION exflib_float_div_z2q2

      !----------------------------------------
      ! power
      !----------------------------------------
      ! pow( exfloat(:,:), int )
      FUNCTION exflib_float_pow_x2i0(a, n) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: n
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(i,j) = a(i,j) **n
           END DO
        END DO
      END FUNCTION exflib_float_pow_x2i0

      ! pow( exfloat(:,:), quadword )
      FUNCTION exflib_float_pow_x2q0(a, n) RESULT(s)
        TYPE(exfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: n
        TYPE(exfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(i,j) = a(i,j) ** n
           END DO
        END DO
      END FUNCTION exflib_float_pow_x2q0

      ! pow( ezfloat(:,:), int )
      FUNCTION exflib_float_pow_z2i0(a, n) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: n
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(i,j) = a(i,j) **n
           END DO
        END DO
      END FUNCTION exflib_float_pow_z2i0

      ! pow( ezfloat(:,:), quadword )
      FUNCTION exflib_float_pow_z2q0(a, n) RESULT(s)
        TYPE(ezfloat), DIMENSION(:,:), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: n
        TYPE(ezfloat), DIMENSION(SIZE(a,1),SIZE(a,2)) :: s
        INTEGER*8 :: i, j
        DO j = 1, SIZE(a,2)
           DO i = 1, SIZE(a,1)
              s(i,j) = a(i,j) ** n
           END DO
        END DO
      END FUNCTION exflib_float_pow_z2q0

    END MODULE exflib_array

!**********************************************************************
! End of file
!**********************************************************************
